#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/environment_setup'
require 'mxx_ru/ex'

require 'mxx_ru/util'

module MxxRu

  # File name of root project file.
  BUILD_ROOT = "build.rb"

  # Target status description.
  class TargetState
    # Target is absent or rebuild is required.
    ABSENT = 0
    # Target was rebuilt.
    REBUILT = 1
    # Target exists and up-to-date.
    EXISTS = 2

    attr_reader :state

    def initialize( a_state = ABSENT )
      @state = a_state
    end

    # Detect target status.
    # TargetState object returned.
    #
    # [_name_] Target name.
    # [_depends_] Dependencies list. Array of String.
    def TargetState.detect( name, depends )

      state = ABSENT

      # In dry-run mode we always return that target is absent.
      if !MxxRu::Util::Mode.instance.is_dry_run
        if FileTest.exist?( name )
          # It's necessary to detect time of target and it's dependencies.
          target_mtime = File.mtime( name )
          if nil != depends.find { |d|
              if !FileTest.exist?( d )
                raise MxxRu::MissingDependFileEx.new( d, name )
              end
              depend_mtime = File.mtime( d )
              target_mtime < depend_mtime
            }
            # More young dependency found.
            state = REBUILT
          else
            # Target is up-to-date.
            state = EXISTS
          end
        end
      end

      return TargetState.new( state )
    end

    # Checks collection of statuses.
    #
    # Returns true if at least one of statuses have value REBUILT or ABSENT.
    #
    # Typical usage is:
    #   subprojects_status = build_subprojects
    #   self_files_status = build_self_files
    #
    #   if TargetState.rebuilding_needed?(
    #       subprojects_status, self_files_status )
    #     do_rebuild
    #   end
    #
    # Since v.1.4.4
    #
    def TargetState.rebuilding_needed?( *statuses )
      nil != statuses.find { |s| EXISTS != s.state }
    end

    # Creates and returns objects of state ABSENT
    #
    # Since v.1.4.4
    #
    def TargetState.make_absent
      TargetState.new( ABSENT )
    end

  end # Class TargetState

  # For compatibility with previous versions.
  Target_state = TargetState

  # Map of targets loaded.
  # Key is target alias -- file name with definition of a target.
  @@defined_targets = Hash.new

  # Root target alias.
  # This target will be used in setup_target function to start building process.
  @@first_target_alias = nil

  def MxxRu.try_set_first_target_alias( a_alias )
    if nil == @@first_target_alias
      @@first_target_alias = a_alias.clone
    end
  end

  def MxxRu.setup_target( target, build_enabled = true )
    begin
      if !( target_defined_for?( target.prj_alias ) )
        @@defined_targets[ target.prj_alias ] = target;
      end

      if build_enabled && nil != @@first_target_alias &&
        @@first_target_alias == target.prj_alias
        # It is a root project, required to build.

        # If this project is not root project file and root project file
        # exists, then we will load root project file and start build process
        # after that.
        if BUILD_ROOT != target.prj_alias
          if FileTest.exist?( BUILD_ROOT )
            require BUILD_ROOT
          end
        end

        target.clean if MxxRu::Util::Mode.instance.is_clean or
            MxxRu::Util::Mode.instance.is_rebuild

        target.reset if MxxRu::Util::Mode.instance.is_rebuild

        MxxRu::Util::build_call_wrapper( target ) if
            !MxxRu::Util::Mode.instance.is_clean
      end

      return target
    rescue MxxRu::Ex => ex
      $stderr.print "<<<[#{ex.class.name}]\t#{ex}>>>\n"
      exit( -2 )
    end
  end

  def MxxRu.query_target( prj_alias, mandatory )
    target = @@defined_targets[ prj_alias ]

    # If target was required, we already should found it.
    # Build should be stopper otherwise.
    if target.nil? && mandatory
      raise TargetUnknownEx.new( prj_alias )
    end

    return target
  end

  def MxxRu.target_defined_for?(prj_alias)
    return true if @@defined_targets.has_key?( prj_alias )
    return false
  end

  # Base class for all source code generators.
  class AbstractGenerator
    # This method is called to generate source code required.
    # Should be redefined in a child class.
    #
    # [_target_] Target, generator belongs to.
    def build( target )
      raise AbstractMethodEx.new( "MxxRu::AbstractGenerator::build" )
    end

    # This method is called to cleanup of generated source code files.
    #
    # [_target_] Target, generator belongs to.
    def clean( target )
      raise AbstractMethodEx.new( "MxxRu::AbstractGenerator::clean" )
    end
  end

  # For compatibility with previous versions.
  Abstract_generator = AbstractGenerator

  # Base class for all targets.
  class AbstractTarget
    # List of full file names, created during a build of the target.
    # It may be an empty array if target doesn't create any files.
    # For example, if target is a composite or unittest.
    attr_reader :mxx_full_targets_names

    # Vector of subordinated projects.
    attr_reader :mxx_required_prjs

    # Vector of source code generators.
    attr_reader :mxx_generators

    # Metacode for generating 'plural' version of singular method.
    # Plural version accept Enumerable-type argument and call
    # singular version for each member of Enumerable argument.
    #
    # For example:
    #  define_plural_form_method :required_prj
    #  define_plural_form_method :cpp_source
    # produces methods
    #   required_prjs( items )
    #   cpp_sources( items )
    #
    def AbstractTarget.define_plural_form_method( singular_method )
      class_eval %Q{
        def #{singular_method}s(a)
          a.each { |e| #{singular_method}( e ) }
        end
      }
    end

    def initialize( a_prj_alias )
      @mxx_full_targets_names = Array.new
      @mxx_prj_alias = a_prj_alias
      @mxx_required_prjs = Array.new
      @mxx_generators = Array.new

      MxxRu.try_set_first_target_alias( a_prj_alias )
    end

    # Add one more file, created during a target build.
    def mxx_add_full_target_name( a_full_name )
      @mxx_full_targets_names << a_full_name
    end

    def prj_alias
      return @mxx_prj_alias
    end

    # Add one more project in a list of subordinated projects.
    #
    # Returns target object, which is defined in subordinated project.
    def required_prj( a_prj )
      # If project is not loaded yet, loading it.
      # Then adding created target object to the list ofd subordinated projects.
      # If no target will be created, exception would be thrown.
      if !MxxRu::target_defined_for?( a_prj )
        # expand_path is called because $: in Ruby 1.9.2 doesn't contain
        # the current path. So the form "required_prj 'some/project/prj.rb'"
        # doesn't work in Ruby 1.9.2.
        require File.expand_path( a_prj )
      end

      target = MxxRu::query_target( a_prj, true )
      @mxx_required_prjs << target

      return target
    end

    # Plural form of required_prj.
    #
    # For example:
    #  required_prjs [ 'module1/prj.rb', 'module2/prj.rb' ]
    #  required_prjs Dir.glob( 'modules/**/prj.rb' )
    define_plural_form_method :required_prj

    # Add one more generator to the target.
    # Reference to the target is returned.
    def generator( a_generator )
      @mxx_generators << a_generator
      return a_generator
    end

    # Child classes should redefine this method.
    # TargetState object is returned.
    def build
      raise AbstractMethodEx.new( "AbstractTarget::build" )
    end

    # Child classes should redefine this method.
    def clean
      raise AbstractMethodEx.new( "AbstractTarget::clean" )
    end

    # Child classes should redefine this method.
    # Base class does not implement it.
    def reset
    end

    # Executing command line given.
    #
    # [a_cmd_lines] Array of String
    # [a_to_destroy_on_fail] Files required to be cleaned up if any of given
    #                        command fail.
    # [brief_desc] Brief description of command (will be shown if
    #              --mxx-brief-show specified in command line)
    def AbstractTarget.run(
        a_cmd_lines, a_to_destroy_on_fail, brief_desc = nil )

      MxxRu.show_brief( brief_desc )

      a_cmd_lines.each { |c|
          puts "<<< #{c} >>>" \
              if MxxRu::Util::Mode.instance.is_show_cmd

          # Do not actually execute in dry-run mode.
          if !MxxRu::Util::Mode.instance.is_dry_run
            if !system( c )
              a_to_destroy_on_fail.each { |d|
                MxxRu::Util::delete_file( d ) }
              raise BuildEx.new( c, $? )
            end
          end
        }
    end

  protected
    # Build of all projects required for the target.
    # Returns MxxRu::TargetState.
    def build_required_projects
      state = MxxRu::TargetState::EXISTS
      @mxx_required_prjs.each { |p|
          prj_state = MxxRu::Util::build_call_wrapper( p )
          if MxxRu::TargetState::EXISTS != prj_state.state
            state = prj_state.state
          end
        }

      return MxxRu::TargetState.new( state )
    end

    # Cleanup of all projects required for the target.
    # Executed only if dry-run mode is disabled.
    def clean_required_prjs
      if !MxxRu::Util::Mode.instance.is_dry_run
        @mxx_required_prjs.each { |p| p.clean }
      end
    end

    # Reset build status for all projects required for the target.
    def reset_required_projects
      @mxx_required_prjs.each { |p| p.reset }
    end

  end

  # For compatibility with previous versions.
  Abstract_target = AbstractTarget

end
