#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/environment_setup'
require 'mxx_ru/ex'

module MxxRu

# Class for description of binary library required to be linked
# with binary target.
#
# Since 1.3.0
#
# On Unix platform static and shared library can be specified to linker with
# same name. For example to link libA.a or libA.so user must specify '-lA'
# option. Linker selects appropriate library (static or shared version) by
# it own rules (on most Linux platform ld first looks for shared version, than
# for static version; this approach can be overriden by '-static' option for
# gcc or '-Bstatic' for ld).
#
# In some cases user may wants to explicitly specify kind of library. For
# example, if ld by default searches shared libraries than user can specify
# sequence '-Bstatic -lA -Bdynamic' for ld (or '-Wl,-Bstatic,-lA,-Bdynamic'
# for gcc). This forces linker for looking only static version of library.
#
# In version 1.3.0 Mxx_ru support for explicit static/shared library selection
# was added. To implement this support required library description was
# changed. Instead of storing names of required libraries as String they are
# stored as BinaryLibrary objects.
class BinaryLibrary
  # Identifier for static library type.
  STATIC = :static
  # Identifier for shared library type.
  SHARED = :shared
  # Identifier for library which type if unknown yet (not explicitly specified).
  ANY = :any

  # Name of library.
  attr_reader :name
  # Type of library (STATIC/SHARED/ANY).
  attr_reader :type

  def initialize( name, type )
    raise InvalidValueEx.new( "Unknown BinaryLibrary type: #{type}" ) unless
        [ STATIC, SHARED, ANY ].member?( type )

    @name = name
    @type = type
  end

  def to_s
    @name
  end

  def eql?( another )
    @name.eql?( another.name ) and @type.eql?( another.type )
  end

  def ==( another )
    self.eql?( another )
  end

  def <=>( another )
    r = @name <=> another.name
    r = @type <=> another.type if 0 == r
    r
  end

end # class BinaryLibrary

# Exception for the case where conflict of BinaryLibrary usage detected.
#
# Conflict detected if STATIC and SHARED library with same name specified
# for one target.
class BinaryLibraryTypeConflictEx < Ex
  def initialize( target_name, library_name )
    super( "Binary library conflict with library '#{library_name}' for " +
        "target '#{target_name}'" )
  end
end

end # module MxxRu
