#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/abstract_target'
require 'mxx_ru/binary_library'
require 'mxx_ru/ex'

# Base class for all targets, which are binary executables, binary shared
# libraries or binary static libraries.
#
# In version v.1.3 extended by methods lib_path, lib_paths which are
# user-frendly synonyms for mxx_add_required_lib_path.
module MxxRu

  class BinaryTarget < MxxRu::AbstractTarget
    def initialize( a_prj_alias )
      super( a_prj_alias )

      # Libraries list, which would be required to the user of current target.
      @mxx_required_libs = Array.new
      # Attribute showing that before return from mxx_required_libs
      # dublicates should be removed.
      @mxx_required_libs_changed = true
      # Folder list, where libraries, encountered in mxx_required_libs, 
      # should be searched for. Also contain paths specified by
      # lib_path, lib_paths methods.
      @mxx_required_lib_paths = Array.new
      # Attribute showing that before return from mxx_required_lib_paths
      # dublicates should be removed.
      @mxx_required_lib_paths_changed = true
    end

    # Add the library required.
    #
    # Type of library must be detect implicitly.
    #
    # If a_path is other then nil, then it's value is added to the list
    # of a folders, the libraries should be searched in.
    def lib( a_library, a_path = nil )
      mxx_add_required_lib(
          BinaryLibrary.new( a_library, BinaryLibrary::ANY ) )
      if a_path
        mxx_add_required_lib_path( a_path )
      end
    end

    # Add static library required.
    #
    # If a_path is other then nil, then it's value is added to the list
    # of a folders, the libraries should be searched in.
    def lib_static( a_library, a_path = nil )
      mxx_add_required_lib(
          BinaryLibrary.new( a_library, BinaryLibrary::STATIC ) )
      if a_path
        mxx_add_required_lib_path( a_path )
      end
    end

    # Add shared library required.
    #
    # If a_path is other then nil, then it's value is added to the list
    # of a folders, the libraries should be searched in.
    def lib_shared( a_library, a_path = nil )
      mxx_add_required_lib(
          BinaryLibrary.new( a_library, BinaryLibrary::SHARED ) )
      if a_path
        mxx_add_required_lib_path( a_path )
      end
    end

    # Get all libraries list, which are required to be linked to for
    # correct usage of given target.
    # Returns array of strings.
    def mxx_required_libs
      if @mxx_required_libs_changed
        @mxx_required_libs.flatten!
        BinaryTarget.check_libraries_types( prj_alias, @mxx_required_libs )
        @mxx_required_libs_changed = false
      end
      return @mxx_required_libs
    end

    # Add library to list of requirements.
    def mxx_add_required_lib( a_lib )
      a_lib = BinaryLibrary.new( a_lib, BinaryLibrary::ANY ) if
          a_lib.kind_of?( String )

      @mxx_required_libs << a_lib
      @mxx_required_libs_changed = true
    end

    # Plural form of mxx_add_required_lib.
    define_plural_form_method :mxx_add_required_lib

    # Get list of all paths to libraries, which should be used to search
    # libraries, returned by mxx_required_libs method.
    # Returns array of strings.
    def mxx_required_lib_paths
      if @mxx_required_lib_paths_changed
        @mxx_required_lib_paths.flatten!
        @mxx_required_lib_paths_changed = false
      end
      return @mxx_required_lib_paths
    end

    # Add folder to search folders list.
    def mxx_add_required_lib_path( a_path )
      @mxx_required_lib_paths << a_path
      @mxx_required_lib_paths_changed = true
    end

    # Plural form of mxx_add_required_lib_path.
    define_plural_form_method :mxx_add_required_lib_path

    alias_method :lib_path, :mxx_add_required_lib_path

    # Plural form of lib_path.
    define_plural_form_method :lib_path

    # Checks types of libraries. Should not exists two instance of
    # same library with STATIC and SHARED types at one time.
    #
    # Raises BinaryLibraryTypeConflictEx if conflict detected.
    def BinaryTarget.check_libraries_types( prj_alias, libs )
      if libs.size
        names_and_types = Hash.new( BinaryLibrary::ANY )
        libs.each { |l|
          k = names_and_types[ l.name ]
          if BinaryLibrary::ANY != l.type
          
            if BinaryLibrary::ANY != k
              raise BinaryLibraryTypeConflictEx.new( prj_alias, l.name ) if
                  k != l.type
            else
              names_and_types[ l.name ] = l.type
            end

          end
        }
      end
    end

  end

  # For compatilibily with previous versions.
  Binary_target = BinaryTarget

end
