#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/ex'
require 'mxx_ru/abstract_target'

module MxxRu

# Class describing a target, which is unit-test binary application.
#
# The idea is, that there is project file, which controls the build process
# of unit-test application. In that project file target object is created,
# inherited from MxxRu::BinaryTarget.
# To initialize unit-test it's required to create one more file, where
# target object of MxxRu::BinaryUnittestTarget class is created.
# For example:
#
# File to compile unit-test.
#  MxxRu::setup_target(
#    MxxRu::Cpp::ExeTarget.new( "test/pack/prj.rb" ) {
#      ...
#    }
#  )
#
# File to run unit-test.
#  MxxRu::setup_target(
#    MxxRu::BinaryUnittestTarget.new( 
#      "test/pack/ut.rb",
#      "test/pack/prj.rb ) )
# 
# File to use unit-test.
#  MxxRu::setup_target(
#    MxxRu::Cpp::CompositeTarget.new( MxxRu::BUILD_ROOT ) {
#      required_prj( "some/project/prj.rb" )
#      required_prj( "test/pack/ut.rb" )
#    }
#  )
#
class BinaryUnittestTarget < AbstractTarget

  # Exception, thrown if subordinated project created more then one result
  # name.
  class MustBeOneTargetNameEx < Ex
    # [_a_names_] Generated result names.
    def initialize( a_names )
      super( "Only one name expected in [#{a_names.join(';')}]" )
    end
  end

  # For compatibility with previous versions.
  Must_be_one_target_name_ex = MustBeOneTargetNameEx

  # Target, responsible for build of unit-test application.
  attr_reader :mxx_target_project
  # Attribute, showing that build method was already executed.
  attr_reader :mxx_build_state

  # Constructor.
  #
  # [_a_alias_] Self alias.
  # [_a_target_project_] Project, responsible for build of unit-test
  #                      application.
  def initialize( a_alias, a_target_project )
    super( a_alias )

    @mxx_build_state = nil

    @mxx_target_project = required_prj( a_target_project )

    # By default the expected result of binary command is zero.
    # Kernel#system will return 'true' in that case.
    # In the case of NegativeBinaryUnittestTarget this value
    # must be changed to 'false'
    @expected_system_retval = true
  end

  # Build subordinated project and run it.
  #
  # If unit-test application returns result, other then 0,
  # exception is thrown.
  def build
    if !@mxx_build_state
      @mxx_target_project.build

      # Determining application name.
      full_names = @mxx_target_project.mxx_full_targets_names
      if 1 != full_names.size
        raise MustBeOneTargetNameEx.new( full_names )
      end

      # In dry-run do not actually execute anything.
      if !MxxRu::Util::Mode.instance.is_dry_run
        puts "running unit test: #{full_names[0]}..."
        if @expected_system_retval != system( full_names[ 0 ] )
          puts "\n\nunit test '#{full_names[0]}' FAILED! #{$?}"
          raise BuildEx.new( full_names[ 0 ], $? )
        end
      end

      @mxx_build_state = TargetState.new( TargetState::REBUILT )
    end

    return @mxx_build_state
  end

  # Project cleanup.
  #
  # Just execute clean method from subordinated project.
  def clean
    @mxx_target_project.clean
  end

  # Reset build state.
  def reset
    @mxx_target_project.reset
    @mxx_build_state = nil
  end

end

# For compatibility with previous versions.
Binary_unittest_target = BinaryUnittestTarget 

# Class describing a target, which is unit-test binary application
# with non-zero exit code.
#
# This class is similar to BinaryUnittestTarget but intended for the
# cases where binary unit test expected to fail (sometimes failure is
# a expected result).
#
# File to run unit-test.
#  MxxRu::setup_target(
#    MxxRu::NegativeBinaryUnittestTarget.new( 
#      "test/pack/ut.rb",
#      "test/pack/prj.rb ) )
# 
#
class NegativeBinaryUnittestTarget < BinaryUnittestTarget

  # Constructor.
  #
  # [_a_alias_] Self alias.
  # [_a_target_project_] Project, responsible for build of unit-test
  #                      application.
  def initialize( a_alias, a_target_project )
    super( a_alias, a_target_project )

    @expected_system_retval = false
  end
end

end # module MxxRu
