#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'optparse'
require 'singleton'

require 'mxx_ru/version'

module MxxRu

# Singleton for centralized mechanizm of command line options parsing.
#
# Since v.1.4.0
#
# Each part of Mxx_ru which need some command line option must provider
# special option processor for CmdLineOptionProcessor singleton:
#
#   module MxxRu
#   module SomeLangSupport
#     class OptionParser
#       include Singleton
#       ...
#       def prepare( parser ) # Instance of OptionParser from stdlib.
#         parser.separator ''
#         parser.separator 'Some Language Support options:'
#
#         parser.on( '-i', '--include PATH', 'Set up include path' ) do
#           |path|
#           ... some processig ...
#         end
#         ... setting another option handlers ...
#       end
#     end
#   end
#   end
#
#   # OptionParser must be added to CmdLineOptionProcessor.
#   MxxRu::CmdLineOptionProcessor.add_processor(
#       MxxRu::SomeLangSupport::OptionParser.instance )
#
# See examples of using this technique in MxxRu::Util::Mode and
# MxxRu::Cpp::Mode classes.
#
class CmdLineOptionProcessor
  include Singleton

  def initialize
    @processors = []
    @parsed = false
  end

  # Add yet another option processor for processors list.
  def add_processor( processor )
    @processors << processor
    @parsed = false
  end

  # Performs parsing of ARGV content by current processors.
  #
  # Finishes work of script if --help or --version found in ARGV.
  def parse
    unless @parsed
      parser = OptionParser.new do |p|
        prepare_option_parser( p )
      end

      begin
        parser.parse( ARGV )
      rescue => x
        $stderr.puts "command line parsing error '#{x.message}'"
        exit( 2 )
      end

      @parsed = true
    end
  end

protected 
  # Calls preparse() method for each processor. And then add common
  # command line options to the parser.
  def prepare_option_parser( parser )
    @processors.each do |p| p.prepare( parser ) end

    parser.separator ''
    parser.separator 'Auxilary options:'

    parser.on_tail( '-h', '--help', 'Show usage help and options summary' ) do
      puts parser
      exit
    end

    parser.on_tail( '--version', 'Show version' ) do
      puts 'Mxx_ru v.' + MXX_RU_VERSION
      exit
    end
  end

end

end

