#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2018, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'singleton'
require 'pathname'

module  MxxRu
  module  Cpp

    module  Analyzer
      # Class, describing one source file.
      class SourceFile
        # Already processed file sign.
        attr_reader :is_processed
        # A set of files-dependecies.
        # Hash, key is a String, no value.
        attr_reader :depends

        def initialize
          @is_processed = false
          @depends = Hash.new
        end

        # Set processed file sign.
        def processed
          @is_processed = true
        end

        # Add a dependency.
        def <<( a_name )
          @depends[ a_name ]= nil
        end

      end # class SourceFile

      # For compatibility with previous versions.
      Source_file = SourceFile

      class CppAnalyzer
        include Singleton

        @@include_regex = Regexp.new(
          '^[ \t]*#[ \t]*include[ \t]*(<|")([^>"]+)(>|")' )

        def initialize
          # Key is a string, value - SourceFile.
          @sources = Hash.new
          # A list of folders to look header files in.
          @include_paths = Array.new
        end

        # Add files to view.
        # [_a_sources_] Should be Array of Strings.
        def add_sources( a_sources )

          a_sources.each { |s|
              if !@sources.key?( s )
                @sources[ s ] = SourceFile.new
              end
            }
        end

        # Add paths to header files.
        # [_a_paths_] Should be Array of Strings.
        def add_include_paths( a_paths )
          a_paths.each { |p|
              if !@include_paths.include?( p )
                @include_paths << p
              end
            }
        end

        # Perform file analisys.
        def analyze ( a_encoding )
          # Current list of keys should be taken to avoid 
          # dynamic growth of file list.
          # Repeat iterations until source file list changes stop      
          old_size = 0
          files = @sources.keys
          while old_size != files.size
            files.each { |file_name|
                current = @sources[ file_name ]
                if !current.is_processed
                  process_file( file_name, current, a_encoding )
                end
              }

            old_size = files.size
            files = @sources.keys
          end
        end

        # Get full dependencies list for a file.
        # Returns Array of Strings.
        def get_depends_for( a_file_name )
          result_hash = Hash.new
          file_obj = @sources[ a_file_name ]
          if nil != file_obj
            to_review = file_obj.depends
            to_review_next = Hash.new

            while to_review.size != to_review_next.size
              to_review.each_key { |f|
                  if !result_hash.key?( f )
                    result_hash[ f ] = nil
                    to_review_next_obj = @sources[ f ]
                    to_review_next_obj.depends.each_key { |k|
                        to_review_next[ k ] = nil }
                  end
                }

              to_review = to_review_next
              to_review_next = Hash.new
            end
          end

          return result_hash.keys
        end

      protected
        # Process next file.
        # [_a_file_name_] File name.
        # [_a_file_obj_] Corresponding object.
        def process_file( a_file_name, a_file_obj, a_encoding )

          # Gather file names list from include directives.
          std_path_included = Array.new
          local_path_included = Array.new
          IO.foreach( a_file_name , encoding: "#{a_encoding}" ) { |line|
              r = @@include_regex.match( line )
              if nil != r
                if "<" == r[ 1 ]
                  std_path_included << r[ 2 ]
                else
                  local_path_included << r[ 2 ]
                end
              end
            }

          # All files, loaded through include <>
          # should be searched in standard search paths only.
          std_path_included.each { |f|
              process_std_include_path_file( f, a_file_obj )
            }

          if local_path_included.size
            # Files loaded through include ""
            # should be searched in folder with the source file first.
            local_path = File.dirname( a_file_name )
            local_path_included.each { |f|
                process_local_include_path_file( f, local_path, a_file_obj )
              }
          end

          a_file_obj.processed
        end

        # Processing file loaded through include <>
        # [_a_file_name_] File name.
        # [_a_owner_obj_] File object contained include directive.
        def process_std_include_path_file( a_file_name, a_owner_obj )
          @include_paths.find { |path|
              test_name = File.join( [ path, a_file_name ] )
              if File.exists?( test_name )
                add_depend_to( a_owner_obj, test_name )
                true
              else
                false
              end
            }
        end

        # Processing file loaded through include ""
        # [_a_file_name_] File name.
        # [_a_local_path_] Folder with owner_obj.
        # [_a_owner_obj_] File object contained include directive.
        def process_local_include_path_file(
          a_file_name,
          a_local_path,
          a_owner_obj )

          test_name = # This trick is necessary to handle relative includes
                      # like ../details/some_header.h.
                      # Without calling cleanpath there could be created many
                      # names in form local/path/details/../details/../details
            Pathname.new(
              File.join( a_local_path, a_file_name ) ).cleanpath
          if File.exists?( test_name )
            add_depend_to( a_owner_obj, test_name )
          else
            process_std_include_path_file( a_file_name, a_owner_obj )
          end
        end

        # Add dependency to a file.
        # [_a_owner_obj_] Dependency owner.
        # [_a_depend_name_] Dependency file name.
        def add_depend_to( a_owner_obj, a_depend_name )

          a_owner_obj << a_depend_name

          if !@sources.key?( a_depend_name )
            # This file wasn't viewed yet.
            @sources[ a_depend_name ] = SourceFile.new
          end
        end
      end # class CppAnalyzer

      # For compatibility with previous versions.
      Cpp_analyzer = CppAnalyzer

    end # module Analyzer
  end # module cpp
end # module MxxRu

load File.dirname( __FILE__ ) + '/../compat.rb'
