#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2019, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'rbconfig'

require 'mxx_ru/cpp/toolset'

module MxxRu
  module Cpp
    ENV_VAR = "MXX_RU_CPP_TOOLSET"

    # Detect current toolset based on values of environment variables
    #
    # Following algorithm is used:
    # Getting the value of MXX_RU_CPP_TOOLSET envonment variable.
    # It should be in a following format:
    #  <file> [tag=value [tag=value [...]]
    # <file> value is used in require directive.
    #
    # After loading of file defined, toolset method should return
    # a pointer to toolset object.
    #
    # If toolset object was created, pairs of tag=value are parsed and
    # saved to the toolset created using a setup_tag method.
    def Cpp.detect_toolset
      begin
        env = get_or_detect_toolset_variable_content

        # Exctracting file name, responsible for toolset object creation.
        toolset_file_name_regexp = /^(\S+)(.*)$/
        file_name_match = toolset_file_name_regexp.match( env )
        if nil == file_name_match
          raise MxxRu::InvalidValueEx.new(
            "Invalid format of environment variable '#{env_var}'. " +
            "Expected format: #{env_var}=<file>[ <tag>=<value>[ " +
            "<tag>=<value>[ ...]]]" )
        end

        file_name = file_name_match[ 1 ]
        unparsed_tags = file_name_match[ 2 ]

        # Loading the file
        # Try to loading specified file first, if it fails try to
        # loading file with 'mxx_ru/cpp/toolsets' added...
        try_load_toolset_file file_name

        if nil == toolset()
          raise MxxRu::InvalidValueEx.new(
            "C/C++ Toolset is not set by file '#{file_name}'" )
        end

        # Setting up toolset tags
        setup_toolset_tags( toolset(), unparsed_tags )

      rescue MxxRu::Ex => ex
        $stderr.print "<<<[#{ex.class.name}]\t#{ex}>>>\n"
        exit( -1 )
      end
    end

    # [_a_toolset_] Toolset, tags are set for
    # [_a_unparsed_tags_] A string, formatted that way:
    #                     <tag>=<value>[ <tag>=<value> [<tag>=<value> ...]]
    def Cpp.setup_toolset_tags( a_toolset,
      a_unparsed_tags )

      a_unparsed_tags.scan( /(\S+)=(\S+)/ ) { |t, v|
        a_toolset.setup_tag( t, v )
      }

    end

    # Try load toolset file (without and with 'mxx_ru/cpp/toolsets' prefix).
    #
    # Raises LoadError if loading failed
    def Cpp.try_load_toolset_file( file_name )
      files_to_load = [ file_name,
          File.join( 'mxx_ru/cpp/toolsets', file_name ) ]
      load_result = files_to_load.detect { |file_to_load|
          begin
            require file_to_load
          rescue LoadError
            false
          end
        }
      raise MxxRu::InvalidValueEx.new(
        "Unable to load Toolset from files: " +
          "'#{files_to_load.join('; ')}'" ) unless load_result
    end

    # Gets or detects value for MXX_RU_CPP_TOOLSET.
    #
    # Since v.1.5.1
    #
    def Cpp.get_or_detect_toolset_variable_content
      env = ENV[ ENV_VAR ]
      if env.nil?
        $stderr.puts "Environment variable #{ENV_VAR} not found. " +
            "Try to detect toolset by HOST_OS and/or PATH..."
        env = try_detect_toolset_myself
        if env
          $stderr.puts "Detection successed. Toolset is: #{env}"
        else
          $stderr.puts "Detection failed."
          raise MxxRu::InvalidValueEx.new(
            "Environment variable '#{ENV_VAR}' not found (not set " +
            "or empty" )
        end
      end

      env
    end

    # Try to detect toolset from host_os and PATH.
    #
    # Since v.1.5.1
    #
    # Only Windows platform and MS C++, Linux and GCC is detected now.
    #
    # Returns name of toolset of throw exception if toolset is not detectable.
    def Cpp.try_detect_toolset_myself
      if RbConfig::CONFIG['host_os'] =~ /linux/
        # Assume that GCC is default compiler on Linux.
        "gcc_linux"
      elsif RbConfig::CONFIG['host_os'] =~ /^freebsd(9|10|11|12)$/
        "clang_freebsd"
      else
        if ENV[ 'COMP_ENV' ] and ENV[ 'VC_ARCH' ]
          'icc_win'
        elsif ENV[ 'VS160COMNTOOLS' ]
          'vc16'
        elsif ENV[ 'VS140COMNTOOLS' ]
          'vc14'
        elsif ENV[ 'VS120COMNTOOLS' ]
          'vc12'
        elsif ENV[ 'VS110COMNTOOLS' ]
          'vc11'
        elsif ENV[ 'VS100COMNTOOLS' ]
          'vc10'
        elsif ENV[ 'VS90COMNTOOLS' ]
          'vc9'
        elsif ENV[ 'VS80COMNTOOLS' ]
          'vc8'
        elsif ENV[ 'VS71COMNTOOLS' ]
          'vc7'
        end
      end
    end

  end # module Cpp
end # module MxxRu

