#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'singleton'

require 'mxx_ru/cmd_line_option_processor'

module  MxxRu
module  Cpp

  # Arguments passed into Ruby interpreter, intended to control MxxRu::Cpp.
  MXXCPPARG_RELEASE = "--mxx-cpp-release"
  MXXCPPARG_DEBUG = "--mxx-cpp-debug"
  MXXCPPARG_ONLY_ONE = "--mxx-cpp-1"
  MXXCPPARG_NO_DEPENDS_ANALYZER = "--mxx-cpp-no-depends-analyzer"
  MXXCPPARG_EXTRACT_OPTIONS = "--mxx-cpp-extract-options"

  # Class, detecting a mode of MxxRu::Cpp is running
  class Mode
    include Singleton

    # Option parser for C++ specific Mxx_ru options.
    #
    # Since v.1.4.0
    class OptionParser
      include Singleton

      attr_reader :is_release
      attr_reader :is_debug
      attr_reader :is_only_one
      attr_reader :is_no_depends_analyzer
      attr_reader :is_option_extraction

      def initialize
        @is_release = false
        @is_debug = false
        @is_only_one = false
        @is_no_depends_analyzer = false
        @is_option_extraction = false
      end

      def prepare( parser )
        parser.separator ''
        parser.separator 'C++ mode options:'

        parser.on( MXXCPPARG_RELEASE, 'Compile in release mode' ) do
          @is_release = true
        end

        parser.on( MXXCPPARG_DEBUG, 'Compile in debug mode' ) do
          @is_debug = true
        end

        parser.on( MXXCPPARG_ONLY_ONE,
            'Process only one project, without required projects' ) do
          @is_only_one = true
        end

        parser.on( MXXCPPARG_NO_DEPENDS_ANALYZER,
            'Do not use C++ dependencies analyzer' ) do
          @is_no_depends_analyzer = true
        end

        parser.on( MXXCPPARG_EXTRACT_OPTIONS,
            'Only extract and show tools (like compiler, linker) options' ) do
          @is_option_extraction = true
        end
      end

    end

    # Method for generation methods like is_clean, is_rebuild, etc.
    def Mode.generate_option_reader( option )
      class_eval do
        define_method( option ) do
          OptionParser.instance.send( option )
        end
      end
    end

    generate_option_reader :is_release
    generate_option_reader :is_debug
    generate_option_reader :is_only_one
    generate_option_reader :is_no_depends_analyzer
    generate_option_reader :is_option_extraction

    # Constructor checks for a special arguments in a command line
    def initialize
      CmdLineOptionProcessor.instance.parse

      @manual_dry_run_counter = 0
    end

    # It should be executed before build/clean of subprojects.
    # If is_only_one option is set and manual_dry_run_counter is equal to 0,
    # manual dry_run mode is activated.
    def before_subprj_processing
      if is_only_one
        if 0 == @manual_dry_run_counter
          MxxRu::Util::Mode.instance.manual_dry_run( true )
        end
        @manual_dry_run_counter += 1
      end
    end

    # It should be executed after build/clean of subprojects.
    # If is_only_one option is set and manual_dry_run_counter is equal to 0,
    # manual dry_run mode is activated.
    def after_subprj_processing
      if is_only_one
        @manual_dry_run_counter -= 1
        if 0 == @manual_dry_run_counter
          MxxRu::Util::Mode.instance.manual_dry_run( false )
        end
      end
    end

  end

end # module Cpp
end # module MxxRu

# C++ mode option parser must be added.
MxxRu::CmdLineOptionProcessor.instance.add_processor(
    MxxRu::Cpp::Mode::OptionParser.instance )

load File.dirname( __FILE__ ) + '/../compat.rb'
