#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2016, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/ex'

module  MxxRu

  module Cpp

    # Base class for a folder name generator.
    # Results of compilation and linking would be located in that folder.
    class ObjPlacement
      # Method, that returns the name of a folder, object file would be located in.
      #
      # [_source_path_name_] Folder name, where source file located is.
      # [_toolset_] Current toolset.
      # [_target_]  Current target.
      def get_obj(
        source_path_name,
        toolset,
        target )

        raise AbstractMethodEx.new(
          "MxxRu::Cpp::ObjPlacement::get_obj" )
      end

      # Method, that returns the name of a folder, 
      # compiled mswin-resource file would be located in.
      #
      # [_source_path_name_] Folder name, where source file located is.
      # [_toolset_] Current toolset.
      # [_target_]  Current target.
      def get_mswin_res(
        source_path_name,
        toolset,
        target )

        raise AbstractMethodEx.new(
          "MxxRu::Cpp::ObjPlacement::get_mswin_res" )
      end

      # Method, that returns the name of a folder, 
      # static library file would be located in.
      #
      # [_source_path_name_] Folder name, where source file located is.
      # [_toolset_] Current toolset.
      # [_target_]  Current target.
      def get_lib(
        source_path_name,
        toolset,
        target )

        raise AbstractMethodEx.new(
          "MxxRu::Cpp::ObjPlacement::get_lib" )
      end

      # Method, that returns the name of a folder, 
      # shared library file would be located in.
      #
      # [_source_path_name_] Folder name, where source file located is.
      # [_toolset_] Current toolset.
      # [_target_]  Current target.
      def get_dll(
        source_path_name,
        toolset,
        target )

        raise AbstractMethodEx.new(
          "MxxRu::Cpp::ObjPlacement::get_dll" )
      end

      # Method, that returns the name of a folder, 
      # executable file would be located in.
      #
      # [_source_path_name_] Folder name, where source file located is.
      # [_toolset_] Current toolset.
      # [_target_]  Current target.
      def get_exe(
        source_path_name,
        toolset,
        target )

        raise AbstractMethodEx.new(
          "MxxRu::Cpp::ObjPlacement::get_exe" )
      end
    end # class ObjPlacement

    # For compatibility with previous versions.
    Obj_placement = ObjPlacement

    # The generator of folder name for an object file, 
    # using a subfolder with a fixed name in a folder with source file.
    # Folder for lib, dll and exe files are not modified.
    # get_obj and get_mswin_res methods are checking if result folder exists.
    # If it's not, they are trying to create it.
    class SourceSubdirObjPlacement < ObjPlacement
      attr  :subdir

      def initialize( a_subdir = "o" )
        # The name of subfolder, object files should be located in.
        @subdir = a_subdir.clone
      end

      def get_obj(
        source_path_name,
        toolset,
        target )

        result = File.join( source_path_name, subdir )
        MxxRu::Util.ensure_path_exists( result )

        return result
      end

      def get_mswin_res(
        source_path_name,
        toolset,
        target )

        result = File.join( source_path_name, subdir )
        MxxRu::Util.ensure_path_exists( result )

        return result
      end

      def get_lib(
        source_path_name,
        toolset,
        target )

        return String.new( source_path_name )
      end

      def get_dll(
        source_path_name,
        toolset,
        target )

        return String.new( source_path_name )
      end

      def get_exe(
        source_path_name,
        toolset,
        target )

        return String.new( source_path_name )
      end

    end # class SourceSubdirObjPlacement

    # For compatibility with previous versions.
    Source_subdir_obj_placement = SourceSubdirObjPlacement

    # The generator of folder name for compilation results,
    # building a hierarchy of subfolders in a special folder,
    # which name is choosen based on runtime mode.
    #
    # For example, let src/lib/l.cpp and src/main/m.cpp files would be
    # the sources of lib/l.lib library and m.exe application.
    # If project is compiled in RELEASE mode, then following files
    # would be created: release/src/lib/l.obj, release/src/main/m.obj,
    # release/lib/l.lib and release/m.exe.
    # Thus the presence of subfolders required will be supervised
    # (for example, release/src/lib, release/src/main,...).
    # If some subfolder doesn't exist, it will be created.
    #
    # An example of usage:
    #
    #  class  Build < MxxRu::Cpp::Composite_target
    #    def initialize( a_alias = MxxRu::BUILD_ROOT )
    #      global_obj_placement(
    #        MxxRu::Cpp::RuntimeSubdirObjPlacement.new(
    #          "output" ) )
    #
    #      required_prj( "src/lib/prj.rb" )
    #      required_prj( "src/main/prj.rb" )
    #    end
    #  end
    #
    class RuntimeSubdirObjPlacement < ObjPlacement
      # Folder name, where subfolders for exact runtime-modes 
      # will be created.
      attr_reader :root_dir
      # A name of subfolder, which will ve created in a root_dir
      # for MxxRu::Cpp::RUNTIME_DEBUG mode.
      attr_reader :debug_subdir
      # A name of subfolder, which will ve created in a root_dir
      # for MxxRu::Cpp::RUNTIME_DEFAULT mode.
      attr_reader :default_subdir
      # A name of subfolder, which will ve created in a root_dir
      # for MxxRu::Cpp::RUNTIME_RELEASE mode.
      attr_reader :release_subdir

      # a_root_dir  A folder, where subfolders for exact runtime-modes
      # will be created. If contains nil, subfolders are created in
      # current folder.
      #
      # [_a_debug_subdir_] Subfolder name for MxxRu::Cpp::RUNTIME_DEBUG mode.
      # [_a_default_subdir_] Subfolder name for MxxRu::Cpp::RUNTIME_DEFAULT mode.
      # [_a_release_subdir_] Subfolder name for MxxRu::Cpp::RUNTIME_RELEASE mode.
      def initialize(
        a_root_dir = nil,
        a_debug_subdir = "debug",
        a_default_subdir = "default",
        a_release_subdir = "release" )

        if a_root_dir
          @root_dir = a_root_dir
        else
          @root_dir = "./"
        end

        @debug_subdir = a_debug_subdir
        @default_subdir = a_default_subdir
        @release_subdir = a_release_subdir
      end

      # It's the only method running something.
      def get_obj(
        source_path_name,
        toolset,
        target )

        if source_path_name &&
          "" != source_path_name &&
          "." != source_path_name
          result = File.join( @root_dir, runtime_mode_path( target ),
            source_path_name )
        else
          result = File.join( @root_dir, runtime_mode_path( target ) )
        end

        MxxRu::Util.ensure_path_exists( result )

        return result
      end

      # Returns result of get_obj method.
      def get_mswin_res(
        source_path_name,
        toolset,
        target )

        return get_obj( source_path_name, toolset, target )
      end

      # Returns result of get_obj method.
      def get_lib(
        source_path_name,
        toolset,
        target )

        return get_obj( source_path_name, toolset, target )
      end

      # Returns result of get_obj method.
      def get_dll(
        source_path_name,
        toolset,
        target )

        return get_obj( source_path_name, toolset, target )
      end

      # Returns result of get_obj method.
      def get_exe(
        source_path_name,
        toolset,
        target )

        return get_obj( source_path_name, toolset, target )
      end

    protected
      # Returns folder name, which is used for target's runtime mode.
      #
      # [_a_target_] Target, actions are performed for.
      def runtime_mode_path( a_target )
        case a_target.mxx_runtime_mode
          when MxxRu::Cpp::RUNTIME_DEBUG
            return @debug_subdir

          when MxxRu::Cpp::RUNTIME_RELEASE
            return @release_subdir

          else
            return @default_subdir
        end
      end

    end # class RuntimeSubdirObjPlacement

    # For compatibility with older versions.
    Runtime_subdir_obj_placement = RuntimeSubdirObjPlacement

    # The generator of folder name for compilation results,
    # building a hierarchy of subfolders in a special folder,
    # which name is choosen based on toolset and runtime mode.
    #
    # For example, let src/lib/l.cpp and src/main/m.cpp files would be
    # the sources of lib/l.lib library and m.exe application.
    # If project is compiled in RELEASE mode, then following files
    # would be created: toolset-id/release/src/lib/l.obj,
    # toolset-id/release/src/main/m.obj,
    # toolset-id/release/lib/l.lib and toolset-id/release/m.exe.
    # Thus the presence of subfolders required will be supervised
    # (for example, toolset-id/release/src/lib,
    # toolset-id/release/src/main,...).
    # If some subfolder doesn't exist, it will be created.
    #
    # An example of usage:
    #
    #  MxxRu::Cpp::composite_target( MxxRu::BUILD_ROOT )
    #      global_obj_placement(
    #        MxxRu::Cpp::ToolsetRuntimeSubdirObjPlacement.new(
    #          "output" ) )
    #
    #      required_prj( "src/lib/prj.rb" )
    #      required_prj( "src/main/prj.rb" )
    #    end
    #  end
    #
    class ToolsetRuntimeSubdirObjPlacement < RuntimeSubdirObjPlacement
      # a_root_dir  A folder, where subfolders for exact toolset and
      # runtime-modes will be created. If contains nil, subfolders are created
      # in current folder.
      #
      # [_a_debug_subdir_] Subfolder name for MxxRu::Cpp::RUNTIME_DEBUG mode.
      # [_a_default_subdir_] Subfolder name for MxxRu::Cpp::RUNTIME_DEFAULT mode.
      # [_a_release_subdir_] Subfolder name for MxxRu::Cpp::RUNTIME_RELEASE mode.
      def initialize(
        a_root_dir = nil,
        a_debug_subdir = "debug",
        a_default_subdir = "default",
        a_release_subdir = "release" )

        super( a_root_dir, a_debug_subdir, a_debug_subdir, a_release_subdir )
      end

      def get_obj(
        source_path_name,
        toolset,
        target )

        if source_path_name &&
          "" != source_path_name &&
          "." != source_path_name
          result = File.join( @root_dir,
            toolset.make_identification_string,
            runtime_mode_path( target ),
            source_path_name )
        else
          result = File.join( @root_dir,
            toolset.make_identification_string,
            runtime_mode_path( target ) )
        end

        MxxRu::Util.ensure_path_exists( result )

        return result
      end
    end # class ToolsetRuntimeSubdirObjPlacement

    # Creation of obj_placement object, which should be used by default.
    def Cpp.default_obj_placement
      return SourceSubdirObjPlacement.new
    end

  end # module Cpp

end # module MxxRu

