#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/util'
require 'mxx_ru/cpp/obj_placement'

module MxxRu

module Cpp

# Analog of RuntimeSubdirObjPlacement, but allow to specify
# paths for final results (EXE, LIB, DLL) and intermediate files (OBJ, RES).
# Unlike RuntimeSubdirObjPlacement these paths are independent of
# runtime_mode.
#
# Example:
#   MxxRu::Cpp::composite_target {
#     global_obj_placement MxxRu::Cpp::CustomSubdirObjPlacement.new(
#       # Final resuls going here.
#       'bin32',
#       # All intermediate files going here.
#       'tmp/output32' )
#
#     required_prj ...
#   }
# If this composite project will be applied for project structure:
#   prj_1/
#   `- src/
#   prj_2/
#   `- module_1/
#   `- module_2/
# Then after build project structructure will be:
#   prj_1/
#   `- src/
#   prj_2/
#   `- module_1/
#   `- module_2/
#   bin32/
#   tmp/
#   `- output32/
#      `- prj_1/
#      |  `- src/
#      `- prj_2/
#         `- module_1/
#         `- module_2/
#
class CustomSubdirObjPlacement < ObjPlacement
  # Constructor
  #
  # [_final_results_path_] path for storing final results (EXE, LIB/A, DLL/SO).
  # [_intermediate_path_] path for storing intermediate files (OBJ/O, RES).
  def initialize( final_results_path, intermediate_path )
    @final_results_path = final_results_path
    @intermediate_path = intermediate_path
  end

  # Make name for obj file.
  def get_obj(
    source_path_name,
    toolset,
    target )

    if source_path_name &&
      "" != source_path_name &&
      "." != source_path_name
      result = File.join( @intermediate_path, source_path_name )
    else
      result = @intermediate_path
    end

    MxxRu::Util.ensure_path_exists( result )

    return result
  end

  # Returns result of get_obj method.
  def get_mswin_res(
    source_path_name,
    toolset,
    target )

    return get_obj( source_path_name, toolset, target )
  end

  # Returns final_results_path
  def get_lib(
    source_path_name,
    toolset,
    target )

    final_result_path_component( source_path_name )
  end

  # Returns final_results_path
  def get_dll(
    source_path_name,
    toolset,
    target )

    final_result_path_component( source_path_name )
  end

  # Returns final_results_path
  def get_exe(
    source_path_name,
    toolset,
    target )

    final_result_path_component( source_path_name )
  end

protected
  # Make final_results_path if needed and return name of it
  def final_result_path_component( target_root )
    r = @final_results_path
    r = File.join( @final_results_path, target_root ) if
        target_root and target_root.size
    MxxRu::Util::ensure_path_exists( r )
    r
  end

end # class CustomSubdirObjPlacement

# For compatibility with old naming convention.
Custom_subdir_obj_placement = CustomSubdirObjPlacement

end # module Cpp

end # module MxxRu

