#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2016, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/util'
require 'mxx_ru/cpp/obj_placement'
require 'mxx_ru/cpp/target'

require 'digest'

module MxxRu

module Cpp

# Since v.1.6.12
#
# This class is similar to RuntimeSubdirObjPlacement and
# ToolsetRuntimeSubdirObjPlacement but stores intermediate files (like
# object files, compiled resources and so on) in special subfolders.
# 
# These subfoldes use special naming in forms:
# 
#    <final_path>/_objs/<runtime>/<prj_alias>/<subdir>
#    <final_path>/_objs/<toolset>/<runtime>/<prj_alias>/<subdir>
#
# where subdir is created from a path to source file. A path is
# transformed (all separators like / or :) is changed to _. It
# means that so_5/rt/impl will be transformed to so_5_rt_impl.
# If the result is too long it is transformed another time: part
# of result name will be changed to SHA1 digest. For example a name
# so_5/disp/prio_one_thread/quoted_round_robin will be transformed to
# so_5_disp_prio_one_thread_quoted_round_robin and then to
# so_5__robin-8f160349ec35.
#    
# Usage example:
#
#   MxxRu::Cpp::composite_target {
#     global_obj_placement MxxRu::Cpp::PrjAwareRuntimeSubdirObjPlacement.new(
#       # Final resuls going here.
#       'bin32' )
#
#     required_prj 'so_5/prj.rb'
#     required_prj 'so_5/prj_s.rb'
#     ...
#   }
#
# The directory structure could looks like:
#
#   bin32/
#   `- _objs/
#      `- release/
#         `- so_5_prj_rb/
#            `- ...
#         `- so_5_prj_s_rb/
#            `- ...
#   `- release/
#
# This ObjPlacement also allows to use toolset id in directory structure:
#
#   MxxRu::Cpp::composite_target {
#     global_obj_placement MxxRu::Cpp::PrjAwareRuntimeSubdirObjPlacement.new(
#       # Final resuls going here.
#       'bin32',
#       # Enable toolset identification
#       MxxRu::Cpp::PrjAwareRuntimeSubdirObjPlacement::USE_COMPILER_ID )
#
#     required_prj 'so_5/prj.rb'
#     required_prj 'so_5/prj_s.rb'
#     ...
#   }
# 
# The directory structure could looks like:
#
#   bin32/
#   `- _objs/
#      `- gcc_4_8_2__x86_64_w64_mingw32/
#         `- release/
#            `- so_5_prj_rb/
#               `- ...
#            `- so_5_prj_s_rb/
#               `- ...
#   `- gcc_4_8_2__x86_64_w64_mingw32/
#      `- release/
#
class PrjAwareRuntimeSubdirObjPlacement < MxxRu::Cpp::ObjPlacement
  USE_COMPILER_ID = true
  DO_NOT_USE_COMPILER_ID = false

  # Constructor
  def initialize( final_results_path, use_compiler_id = DO_NOT_USE_COMPILER_ID )
    @final_results_path = final_results_path
    @intermediate_path = File.join( @final_results_path, '_objs' )
    @use_compiler_id = use_compiler_id
  end

  # Make name for obj file.
  def get_obj( source_path_name, toolset, target )
    root_path = USE_COMPILER_ID == @use_compiler_id ?
        File.join( @intermediate_path, toolset.make_identification_string ) :
        @intermediate_path

    target_tmps = create_target_tmps_name( target )
    result = if source_path_name &&
        "" != source_path_name &&
        "." != source_path_name
      File.join( root_path, runtime_mode_path( target ), target_tmps,
        transform_path( source_path_name ) )
    else
      File.join( root_path, target_tmps, runtime_mode_path( target ) )
    end

    MxxRu::Util.ensure_path_exists( result )

    result
  end

  # Returns result of get_obj method.
  def get_mswin_res( source_path_name, toolset, target )
    get_obj( source_path_name, toolset, target )
  end

  # Returns final_results_path
  def get_lib( source_path_name, toolset, target )
    final_result_path_component( source_path_name, toolset, target )
  end

  # Returns final_results_path
  def get_dll( source_path_name, toolset, target )
    final_result_path_component( source_path_name, toolset, target )
  end

  # Returns final_results_path
  def get_exe( source_path_name, toolset, target )
    final_result_path_component( source_path_name, toolset, target )
  end

protected
  # Make final_results_path if needed and return name of it
  def final_result_path_component( target_root, toolset, target )
    root_path = USE_COMPILER_ID == @use_compiler_id ?
        File.join( @final_results_path, toolset.make_identification_string ) :
        @final_results_path

    result = if target_root &&
        "" != target_root &&
        "." != target_root
      File.join( root_path, runtime_mode_path( target ), target_root )
    else
      File.join( root_path, runtime_mode_path( target ) )
    end

    MxxRu::Util.ensure_path_exists( result )

    result
  end

  # Returns folder name, which is used for target's runtime mode.
  #
  # [_a_target_] Target, actions are performed for.
  def runtime_mode_path( a_target )
    case a_target.mxx_runtime_mode
      when MxxRu::Cpp::RUNTIME_DEBUG
        return 'debug'

      when MxxRu::Cpp::RUNTIME_RELEASE
        return 'release'

      else
        return 'default'
    end
  end

  def transform_path( path )
    normalized_value = path.gsub( /[\\\/\.:]/, '_' )
    if 24 < normalized_value.size
      r = normalized_value[ 0..3 ] + '__' + normalized_value[ -5..-1 ] + '-' +
        Digest::SHA1.new.update(normalized_value).hexdigest[ 0..11 ]
      normalized_value = r
    end
    normalized_value
  end

  def create_target_tmps_name( target )
    transform_path( target.prj_alias )
  end

end # class PrjAwareRuntimeSubdirObjPlacement

end # module Cpp

end # module MxxRu

# vim:ts=2:sts=2:sw=2:expandtab

