#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'set'

require 'mxx_ru/cpp'

module MxxRu
module Cpp

  # Files generator for Qt class.
  #
  # Main features:
  #
  # - building of moc-files using moc tool.
  #   Generation from source and header files is supported.
  # - building source files from ui-files using uic-compiler.
  #   For all files generated this way moc tool executed automatically.
  #
  # Generated source files automatically added into
  # cpp_source list of target.
  #
  # Local list defines is supported.
  #
  # If only pointer to the target passed into contructor,
  # default list of defines is used. If it's required to change
  # default list, new list should be passed through a second argument:
  #
  #  generator( MxxRu::Cpp::QtGen.new( self, [ "QT_THREAD_SUPPORT" ] ) )
  #
  class QtGen < MxxRu::AbstractGenerator
    # Default list of defines.
    # QT_DLL, QT_THREAD_SUPPORT
    DEFAULT_DEFINES = [ "QT_DLL", "QT_THREAD_SUPPORT" ]

    UiNames = Struct.new( :header, :source )

    # moc tool executable
    # Default: moc
    attr_accessor :moc_name 

    # uic tool executable
    # Default: uic.
    attr_accessor :uic_name

    # File extension for source files generated.
    # Used in both moc and uic tools.
    # Default: .cpp
    attr_accessor :cpp_ext

    # File extension for header files generated.
    # Used in uic tool.
    # Default: .hpp
    attr_accessor :hpp_ext

    # File extension for moc files generated.
    # Used in moc tool.
    # Default: .moc
    attr_accessor :moc_ext

    # Subfolder name where moc generated files would be located.
    # If nil, generated files would be in the same folder
    # where source files were.
    # Default: nil
    attr_accessor :moc_result_subdir

    # Subfolder name where results of uic invokation must be placed.
    # If nil, generated files would be in the same folder where
    # source ui files were.
    attr_accessor :uic_result_subdir

    # Files list for generating moc files from header files.
    #
    # More exact: a.hpp -> moc_a.cpp. moc_a.cpp file is added
    # to cpp_source list of a target.
    attr_reader :qt_h2moc_files

    # Files list for generating moc files from source files.
    #
    # More exact: a.cpp -> a.moc.
    attr_reader :qt_cpp2moc_files

    # Files list for generating hpp, cpp and moc files from ui files.
    #
    # More exact: a.ui -> a.hpp, a.cpp, moc_a.cpp. 
    # a.cpp, moc_a.cpp files are added to cpp_source of a target.
    attr_reader :qt_ui_files

    # Target, generator is created for.
    attr  :target

    # Constructor.
    def initialize( a_target, a_defines = DEFAULT_DEFINES )
      @moc_name = "moc"
      @uic_name = "uic"
      @cpp_ext = ".cpp"
      @hpp_ext = ".hpp"
      @moc_ext = ".moc"
      @moc_result_subdir = nil
      @uic_result_subdir = nil

      # If uic_result_subdir not nil than we must add
      # uic_result_subdir into
      # target include_path. But must do this only one time.
      @uic_result_subdir_include_paths = Set.new

      @qt_h2moc_files = Array.new
      @qt_cpp2moc_files = Array.new
      @qt_ui_files = Array.new

      defines_to_set = a_defines.flatten
      defines_to_set.each { |d| a_target.define( d ) }

      @target = a_target
    end

    # Add a file for hpp to moc generation.
    def h2moc( a_file )
      @qt_h2moc_files << @target.create_full_src_file_name( a_file )
    end

    # Add a file for cpp to moc generation.
    def cpp2moc( a_file )
      @qt_cpp2moc_files << @target.create_full_src_file_name( a_file )
    end

    # Add a file for ui to hpp, cpp and moc generation.
    def ui( a_file )
      full_file_name = @target.create_full_src_file_name( a_file )
      @qt_ui_files << full_file_name
      if @uic_result_subdir
        uic_result_path = make_uic_result_path_for( full_file_name )

        if !@uic_result_subdir_include_paths.member?( uic_result_path )
          @target.include_path( uic_result_path )
          @uic_result_subdir_include_paths.add( uic_result_path )
        end
      end
    end

    # Perform files generation.
    def build( a_target )
      build_from_ui( a_target )
      build_from_h( a_target )
      build_from_cpp( a_target )
    end

    # Perform generated files cleanup.
    def clean( a_target )
      clean_from_ui( a_target )
      clean_from_h( a_target )
      clean_from_cpp( a_target )
    end

  protected

    # Perform generation from ui-files.
    def build_from_ui( a_target )
      @qt_ui_files.each { |ui|
        files_from_ui = names_from_ui( ui, a_target )
        MxxRu::Util::ensure_path_exists(
            File.dirname( files_from_ui.header ) )
        MxxRu::Util::ensure_path_exists(
            File.dirname( files_from_ui.source ) )

        # Full file path detection.
        header_file = files_from_ui.header
        cpp_file = files_from_ui.source

        @qt_h2moc_files << header_file 
        add_cpp_source( a_target, cpp_file )

        # Checking what files are changed and running generation
        # from them if necessary.
        if TargetState::EXISTS != TargetState.detect(
          header_file, [ ui ] ).state
          MxxRu::AbstractTarget::run(
            [ "#{@uic_name} -o #{header_file} #{ui}" ],
            [ header_file ],
            "building header file #{header_file}" )
        end
        if TargetState::EXISTS != TargetState.detect(
          cpp_file, [ ui, header_file ] ).state
          MxxRu::AbstractTarget::run(
            [ "#{@uic_name} -i #{File.basename(header_file)} " +
              "-o #{cpp_file} #{ui}" ],
            [ cpp_file ],
            "building source file #{cpp_file}" )
        end
      }
    end

    # Perform cleanup of files, generated from ui files.
    def clean_from_ui( a_target )
      @qt_ui_files.each { |ui|
        files_from_ui = names_from_ui( ui, a_target )

        MxxRu::Util::delete_file( files_from_ui.header )
        MxxRu::Util::delete_file( files_from_ui.source )

        # To remove object files.
        @qt_h2moc_files << files_from_ui.header
        add_cpp_source( a_target, files_from_ui.source )
      }
    end

    # Getting file names, generated by uic tool.
    #
    # Returns UiNames struct object.
    def names_from_ui( a_ui_file, a_target )
      if @uic_result_subdir then
        uic_result_path = make_uic_result_path_for( a_ui_file )

        short_ui_name = MxxRu::Util::remove_file_ext(
            File.basename( a_ui_file ) )
        result = UiNames.new(
            File.join( uic_result_path, short_ui_name + @hpp_ext ),
            File.join( uic_result_path, short_ui_name + @cpp_ext ) )
      else
        full_ui_name = MxxRu::Util::remove_file_ext( a_ui_file )
        result = UiNames.new(
            full_ui_name + @hpp_ext,
            full_ui_name + @cpp_ext )
      end

      return result
    end

    # Perform generation from hpp file using moc tool.
    def build_from_h( a_target )
      @qt_h2moc_files.each { |header_full|
        moc_full = moc_from_h( header_full, a_target )

        if TargetState::EXISTS != TargetState.detect(
          moc_full, [ header_full ] ).state
          MxxRu::AbstractTarget::run(
            [ "#{moc_name} -o #{moc_full} " + 
              "#{header_full}" ],
            [ moc_full ],
            "building moc file #{moc_full}" )
        end

        add_cpp_source( a_target, moc_full )
      }
    end

    # Perform cleanup of files produced by generation from hpp file.
    def clean_from_h( a_target )
      @qt_h2moc_files.each { |h|
        moc = moc_from_h( h, a_target )

        MxxRu::Util::delete_file( moc )

        add_cpp_source( a_target, moc )
      }
    end

    # Formatting file name, which is generated from hpp file by moc tool
    def moc_from_h( a_h_file, a_target )
      path = File.dirname( a_h_file )

      # If path == ".", then Qt will generate wrong include directive
      path = "" if path == "./"
      path = make_moc_result_path_for( path )

      MxxRu::Util::ensure_path_exists( path )

      r = File.join( path, "moc_" + File.basename(
        MxxRu::Util::remove_file_ext( a_h_file ) ) + @cpp_ext )
    end

    # Perform generation from cpp file using moc tool.
    def build_from_cpp( a_target )
      @qt_cpp2moc_files.each { |cpp_full|
        moc_full = moc_from_cpp( cpp_full, a_target )

        if TargetState::EXISTS != TargetState.detect(
          moc_full, [ cpp_full ] ).state
          MxxRu::AbstractTarget::run(
            [ "#{moc_name} -o #{moc_full} " + 
              "#{cpp_full}" ],
            [ moc_full ],
            "building moc file #{moc_full}" )
        end
      }
    end

    # Perform cleanup of files produced by generation from cpp file.
    def clean_from_cpp( a_target )
      @qt_cpp2moc_files.each { |cpp|
        moc = moc_from_cpp( cpp, a_target )

        MxxRu::Util::delete_file( moc )
      }
    end

    # Formatting file name, which is generated from cpp file by moc tool
    def moc_from_cpp( a_cpp_file, a_target )
      path = make_moc_result_path_for( File.dirname( a_cpp_file ) )

      MxxRu::Util::ensure_path_exists( path )

      r = File.join( path, File.basename(
        MxxRu::Util::remove_file_ext( a_cpp_file ) ) + @moc_ext )

      return r
    end

    # Setting C++ file name, ignoring current sources_root value.
    def add_cpp_source( a_target, a_file )
      old_root = a_target.sources_root( "" )
      a_target.cpp_source( a_file )
      a_target.sources_root( old_root )
    end

    # Calculating path for moc results.
    def make_moc_result_path_for( path )
      if @moc_result_subdir
        if path.length > 0
          File.join( path, @moc_result_subdir )
        else
          @moc_result_subdir
        end
      else
        path
      end
    end

    # Calculating path for uic results.
    def make_uic_result_path_for( file )
      File.join( File.dirname( file ), @uic_result_subdir )
    end

  end # class QtGen

  # For compatibility with previous versions
  Qt_gen = QtGen

end
end

