#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/abstract_target.rb'

module MxxRu
module Cpp

# Generator intended to use generators based on 
# {RuCodeGen}[http://eao197.narod.ru/rucodegen] in C++ projects.
#
# The attachment of the given generator to the C++ projects consists
# in that it allows sources_root value be taken into account during
# a work with names of code generation scripts.
#
# For example, for my_prj/pub.rb and my_prj/impl/details.rb code 
# generation scripts:
#
#  require 'mxx_ru/cpp'
#  require 'mxx_ru/cpp/rucodegen'
# 
#  MxxRu::setup_target( MxxRu::Cpp::ExeTarget.new( "my_prj/prj.rb" ) {
#    ...
#    rucodegen = generator( MxxRu::Cpp::RuCodeGen.new( self ) )
#    ...
#    rucodegen << "pub.rb"
#    ...
#    sources_root( "impl" ) {
#      rucodegen << "details.rb"
#      ...
#    }
#    ...
#  })
#
# <b>NOTE</b>. RuCodeGen#add method may be used instead of shift operator
# for adding next code generation script
#
#  rucodegen.add "details.rb"
#
# Since {RuCodeGen}[http://eao197.narod.ru/rucodegen] 0.2.0 embeded scripts
# supported by RuCodeGen#add_embedded method:
#
#  rucodegen.add_embedded 'cfg.hpp'
#
class RuCodeGen < AbstractGenerator
  # Base class for various code generation script types.
  class Script
    def initialize( name )
      @name = name
    end

    def to_s
      @name
    end
  end

  # Description for standalone RuCodeGen script.
  class Standalone < Script
    def initialize( name )
      super( name )
    end

    # Returns command line for run code generation.
    def build_cmd( mode )
      "ruby -I. #{@name} --mode #{mode}"
    end
  end

  # Description for embeded RuCodeGen script.
  class Embedded < Script
    def initialize( name )
      super( name )
    end

    # Returns command line for run code generation.
    def build_cmd( mode )
      "ruby -I. -e\"require 'rubygems'; gem 'RuCodeGen'; require 'rucodegen'; " +
        "process_embedded_rucodegen('#{@name}',:#{mode})\""
    end
  end

  # Target, code generation is performed for.
  attr_reader :target

  # A list of code generation script names.
  attr_reader :scripts


  # Constructor.
  #
  # [_target_] Target, code generation is performed for.
  def initialize( target )
    @target = target
    @scripts = []
  end

  # Adding a name of next code generation script.
  def <<( script )
    @scripts << Standalone.new( @target.create_full_src_file_name( script ) )
  end

  alias_method :add, :<<

  # Adding a name of next embeded code generation script.
  def add_embedded( script )
    @scripts << Embedded.new( @target.create_full_src_file_name( script ) )
  end

  alias_method :add_embeded, :add_embedded

  # Perform code generation.
  def build( target )
    @scripts.each do |s|
      AbstractTarget.run(
          [ s.build_cmd( :build ) ],
          [],
          "running code generation script #{s} (build mode)" )
    end
  end

  # Perform cleanup of generated files.
  def clean( target )
    @scripts.each do |s|
      AbstractTarget.run(
          [ s.build_cmd( :clean ) ],
          [],
          "running code generation script #{s} (clean mode)" )
    end
  end
end

end # module Cpp
end # module MxxRu

