#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/binary_target'
require 'mxx_ru/cpp/mode'
require 'mxx_ru/cpp/source_file'
require 'mxx_ru/cpp/obj_placement'
require 'mxx_ru/cpp/analyzer'

# Base class for all targets of C/C++ projects.
module MxxRu

  module Cpp
    # Runtime mode types.

    # Default mode.
    # Nor debug information nor optimization options are set.
    RUNTIME_DEFAULT = "default"
    # Release mode. Optimization is turned on.
    RUNTIME_RELEASE = "release"
    # Debug mode. Debug information is generated.
    RUNTIME_DEBUG = "debug"

    # RTTI mode types.

    # Default RTTI mode. No RTTI options are passed to compiler.
    RTTI_DEFAULT = "default"
    # RTTI mode should be enabled.
    RTTI_ENABLED = "enabled"
    # RTTI mode should be disabled.
    RTTI_DISABLED = "disabled"

    # RTL mode types.

    # Default compiler libraries are used.
    RTL_DEFAULT = "default"
    # Use shared runtime library.
    RTL_SHARED = "shared"
    # Use static runtime library.
    RTL_STATIC = "static"

    # Threading mode types.

    # Default compiler threading mode.
    THREADING_DEFAULT = "default"
    # Multithreading is used.
    THREADING_MULTI = "multi"
    # Multithreading is not used.
    THREADING_SINGLE = "single"

    # Screen mode types.

    # Console application.
    SCREEN_CONSOLE = "console"
    # Windowed application.
    SCREEN_WINDOW = "window"

    # Optimization modes.

    # Optimization by size.
    OPTIM_SIZE = "size"
    # Optimization by speed.
    OPTIM_SPEED = "speed"

    # Auxiliary class, intended to store objects like
    # runtime_mode, rtl_mode, etc, with information about
    # who has set this or that mode.
    class GlobalSingleValueOption
      attr_reader :name
      attr_reader :default_value
      attr_reader :value
      attr_reader :who

      # [_a_name_] Name.
      # [_a_default_value_] Default value.
      # [_a_available_values_] Available values. Array of String.
      def initialize(
        a_name,
        a_default_value,
        a_available_values )

        @name = a_name
        @default_value = a_default_value
        @default_value_changer = nil
        @available_values = a_available_values
        @value = @default_value
        @who = "n/a"
      end

      # Change default value.
      #
      # Description of this change is printed to standard output.
      #
      # [_a_value_] new default value.
      # [_a_who_] who tries to change value.
      def change_default( a_value, a_who )
        check_value( a_value, a_who )

        if nil == @default_value_changer
          need_change_current = ( @value == @default_value )
          @default_value = a_value
          @default_value_changer = a_who
          @value = @default_value if need_change_current
        elsif a_value != @default_value
          puts "WARNING: ignoring change default value for '#{@name}' to " +
            "'#{a_value}' by '#{a_who}'; default value already changed by " +
            "'#{@default_value_changer}'"
        end
      end

      # Value change.
      # If new value is the same as default, value change try is ignored.
      #
      # [_a_value_] New value.
      # [_a_who_] Who tries to change value.
      def change(
        a_value,
        a_who )

        if a_value != @default_value
          check_value( a_value, a_who )

          if @value == @default_value
            # Value did not change till now.
            @value = a_value
            @who = a_who
          else
            # Value is already changed. It is necessary to be convinced,
            # that values are the same. Otherwise there is a conflict.
            if a_value != @value
              raise MxxRu::GlobalOptionConflictEx.new(
                @name, @value, @who, a_value, a_who )
            end
          end
        end
      end

    protected
      # Search value in available values. Raise exceptions if value
      # not found.
      def check_value( value, who )
        if !@available_values.include?( value )
          raise MxxRu::GlobalOptionDisallowedValueEx.new(
            @name, @available_values, value, who )
        end
      end
    end # class GlobalSingleValueOption

    # For compatibility with previous versions.
    Global_single_value_option = GlobalSingleValueOption

    # Auxiliary class for a storage of options, which can be broken on 3 groups:
    # global, local and upspread.
    # Examples of such options are: defines, include_paths,
    # compiler_options, linker_options, ...
    class SpreadableOption
      # A name of a method, which should be used for reception of all options.
      ALL_OPT_METHOD = "all"
      # A name of a method, which should be used for reception of all upspread
      # options.
      ALL_UPSPREAD_METHOD = "all_upspreads"

      # The global tick counter which is used for check that cached values are
      # still correct.  Is increased at each reference to add.
      @@last_change_time = 1
      # Global hash, in which the time of last change of options with a
      # concrete name is fixed.
      @@cache_times = Hash.new

      # Set new time of change of value of an option.  This method is intended
      # for use at change of global options, which are stored separately from
      # spreadable-options, but which values are included in values of
      # spreadable-options.
      def SpreadableOption.refresh_option_change_time( a_option_name )
        @@last_change_time = @@last_change_time + 1
        @@cache_times[ a_option_name ] = @@last_change_time
      end 

      # Constructor.
      #
      # [_a_option_name_] The name of an option, which serves the given object.
      #                   This name will be used for fixation of time of caching
      #                   in @@cache_times object.
      def initialize( a_option_name )

        @@cache_times[ a_option_name ] = 1

        # Own name.
        @name = a_option_name

        # Own options.
        @locals = Array.new
        # Spreadable options.
        @upspreads = Array.new

        # Cached value of all method.
        @all_cache_value = Array.new
        # @@last_change_time value at caching moment in all method.
        @all_cache_time = 0

        # Cached value of all_upspread method.
        @all_upspread_cache_value = Array.new
        # @@last_change_time value at caching moment in all_upspread method.
        @all_upspread_cache_time = 0

      end

      # Add option.
      #
      # [_a_option_] Added value.
      # [_a_mode_] Local or upspread option.
      #            Is set by values MxxRu::Cpp::Target::OPT_LOCAL,
      #            MxxRu::Cpp::Target::OPT_UPSPEAD.
      def add(
        a_option,
        a_mode = MxxRu::Cpp::Target::OPT_LOCAL )
        add_unique_to( a_option, check_opt_mode( a_mode ) )
      end

      # To get only upspread options..
      def upspeads_only
        return @upspreads
      end

      # To generate the list of all options.
      #
      # [_a_globals_] The storage of global options, from which is necessary
      #               to take values.
      # [_a_subprojects_] The list of all subordinated projects.
      #                   Array of MxxRu::AbstractTarget.
      # [_a_upspreads_method_] The name of a method, which needs to be called
      #                        from the subordinated project, to get it's list
      #                        of all upspread options.
      def all(
        a_globals,
        a_subprojects,
        a_upspreads_method )

        if @all_cache_time < @@cache_times[ @name ]
          # It is necessary to calculate cache anew.
          r = Array.new
          r << a_globals << @locals

          # At the same time it is possible to update cache of upspread-options.
          try_update_upspread_options( a_subprojects, a_upspreads_method )

          r << @all_upspread_cache_value
            
          @all_cache_value = r.flatten.uniq
          @all_cache_time = @@last_change_time
        end

        return @all_cache_value
      end

      # To generate the list of all upspread options.
      #
      # [_a_subprojects_] The list of all subordinated projects.
      #                   Array of MxxRu::AbstractTarget.
      # [_a_upspreads_method_] The name of a method, which needs to be called
      #                        from the subordinated project, to get it's list
      #                        of all upspread options.
      def all_upspreads(
        a_subprojects,
        a_upspreads_method )

        try_update_upspread_options( a_subprojects, a_upspreads_method )

        return @all_upspread_cache_value
      end

    protected
      # Add value to Array, if that value is still is not present there.
      def add_unique_to( what, to )
        if !to.include?( what )
          to.push( what )

          SpreadableOption::refresh_option_change_time( @name )
        end
      end

      # Check up value of a mode parameter on equality 
      # either OPT_UPSPREAD, or OPT_LOCAL and return object of Array type,
      # in which it is necessary to place value.
      #
      # [_a_mode_] mode parameter value for check.
      def check_opt_mode(
        a_mode )
        if MxxRu::Cpp::Target::OPT_UPSPREAD == a_mode
          return @upspreads
        elsif MxxRu::Cpp::Target::OPT_LOCAL == a_mode
          return @locals
        else
          raise MxxRu::InvalidValueEx.new(
            "invalid option mode #{a_mode}" )
        end
      end
      
      # Checks upspread options cache time and updates upspread
      # options if needed.
      def try_update_upspread_options( subprojects, upspreads_method )
        if @all_upspread_cache_time < @@cache_times[ @name ]
          # It is necessary to calculate cache anew.
          r = Array.new
          r << @upspreads
   
          subprojects.each { |p|
            if p.kind_of?( MxxRu::Cpp::Target )
              r << p.send( upspreads_method )
            end
          }
   
          @all_upspread_cache_value = r.flatten.uniq
          @all_upspread_cache_time = @@last_change_time
        end
      end

    end # class SpreadableOption

    # For compatibility with previous versions.
    Spreadable_option = SpreadableOption

    # Target type (exe, lib, dll, unittest, composite etc.).
    # The real type of the target is defined by a derivative class.
    class TargetType
      def name
        raise AbstractMethodEx.new( "MxxRu::Cpp::TargetType::name" )
      end
    end

    # For compatibility with previous versions.
    Target_type = TargetType

    # Target type is application.
    class ExeTargetType < TargetType
      TYPE = "exe"

      def name
        return TYPE
      end
    end

    # For compatibility with previous versions.
    Exe_target_type = ExeTargetType

    # Target type is static library (lib).
    class LibTargetType < TargetType
      TYPE = "lib"

      def name
        return TYPE
      end
    end

    # For compatibility with previous versions.
    Lib_target_type = LibTargetType

    # Target type is shared library (dll).
    class DllTargetType < TargetType
      TYPE = "dll"

      def name
        return TYPE
      end
    end

    # For compatibility with previous versions.
    Dll_target_type = DllTargetType

    # Target type is MacOS bundle (.bundle is a kind of dll on MacOS).
    #
    # Since v.1.4.10
    #
    class MacOSBundleTargetType < TargetType
      TYPE = "macos_bundle"

      def name
        return TYPE
      end
    end

    # For compatibility with previous versions.
    Macos_bundle_target_type = MacOSBundleTargetType

    # Base class for all targets of C/C++ projects.
    class Target < MxxRu::BinaryTarget
      # Description of obj_placement object installed.
      # Used to store global obj_placement.
      class GlobalObjPlacementInfo
        # Global obj_placement.
        attr_reader :placement
        # Who was set obj_placement.
        attr_reader :who

        def initialize
          @placement = nil
          @who = nil
        end

        # Change option value.
        # If placement already set, exception thrown.
        def change( a_placement, a_who )
          if nil != a_placement && nil != @placement
            raise MxxRu::GlobalOptionConflictEx.new(
              "global_obj_placement",
              @placement.class.name,
              @who,
              a_placement.class.name,
              a_who )
          end

          @placement = a_placement
          @who = a_who
        end
      end # class GlobalObjPlacementInfo

      # For compatibility with previous versions.
      Global_obj_placement_info = GlobalObjPlacementInfo

      # Runtime mode storage.
      @@mxx_runtime_mode = GlobalSingleValueOption.new(
        "runtime_mode", RUNTIME_DEFAULT,
        [ RUNTIME_RELEASE, RUNTIME_DEBUG ] )
      # RTTI mode storage.
      @@mxx_rtti_mode = GlobalSingleValueOption.new(
        "rtti_mode", RTTI_DEFAULT,
        [ RTTI_ENABLED, RTTI_DISABLED ] )
      # RTL mode storage.
      @@mxx_rtl_mode = GlobalSingleValueOption.new(
        "rtl_mode", RTL_DEFAULT,
        [ RTL_SHARED, RTL_STATIC ] )
      # Threading mode storage.
      @@mxx_threading_mode = GlobalSingleValueOption.new(
        "threading_mode", THREADING_DEFAULT,
        [ THREADING_MULTI, THREADING_SINGLE ] )

      # The storage of global description of obj_placement.
      @@mxx_global_obj_placement = GlobalObjPlacementInfo.new

      # Attribute that the option should be upspreadable.
      OPT_UPSPREAD = "upspread"
      # Attribute that the option should be local.
      OPT_LOCAL = "local"

      # Path, the name of result of a target is related to.
      # If contains an empty value, target should be defined using a full path.
      # Otherwise the value of mxx_target_root is concatenated with a given
      # target name.
      attr_reader :mxx_target_root
      # Custom prefix for target.
      # NIL, if custom prefix not defined.
      #
      # Since v.1.4.2
      attr_reader :mxx_target_prefix
      # Custom extension for target.
      # NIL, if custom extension not defined.
      attr_reader :mxx_target_ext
      # Path, the names of source files are related to.
      attr_reader :mxx_sources_root
      # Screen mode. Console by default.
      attr_reader :mxx_screen_mode
      # The list of source C-files.
      # Array of MxxRu::Cpp::CppSourceFile objects.
      attr_reader :mxx_c_files
      # The list of source C++-files.
      # Array of MxxRu::Cpp::CppSourceFile objects.
      attr_reader :mxx_cpp_files
      # Object files list.
      attr_reader :mxx_obj_files
      # Name of mswin resource file.
      # Should be object of MxxRu::Cpp::MswinRcFile type.
      # May contain nil value, if resources are not used.
      attr_reader :mxx_mswin_rc_file
      # The name of compiled mswin resource file.
      # Should be a string.
      # May contain nil value, if resources are not used.
      attr_reader :mxx_mswin_res_file
      # Import library folder.
      # May contain nil value.
      attr_reader :mxx_implib_path
      # The name to get target result.
      # Should not contain file extension.
      attr_reader :mxx_target_name
      # Optimization mode. By default, speed optimization is used.
      attr_reader :mxx_optimization
      # Default encoding source files.
      attr_reader :mxx_encoding

      # Generator for methods to manipulation with 'runtime_mode',
      # 'rtti_mode', 'rtl_mode', 'threading_mode' and so on.
      #
      # By default generates 3 methods:
      #   # Setter
      #   def <option>( value ); ...; end
      #   # Setter for default value.
      #   def default_<option>( value ); ...; end
      #   # Getter
      #   def mxx_<option>; ...; end
      #
      # If no_getter == :nogetter then getter method not generated.
      def self.global_option_methods( option, getter = :withgetter )
        class_eval %Q{
          def #{option}( value )
            @@mxx_#{option}.change( value, prj_alias )
          end
          def default_#{option}( value )
            @@mxx_#{option}.change_default( value, prj_alias )
          end
        }
        class_eval %Q{
          def mxx_#{option}
            @@mxx_#{option}.value
          end
        } unless getter == :nogetter
      end

      # Generator for methods to manipulation of options those can be
      # 'global', 'local' and 'upspread'.
      #
      # For each option generates following stuff:
      #   # Global values container.
      #   @@mxx_global_<option> = []
      #   # Local and opspread values container.
      #   @mxx_<option> = SpreadableOption.new(...)
      #   # Setter for global values.
      #   def global_<option>( value ); ...; end
      #   # Setter for local values.
      #   def <option>( value, mode ); ...; end
      #   # Getter for all values (including global, local and upspread).
      #   def mxx_all_<option>s; ...; end
      #   # Getter for upspread values only.
      #   def mxx_all_upspread_<option>s; ...; end
      #
      def self.define_spreadable_option_methods( option )

        global_container_name = "@@mxx_globals_#{option}"
        local_container_name = "@mxx_#{option}"
        global_setter_name = "global_#{option}"
        local_setter_name = "#{option}"
        all_getter_name = "mxx_all_#{option}s"
        upspread_getter_name = "mxx_all_upspread_#{option}s"

        metaprogram = %Q{
          #{global_container_name} = []

          def #{global_setter_name}( value )
            #{global_container_name} << value
            SpreadableOption::refresh_option_change_time( :'#{option}' )
          end

          def #{local_setter_name}(
              value,
              mode = Target::OPT_LOCAL )
            #{local_container_name}.add( value, mode )
          end

          def #{all_getter_name}()
            return #{local_container_name}.all(
                #{global_container_name},
                mxx_required_prjs,
                :'#{upspread_getter_name}' )
          end

          def #{upspread_getter_name}()
            return #{local_container_name}.all_upspreads(
                mxx_required_prjs,
                :'#{upspread_getter_name}' )
          end
        }

        class_eval metaprogram
      end

      # Initializator for all instance variables created by
      # self.define_spreadable_option_methods().
      #
      # Must be called in constructor.
      #
      def self.initialize_spreadable_option_instance_variables( target )
        class_variables.grep( /^@@mxx_globals_/ ) do |g|
          option = /^@@mxx_globals_(.+)$/.match(g)[ 1 ]
          name = "@mxx_#{option}"
          target.instance_variable_set( name, SpreadableOption.new( option ) )
        end
      end

      define_spreadable_option_methods :compiler_option
      define_spreadable_option_methods :cpp_compiler_option
      define_spreadable_option_methods :c_compiler_option
      define_spreadable_option_methods :linker_option
      define_spreadable_option_methods :librarian_option
      define_spreadable_option_methods :define
      define_spreadable_option_methods :include_path
      define_spreadable_option_methods :mswin_rc_option
      define_spreadable_option_methods :mswin_rc_include_path
      define_spreadable_option_methods :mswin_rc_define
      define_spreadable_option_methods :mswin_rlink_option

      # Constructor
      # [_a_prj_alias_] Project identifier.
      def initialize( a_prj_alias )

        super( a_prj_alias )

        @mxx_target_root = ""

        @mxx_target_exe = nil

        @mxx_sources_root = File.dirname( a_prj_alias )

        @mxx_screen_mode = SCREEN_CONSOLE

        @mxx_c_files = Array.new
        @mxx_cpp_files = Array.new

        @mxx_mswin_rc_file = nil
        @mxx_mswin_res_file = nil

        @mxx_obj_files = Array.new

        # The storage of local options.
        # The key is a String, the value is SpreadableOption.
        @mxx_locals = Hash.new

        @mxx_implib_path = nil

        # Object, which defines accommodation of object files.
        @mxx_obj_placement = nil

        @mxx_target_name = nil

        @mxx_optimization = OPTIM_SPEED

        # Result of work of the previous reference to a build method.
        # During a repeted calls of build method target build is not performed,
        # and this value is returned.
        @mxx_last_build_result = nil

        # True if toolset has completely defined all compiler and linker options required.
        @mxx_all_options_defined = false

        @mxx_encoding = Encoding.find("Binary")

        Target::initialize_spreadable_option_instance_variables( self )
      end

      # A simple way to get access to the toolset.
      # In a class, inherited from MxxRu::Cpp::Target is more convinient
      # to call toolset method, instead of MxxRu::Cpp::toolset function.
      # Just calling MxxRu::Cpp::toolset function.
      def toolset()
        return MxxRu::Cpp::toolset()
      end

      global_option_methods :runtime_mode, :nogetter

      # runtime_mode accessor (custom implementation).
      def mxx_runtime_mode
        # If runtime_mode is set to RUNTIME_DEFAULT (i.e.
        # wasn't changed in a projects), then check of command line options
        # is required (--mxx-cpp-release, --mxx-cpp-debug).
        result = @@mxx_runtime_mode.value
        if @@mxx_runtime_mode.default_value == result
          if MxxRu::Cpp::Mode.instance.is_release &&
            MxxRu::Cpp::Mode.instance.is_debug
            # This is an invalid combination of flags.
            raise MxxRu::InvalidValueEx.new(
              "#{MxxRu::Cpp::MXXCPPARG_RELEASE} and " +
              "#{MxxRu::Cpp::MXXCPPARG_DEBUG} cannot be used " +
              "together" )
          elsif MxxRu::Cpp::Mode.instance.is_release
            result = RUNTIME_RELEASE
          elsif MxxRu::Cpp::Mode.instance.is_debug
            result = RUNTIME_DEBUG
          end
        end

        return result
      end

      global_option_methods :rtti_mode
      global_option_methods :rtl_mode
      global_option_methods :threading_mode

      # Set global obj_placement.
      # If gloval obj_placement was already set before, exception is thrown.
      def global_obj_placement( a_placement )
        @@mxx_global_obj_placement.change( a_placement, prj_alias )
      end

      # Set local obj_placement.
      # It's impossible to simultaneously set global and local obj_placement.
      def obj_placement( a_placement )
        @mxx_obj_placement = a_placement
      end

      # Get obj_placement object, which should be used.
      #
      # If both global and local obj_placement are set, exception is thrown.
      #
      # If global or local obj_placement is defined, it is returned.
      #
      # If nor global nor local obj_placement is set, local obj_placement is
      # created with a help of MxxRu::Cpp::default_obj_placement function and
      # it's returned.
      def mxx_obj_placement
        if nil != @mxx_obj_placement &&
          nil != @@mxx_global_obj_placement.placement
          raise MxxRu::InvalidValueEx.new(
            "Global and local obj_placement object exists! " +
            "Global created by: #{@@mxx_global_obj_placement.who}. " +
            "Local created by: #{prj_alias}" )
        end

        if nil != @@mxx_global_obj_placement.placement
          return @@mxx_global_obj_placement.placement
        end

        if nil == @mxx_obj_placement
          @mxx_obj_placement = MxxRu::Cpp::default_obj_placement
        end

        return @mxx_obj_placement
      end

      # Change path, the names of target results are related to.
      #
      # NOTE: This method should be called BEFORE target method!
      #
      # If target_root is called after target method,
      # then _a_root_ given is added to the name already was set.
      def target_root( a_root )
        @mxx_target_root = a_root
        if @mxx_target_name
          @mxx_target_name = File.join(
            [ @mxx_target_root, @mxx_target_name ] )
        end
      end

      # Set a name from which target name would be created.
      def target( a_target )
        @mxx_target_name = create_full_target_file_name( a_target )
      end

      # Set a custom prefix for target.
      def target_prefix( ext )
        @mxx_target_prefix = ext
      end

      # Set a custom extension for target.
      def target_ext( ext )
        @mxx_target_ext = ext
      end
      alias_method :target_suffix, :target_ext

      # Change source files path.
      # If empty string is set, then full file names are required.
      # If block is defined, then path is changed only during a block
      # execution, then old path value is restored.
      #
      # Returns previous mxx_sources_root value. Is actual for a case,
      # when method is called without block parameter.
      def sources_root( a_root )
        old_root = @mxx_sources_root.clone
        if block_given?
          @mxx_sources_root = File.join( [ @mxx_sources_root, a_root ] )
          yield
          @mxx_sources_root = old_root
        else
          @mxx_sources_root = a_root
        end

        return old_root
      end

      # Set screen mode.
      def screen_mode( a_mode )
        if SCREEN_CONSOLE == a_mode || SCREEN_WINDOW == a_mode
          @mxx_screen_mode = a_mode
        else
          raise MxxRu::UnsupportedModeEx.new(
            "Unsupported screen mode: #{a_mode}" )
        end
      end

      # Add one C-file to the target.
      #
      # Note: if a_file_name contains prefix equal to current sources_root
      # than mxx_ru does not add sources_root to name of file. For example:
      #   MxxRu::Cpp::exe_target( 'some/module/prj.rb' ) {
      #     ...
      #     c_source 'f1.c' # become 'some/module/f1.c'
      #     c_source 'some/module/impl/f2.c' # remains 'some/module/impl/f2.c
      # This approach will be useful with Dir.glob:
      #   MxxRu::Cpp::exe_target( 'some/module/prj.rb' ) {
      #     ...
      #     c_sources Dir.glob( 'some/module/**/*.c' )
      #   }
      #
      # [_a_file_name_] The name of a file added.
      # [_a_options_] Optional list of a compiler options for that file.
      #               When command line would be created, this options would
      #               be concatenated with others.
      def c_source( a_file_name, a_options = nil )
        @mxx_c_files.push( CppSourceFile.new(
          create_full_src_file_name( a_file_name ), a_options ) )
      end

      # Add one C++-file to the target.
      #
      # Note: if a_file_name contains prefix equal to current sources_root
      # than mxx_ru does not add sources_root to name of file. For example:
      #   MxxRu::Cpp::exe_target( 'some/module/prj.rb' ) {
      #     ...
      #     cpp_source 'f1.cpp' # become 'some/module/f1.cpp'
      #     cpp_source 'some/module/impl/f2.cpp' # remains 'some/module/impl/f2.cpp
      # This approach will be useful with Dir.glob:
      #   MxxRu::Cpp::exe_target( 'some/module/prj.rb' ) {
      #     ...
      #     cpp_sources Dir.glob( 'some/module/**/*.cpp' )
      #   }
      #
      # [_a_file_name_] The name of a file added.
      # [_a_options_] Optional list of a compiler options for that file.
      #               When command line would be created, this options would
      #               be concatenated with others.
      def cpp_source( a_file_name, a_options = nil )
        @mxx_cpp_files.push( CppSourceFile.new(
          create_full_src_file_name( a_file_name ), a_options ) )
      end

      # Plural form for c_source and cpp_source. Does not allow specify options
      # for source files.
      define_plural_form_method :c_source
      define_plural_form_method :cpp_source

      # Set mswin resource file name.
      #
      # [_a_rc_file_] The name of rc-file.
      # [_a_depends_] Optional dependencies list. Should be Array of String.
      def mswin_rc_file(
        a_rc_file,
        a_depends =nil )

        @mxx_mswin_rc_file = MswinRcFile.new(
            create_full_src_file_name( a_rc_file ),
            a_depends )
      end

      # Set compiled mswin resource file name.
      def mswin_res_file( a_res_file )
        @mxx_mswin_res_file = String.new( a_res_file )
      end

      # Add object file to the target.
      def obj_file( file_name )
        @mxx_obj_files.push( String.new( file_name ) )
      end

      # Set import library location.
      def implib_path( name )
        @mxx_implib_path = name.clone
      end

      # Set optimization mode required.
      def optimization( a_mode )
        if OPTIM_SIZE != a_mode && OPTIM_SPEED != a_mode
          raise MxxRu::InvalidValueEx.new(
            "Unknown optimization mode: '#{a_mode}'" )
        else
          @mxx_optimization = a_mode.clone
        end
      end

      # Set encoding source files.
      def source_encoding( a_encoding )
        @mxx_encoding = Encoding.find("#{a_encoding}")
      end

      # Perform build of the target.
      def build
        if nil == @mxx_last_build_result

          builder = self.method( :normal_mode_builder )
          builder = self.method( :option_extraction_mode_builder ) \
              if MxxRu::Cpp::Mode.instance.is_option_extraction
            
          @mxx_last_build_result = builder.call
        end

        return @mxx_last_build_result
      end

      # Perform target's cleanup.
      def clean
        if nil == @mxx_last_build_result

          cleaner = self.method( :normal_mode_cleaner )
          cleaner = self.method( :option_extraction_mode_cleaner ) \
              if MxxRu::Cpp::Mode.instance.is_option_extraction

          cleaner.call

          @mxx_last_build_result = MxxRu::TargetState.new(
            MxxRu::TargetState::ABSENT )
        end
      end

      # Reset build status to enable rebuild.
      def reset
        if @mxx_last_build_result 
          reset_required_projects
          @mxx_last_build_result = nil
        end
      end

      # Get the target type.
      # Should be overloaded in child classes.
      # Object, inherited from MxxRu::Cpp::TargetType is returned.
      def target_type
        raise AbstractMethodEx.new( "MxxRu::Cpp::Target::target_type" )
      end

      # Creation of a complete name of a resulting file
      # taking the value of mxx_target_root into account
      def create_full_target_file_name( a_file_name )
        if "" == mxx_target_root
          return a_file_name
        else
          return File.join( [ mxx_target_root, a_file_name ] )
        end
      end

      # Creation of a complete name of a source file
      # taking the value of mxx_sources_root into account
      #
      # If a_file_name starts with current value of mxx_sources_root
      # than a_file_name remains unchanged.
      def create_full_src_file_name( a_file_name )
        if "" == mxx_sources_root
          return a_file_name
        else
          if a_file_name[ 0..mxx_sources_root.size ] == mxx_sources_root + '/'
            return a_file_name
          else
            return File.join( [ mxx_sources_root, a_file_name ] )
          end
        end
      end

      # Creation of complete result target file name.
      #
      # Result target file name depends on ObjPlacement and Toolset.
      #
      def create_full_result_target_file_name
        do_target_type_depending_actions(
            :lib => lambda { |toolset| toolset.full_lib_name( self ) },
            :dll => lambda { |toolset| toolset.full_dll_name( self ) },
            :macos_bundle => lambda { |toolset| toolset.full_dll_name( self ) },
            :exe => lambda { |toolset| toolset.full_exe_name( self ) } )
      end

    protected
      # Checks value of @mxx_all_options_defined and initiate
      # options setup if it is nil.
      def check_all_options_definition
        if !@mxx_all_options_defined
          # It's necessary, that toolset has defined all the options
          # taking all current modes into account.
          MxxRu::Cpp::toolset.setup_mandatory_options( self )
          @mxx_all_options_defined = true
        end
      end

      # Performs 'build' in normal mode.
      def normal_mode_builder
        MxxRu::Cpp::Mode.instance.before_subprj_processing
        # The build of subordinated projects performed automatically
        # taking dry-run mode into account.
        required_prjs_state = build_required_projects
        MxxRu::Cpp::Mode.instance.after_subprj_processing

        # Since v.1.6.11 options are set only after building
        # of required projects.
        check_all_options_definition

        run_generators
        if !MxxRu::Cpp::Mode.instance.is_no_depends_analyzer
          run_depends_analyzer
        end
        objs_state = compile_sources

        res_state = compile_resources

        link_state = link_target(
          required_prjs_state,
          objs_state, res_state )

        # The status of the target is determined by result of target linking.
        link_state
      end

      # Performs 'clean' in normal mode
      def normal_mode_cleaner
        MxxRu::Cpp::Mode.instance.before_subprj_processing
        # The cleanup of subordinated projects performed automatically
        # taking dry-run mode into account.
        clean_required_prjs
        MxxRu::Cpp::Mode.instance.after_subprj_processing

        # Since v.1.6.11 options are set only after processing
        # of required projects.
        check_all_options_definition

        clean_generators
        clean_objs
        clean_resources
        clean_target
      end

      # Performs 'build' in 'option extraction mode'
      def option_extraction_mode_builder
        options = {
            'Compiler options' => :mxx_all_compiler_options,
            'C++ Compiler Options' => :mxx_all_cpp_compiler_options,
            'C Compiler Options' => :mxx_all_c_compiler_options,
            'Linker Options' => :mxx_all_linker_options,
            'Librarian Options' => :mxx_all_librarian_options,
            'Defines' => :mxx_all_defines,
            'Include Paths' => :mxx_all_include_paths
            }
        if 'mswin' == toolset.tag( 'target_os' )
          options.merge!(
            'MSWindows Resource Compiler Options' => :mxx_all_mswin_rc_options,
            'MSWindows Resource Compiler Include Paths' =>
                :mxx_all_mswin_rc_include_paths,
            'MSWindows Resource Compiler Defines' => :mxx_all_mswin_rc_defines,
            'MSWindows Resource Linker Options' => :mxx_all_mswin_rlink_options
          )
        end

        options.keys.sort.each do |k|
          opts = self.send( options[ k ] )

          puts "#{k}:\n\t#{opts.join("\n\t")}\n"
        end

        return TargetState.new( TargetState::EXISTS )
      end

      # Performs 'clean' in 'option extraction mode'
      def option_extraction_mode_cleaner
        # Nothing to do!
      end

      # Start of all generators of source files.
      def run_generators
        mxx_generators.each { |g| g.build( self ) }
      end

      # Start of C++ dependencies analyzer.
      def run_depends_analyzer
        analyzer = Analyzer::CppAnalyzer.instance

        # We prepare the analyzer to start.
        sources = Array.new
        mxx_c_files.each { |s| sources << s.name }
        mxx_cpp_files.each { |s| sources << s.name }
        analyzer.add_sources( sources )

        analyzer.add_include_paths( mxx_all_include_paths )

        # Performing analisys.
        analyzer.analyze mxx_encoding

        # Getting results of analysis.
        mxx_c_files.each { |s|
          depends = analyzer.get_depends_for( s.name )
          s.depends << depends
          s.depends.flatten!
        }

        mxx_cpp_files.each { |s|
          depends = analyzer.get_depends_for( s.name )
          s.depends << depends
          s.depends.flatten!
        }
      end

      # Start of source files compiler.
      # Returns MxxRu::TargetState.
      def compile_sources
        return MxxRu::Cpp::toolset().make_objs( self )
      end

      # Start of resource files compiler.
      # Returns nothing.
      def compile_resources
        # We proceed from the assumption, that on one platform
        # some different types of resource files simultaneously can not be used.
        if @mxx_mswin_rc_file
          return MxxRu::Cpp::toolset().make_mswin_res( self )
        end

        return MxxRu::TargetState.new( TargetState::EXISTS )
      end

      # Start of target linking.
      #
      # [_required_prjs_state_] Build results of all subprojects required.
      #                         Object of MxxRu::TargetState type.
      # [_objs_state_] Compilation results of all object files
      #                Object of MxxRu::TargetState type.
      # [_res_state_] Compilation results of all resource files.
      #               Object of MxxRu::TargetState type.
      def link_target(
        required_prjs_state,
        objs_state,
        res_state )

        rebuilding_needed = TargetState.rebuilding_needed?(
            required_prjs_state,
            objs_state,
            res_state )

        state = do_target_type_depending_actions(
            :lib => lambda { |toolset|
                toolset.make_lib( self, rebuilding_needed ) },
            :dll => lambda { |toolset|
                toolset.make_dll( self, rebuilding_needed ) },
            :macos_bundle => lambda { |toolset|
                toolset.make_dll( self, rebuilding_needed ) },
            :exe => lambda { |toolset|
                toolset.make_exe( self, rebuilding_needed ) } )

        state
      end

      def clean_generators
        mxx_generators.each { |g| g.clean( self ) }
      end

      def clean_objs
        MxxRu::Cpp::toolset().clean_objs( self )
      end

      def clean_resources
        if @mxx_mswin_rc_file
          MxxRu::Cpp::toolset().clean_mswin_res( self )
        end
      end

      def clean_target
        do_target_type_depending_actions(
          :lib => lambda { |toolset| toolset.clean_lib( self ) },
          :dll => lambda { |toolset| toolset.clean_dll( self ) },
          :macos_bundle => lambda { |toolset| toolset.clean_dll( self ) },
          :exe => lambda { |toolset| toolset.clean_exe( self ) } )
      end

      # Do action with toolset depending on target type.
      #
      # Since v.1.4.7
      #
      # [_a_type_actions_] Hash with keys :lib, :dll and :exe.
      #                    Values should be Proc instances.
      #
      # Usage:
      #   do_target_type_depending_actions(
      #     :lib => lambda { |toolset| toolset.clean_lib(self) },
      #     :dll => lambda { |toolset| toolset.clean_dll(self) },
      #     :macos_bundle => lambda { |toolset| toolset.clean_dll(self) },
      #     :exe => lambda { |toolset| toolset.clean_exe(self) }
      #   )
      #
      def do_target_type_depending_actions( a_type_actions )
        toolset = MxxRu::Cpp::toolset
        if nil == target_type
          raise MxxRu::UnsupportedTargetTypeEx.new(
            self.class.name, "nil" )
        elsif target_type.name == LibTargetType::TYPE
          a_type_actions[ :lib ].call( toolset )
        elsif target_type.name == DllTargetType::TYPE
          a_type_actions[ :dll ].call( toolset )
        elsif target_type.name == MacOSBundleTargetType::TYPE
          a_type_actions[ :macos_bundle ].call( toolset )
        elsif target_type.name == ExeTargetType::TYPE
          a_type_actions[ :exe ].call( toolset )
        else
          raise MxxRu::UnsupportedTargetTypeEx.new(
            self.class.name, target_type.name )
        end
      end

      def add_unique_to( what, to )
        if !to.include?( what )
          to.push( what )
        end
      end

    end # class Target

    # Base class for targets that may be lib or dll, or MacOS bundle.
    #
    # In that cases for a project should be defined special target class.
    # The class constructor should define two blocks:
    # * one of them is called if developer wants to build target 
    #   as a static library. Block is run during as_lib method call;
    # * second is called if developer wants to build target
    #   as a shared library. Block is run during as_dll method call;
    #
    # Starting from version 1.4.10 this approach could also be used for
    # bulding MacOS bundles.
    #
    # For example:
    #
    #  class Prj < MxxRu::Cpp::LibOrDllTarget
    #    TAG = "threads_1/prj.rb"
    #    def initialize( a_alias = TAG )
    #      super( a_alias, TAG )
    # 
    #      target( "threads.1.3" )
    # 
    #      threading_mode( MxxRu::Cpp::THREADING_MULTI )
    # 
    #      init_dll_block(
    #        Proc.new {
    #          rtl_mode( MxxRu::Cpp::RTL_SHARED )
    #          implib_path( "lib" )
    #          define( "THREADS_1__DLL", OPT_UPSPREAD )
    #        })
    #      # This is avaliable from v.1.4.10
    #      init_macos_bundle_block(
    #        Proc.new {
    #          rtl_mode( MxxRu::Cpp::RTL_SHARED )
    #        }
    #      )
    # 
    #      cpp_source( "threads.cpp" )
    #      cpp_source( "micro_time.cpp" )
    # 
    #      if toolset.tag( "target_os" ) == "mswin"
    #        define( "__WIN32__" )
    #        cpp_source( "win32/os_thread.cpp" )
    #      elsif toolset.tag( "target_os" ) == "unix"
    #        cpp_source( "posix/os_thread.cpp" )
    #      end
    # 
    #      define( "THREADS_1__PRJ" )
    #    end
    #  end
    #
    # After definition of such class for the target, project may allow
    # to use itself by the other projects using two methods.
    #
    # == Manual execution of as_lib, as_dll, as_macos_bundle methods
    #
    # If project defines one class only, inherited from LibOrDllTarget,
    # then all clients of that project should explicitly call as_lib or as_dll,
    # or as_macos_bundle method during a reference to required_prj:
    #
    #  class MyPrj < MxxRu::Cpp::ExeTarget
    #    def initialize( a_alias = "my_prj.rb" )
    #      super( a_alias )
    #      required_prj( "threads_1/prj.rb" ).as_dll( self )
    #    end
    #  end
    #
    # Such approach is not good because if in composite project several
    # projects at the same time uses threads_1/prj.rb, then in one of them
    # it's easy to forget to call as_dll. And it wouldn't be incorrect,
    # because as_dll may be called by other projects in composite project.
    #
    # == Definition of auxiliary target classes
    #
    # To make usage of target that may be lib or dll easier, project may
    # define two auxiliary classes, which would call methods as_lib, as_dll
    # and as_macos_bundle inside their constructors:
    #
    #  class Prj < MxxRu::Cpp::LibOrDllTarget
    #    def initialize( a_alias = "threads_1/prj.rb" )
    #      super( a_alias, "threads_1" )
    #      ...
    #    end
    #  end
    #  class Lib < Prj
    #    def super( a_alias = "threads_1/lib.rb" )
    #      super( a_alias )
    #      as_lib
    #    end
    #  end
    #  class Dll < Prj
    #    def super( a_alias = "threads_1/dll.rb" )
    #      super( a_alias )
    #      as_dll
    #    end
    #  end
    #  class MacOSBundle < Prj
    #    def super( a_alias = "threads_1/macos_bundle.rb" )
    #      super( a_alias )
    #      as_macos_bundle
    #    end
    #  end
    #
    # Then for using the target it would be enough to do that:
    #  class MyPrj < MxxRu::Cpp::ExeTarget
    #    def initialize( a_alias = "my_prj.rb" )
    #      super( a_alias )
    #      required_prj( "threads_1/dll.rb" )
    #    end
    #  end
    #
    class LibOrDllTarget < Target
      # Exception type, which is thrown if target type is undefined.
      class TypeNotDefinedEx < MxxRu::Ex
        # [_a_prj_alias_] Project alias, for which target type should be defined.
        def initialize( a_prj_alias )
          super( "#{a_prj_alias}: type of target not defined yet" )
        end
      end

      # For compatibility with previous versions.
      Type_not_defined_ex = TypeNotDefinedEx

      # Exception type, which is thrown if target type was already defined.
      class TypeAlreadyDefinedEx < MxxRu::Ex
        # [_a_prj_tag_] Target identifier, for which type is established.
        # [_a_type_] Already defined target type.
        def initialize( a_prj_alias, a_type )
          super( "#{a_prj_alias}: type of target already defined " +
            "as #{a_type}" )
        end
      end

      # For compatibility with previous versions.
      Type_already_defined_ex = TypeAlreadyDefinedEx

      # Map describing unique target identifiers and their types.
      # Used to avoid conflicts, when the same target is inside composite
      # project as both lib and dll.
      @@mxx_tags = Hash.new

      # Constructor.
      #
      # [_a_prj_alias_] Project alias.
      # [_a_prj_tag_] Unique value, which identifies target given.
      #               This value used to avoid conflicts in the next situations:
      #  class Prj < MxxRu::Cpp::LibOrDllTarget ... end
      #  class Lib < Prj ... end
      #  class Dll < Prj ... end
      #  class MacOSBundle < Prj ... end
      #  class MyExe < MxxRu::Cpp::ExeTarget
      #    def initialize( a_alias = "my_exe.rb" )
      #      super( a_alias )
      #      required_prj( "threads_1/dll.rb" )
      #    end
      #  end
      #  class MyDll < MxxRu::Cpp::DllTarget
      #    def initialize( a_alias = "my_dll.rb" )
      #      super( a_alias )
      #      required_prj( "threads_1/lib.rb" )
      #    end
      #  end
      # If both MyExe and MyDll are used in one composite project,
      # then it's a conflict of usage of threads_1/dll.rb and threads_1/lib.rb.
      def initialize( a_prj_alias, a_prj_tag )
        super( a_prj_alias )

        @mxx_prj_tag = a_prj_tag

        # Block, which is run if lib is a target.
        @mxx_init_lib_block = nil
        # Block, which is run if dll is a target.
        @mxx_init_dll_block = nil
        # Block, which is run if MacOS bundle is a target.
        @mxx_init_macos_bundle_block = nil
        # Target type should be defined or by as_lib method, or by as_dll method,
        # or by as_macos_bundle.
        @mxx_target_type = nil

      end

      # Set lib as target type.
      def as_lib
        if ensure_type_not_set_yet( LibTargetType::TYPE )
          @mxx_target_type = LibTargetType.new
          if nil != @mxx_init_lib_block
            @mxx_init_lib_block.call
          end
        end
      end

      # Set dll as target type.
      def as_dll
        if ensure_type_not_set_yet( DllTargetType::TYPE )
          @mxx_target_type = DllTargetType.new
          if nil != @mxx_init_dll_block
            @mxx_init_dll_block.call
          end
        end
      end

      # Set MacOS bundle as target type.
      def as_macos_bundle
        if ensure_type_not_set_yet( MacOSBundleTargetType::TYPE )
          @mxx_target_type = MacOSBundleTargetType.new
          if nil != @mxx_init_macos_bundle_block
            @mxx_init_macos_bundle_block.call
          end
        end
      end

      # Return current value of mxx_target_type.
      def target_type
        if nil == @mxx_target_type
          raise TypeNotDefinedEx.new( prj_alias )
        end

        return @mxx_target_type
      end

    protected
      # Set initialization block for lib.
      # This block will be executed in as_lib method.
      def init_lib_block( a_block )
        @mxx_init_lib_block = a_block
      end

      # Set initialization block for dll.
      # This block will be executed in as_dll method.
      def init_dll_block( a_block )
        @mxx_init_dll_block = a_block
      end

      # Set initialization block for MacOS bundle.
      # This block will be executed in as_macos_bundle method.
      def init_macos_bundle_block( a_block )
        @mxx_init_macos_bundle_block = a_block
      end

    private
      # If someone already defined other target type, exception is thrown.
      # If already defined target type is the same, false returned.
      # If target type is not defined, true is returned.
      #
      # [_a_type_name_] Name of target type defined.
      def ensure_type_not_set_yet( a_type_name )
        if @@mxx_tags.key?( @mxx_prj_tag ) &&
          @@mxx_tags[ @mxx_prj_tag ] != a_type_name
          raise TypeAlreadyDefinedEx.new(
            @mxx_prj_tag, a_type_name )
        end

        @@mxx_tags[ @mxx_prj_tag ] = a_type_name

        return ( nil == @mxx_target_type )
      end
    end # class LibOrDllTarget

    # For compatibility with previous versions.
    Lib_or_dll_target = LibOrDllTarget

    # Lib is a target type.
    class LibTarget < LibOrDllTarget
      def initialize( a_prj_alias, &block )
        super( a_prj_alias, a_prj_alias )

        as_lib

        instance_eval( &block ) if block
      end
    end

    # For compatibility with previous versions.
    Lib_target = LibTarget

    # Dll is a target type.
    class DllTarget < LibOrDllTarget
      def initialize( a_prj_alias, &block )
        super( a_prj_alias, a_prj_alias )

        as_dll

        instance_eval( &block ) if block
      end
    end

    # For compatibility with previous versions.
    Dll_target = DllTarget

    # MacOS bundle is a target type.
    class MacOSBundleTarget < LibOrDllTarget
      def initialize( a_prj_alias, &block )
        super( a_prj_alias, a_prj_alias )

        as_macos_bundle

        instance_eval( &block ) if block
      end
    end

    # For compatibility with previous versions.
    Macos_bundle_target = MacOSBundleTarget

    # Exe is a target type.
    class ExeTarget < Target

      def initialize( a_alias, &block )
        super( a_alias )

        instance_eval( &block ) if block
      end

      def target_type
        return ExeTargetType.new
      end
    end

    # For compatibility with previous versions.
    Exe_target = ExeTarget

    # Generator for generating method for simple way of describing targets.
    def Cpp.generate_simple_target_method( target_type )
      target_class_name = target_type.to_s.capitalize
      class_eval %Q{
        def Cpp.#{target_type}( prj_alias = nil, &block )
          prj_alias = MxxRu::Util::prj_alias_form_caller( caller ) unless prj_alias
          MxxRu::setup_target #{target_class_name}.new( prj_alias, &block )
        end
      }
    end

    generate_simple_target_method :lib_target
    generate_simple_target_method :dll_target
    generate_simple_target_method :macos_bundle_target
    generate_simple_target_method :exe_target

  end # module Cpp
end # module MxxRu

