#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/cpp/toolset'

module  MxxRu
  module  Cpp
    module Toolsets

      # Toolset implemetation for c89 NonStopKernel.
      class C89NskFamily < MxxRu::Cpp::Toolset
      public
        # Tag name, points to target platform is
        # Open System Services on NonStopKernel.
        TANDEM_OSS = "tandem_oss"

        def initialize( a_name )
          super( a_name )
        end

        # Returns C compiler name
        def c_compiler_name
          tag( [ C_COMPILER_NAME_TAG, COMPILER_NAME_TAG ], "c89" )
        end

        # Returns C++ compiler name.
        def cpp_compiler_name
          tag( [ CPP_COMPILER_NAME_TAG, COMPILER_NAME_TAG ], "c89" )
        end

        # Returns linker name.
        def linker_name
          tag( LINKER_NAME_TAG, "c89" )
        end

        # Returns librarian name.
        def librarian_name
          tag( LIBRARIAN_NAME_TAG, "ar" )
        end

        # See description at MxxRu::Cpp::Toolset#setup_mandatory_options.
        def setup_mandatory_options( target )

          if TANDEM_OSS == tag( "target_os" )
            target.compiler_option( "-Wsystype=oss" )
            target.linker_option( "-Wsystype=oss" )
          end

          target.compiler_option( "-Wcall_shared" )

          if RUNTIME_DEBUG == target.mxx_runtime_mode
            target.compiler_option( "-g" )
            target.linker_option( "-g" )
          elsif RUNTIME_RELEASE == target.mxx_runtime_mode
            target.define( "NDEBUG" )
            target.linker_option( "-s" )
            if OPTIM_SIZE == target.mxx_optimization
              target.compiler_option( "-Woptimize=1" )
            else
              target.compiler_option( "-Woptimize=2" )
            end
          end

          target.compiler_option( "-WIEEE_float" )
          target.compiler_option( "-Wnowarn=1255" )

          target.cpp_compiler_option( "-Wcplusplus" )
          target.linker_option( "-Wcplusplus" )
          target.cpp_compiler_option( "-Wversion3" )
          target.linker_option( "-Wversion3" )

          # No warnings:
          # 1506 - implicit conversion form "unsigned int" to long.
          # 770 - variable set but never used.
          # 734 - function not inlined.
          # 495 - delete of pointer to incomplete class.
          # 270 - pointless comparision of unsigned integer with zero.
          # 262 - variable declared but never referenced.
          # 101 - last line of file ends without a newline.
          target.compiler_option( "-Wnowarn=1506,770,734,495,270,262,101" )

          target.mxx_all_defines.each { |d|
            target.compiler_option( "-D " + d )
          }

          target.mxx_all_include_paths.each { |p|
            target.compiler_option( "-I " + p )
          }
        end

        # See description at MxxRu::Cpp::Toolset#obj_file_ext.
        def obj_file_ext
          return ".o"
        end

        # See description at MxxRu::Cpp::Toolset#make_c_obj_command_lines.
        def make_c_obj_command_lines(
          obj_name,
          source_name,
          compiler_options,
          target )

          cmd_line = "-c -o #{obj_name} " +
            "#{compiler_options.join(' ')} #{source_name}"

          return [ "#{c_compiler_name} #{cmd_line}" ]
        end

        # See description at MxxRu::Cpp::Toolset#make_c_obj_command_lines.
        def make_cpp_obj_command_lines(
          obj_name,
          source_name,
          compiler_options,
          target )

          cmd_line = "-c -o #{obj_name} " +
            "#{compiler_options.join(' ')} #{source_name}"

          return [ "#{cpp_compiler_name} #{cmd_line}" ]
        end

        # See description at MxxRu::Cpp::Toolset#lib_file_name.
        def lib_file_name( source_name, target )
          return construct_target_name( source_name, "lib", ".a", target )
        end

        # See description at MxxRu::Cpp::Toolset#lib_link_name.
        def lib_link_name( source_name, target )
          return source_name
        end

        # See description at MxxRu::Cpp::Toolset#make_lib_command_lines.
        def make_lib_command_lines(
          lib_name,
          obj_files,
          librarian_options,
          target )

          result = "-r #{librarian_options.join(' ')} " +
            "#{lib_name} #{obj_files.join(' ')}"

          return [ "#{librarian_name} #{result}" ]
        end

        # See description at MxxRu::Cpp::Toolset#dll_file_name.
        def dll_file_name( source_name, target )
          return construct_target_name( source_name, 'lib', '.so', target )
        end

        # See description at MxxRu::Cpp::Toolset#implib_link_name.
        def implib_link_name(
          dll_real_name,
          target )

          # Target name should be passed to linker on UNIX platform
          return lib_link_name(
            File.basename( target.mxx_target_name ), target )
        end

        # See description at MxxRu::Cpp::Toolset#implib_link_path.
        def implib_link_path(
          dll_real_name,
          dll_real_path,
          target )

          # DLL name is the import library name on UNIX platform.
          return dll_real_path
        end

        # See description at MxxRu::Cpp::Toolset#make_dll_command_lines.
        def make_dll_command_lines(
          a_dll_name,
          a_dll_info,
          a_linker_lists,
          a_target )

          result = "-Wshared " +
            "#{a_linker_lists.linker_options.join(' ')} " +
            "-Wld=-export_all " +
            "-o #{a_dll_name} "

          a_linker_lists.lib_paths.each { |p| result << "-L #{p} " }

          result << "#{a_linker_lists.objs.join(' ')} "
          a_linker_lists.libs.each { |l| result << "-l#{l} " }

          return [ "#{linker_name} #{result}" ]
        end

        # See description at MxxRu::Cpp::Toolset#make_dll_requirements.
        def make_dll_requirements(
          a_dll_name,
          a_dll_info,
          a_linker_lists,
          a_target )

          result = DllRequirements.new

#FIX ME: This should be checked on a real translator.
          # On UNIX all requirements are dependencies.

          # This library, as a first one.
          result.add_libs( [ a_dll_info.link_name ] )
          result.add_lib_paths( [ a_dll_info.link_path ] )

          # And all required libraries.
          a_target.mxx_required_prjs.each { |d|
            if Toolset::has_linkable_dependecies?( d )
              result.add_libs( d.mxx_required_libs )
              result.add_lib_paths( d.mxx_required_lib_paths )
            end
          }

          return result
        end

        # See description at MxxRu::Cpp::Toolset#exe_file_name.
        def exe_file_name( source_name, target )
          return construct_target_name( source_name, NO_PREFIX, NO_SUFFIX, target )
        end

        # See description at MxxRu::Cpp::Toolset#make_exe_command_lines.
        def make_exe_command_lines(
          a_exe_name,
          a_exe_info,
          a_linker_lists,
          a_target )

          result = "-Wcall_shared " +
            "#{a_linker_lists.linker_options.join(' ')} " +
            "-o #{a_exe_name} "

          a_linker_lists.lib_paths.each { |p| result << "-L #{p} " }

          result << "#{a_linker_lists.objs.join(' ')} "
          a_linker_lists.libs.each { |l| result << "-l#{l} " }

          return [ "#{linker_name} #{result}" ]
        end

      end # class C89NskFamily


      # For compatibility with previous versions.
      C89_nsk_family = C89NskFamily

    end # module Toolsets

  end # module Cpp

end # module MxxRu

