#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2017, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/binary_library'
require 'mxx_ru/cpp/toolsets/gcc_family'

module  MxxRu
  module  Cpp
    module Toolsets

      # Toolset implemetation for Clang compiler.
      class ClangFamily < GccFamily
      public
        def initialize( name )
          super( name )
        end

        # See description at MxxRu::Cpp::Toolset#setup_mandatory_options.
        def setup_mandatory_options( target )

          if RUNTIME_DEBUG == target.mxx_runtime_mode
            target.compiler_option( "-g" )
            target.linker_option( "-g" )
          elsif RUNTIME_RELEASE == target.mxx_runtime_mode
            target.define( "NDEBUG" )

            # -s switch is non-supported on clang+msvc.
            if 'mswin' != tag( 'host_os' )
              target.linker_option( "-s" )
            end

            if OPTIM_SIZE == target.mxx_optimization
              target.compiler_option( "-Os" )
            else
              target.compiler_option( "-O2" )
            end
          end

          if RTTI_DISABLED == target.mxx_rtti_mode
            target.cpp_compiler_option( "-fno-rtti" )
          end

          # -fPIC switch is not supported on clang+msvc since clang-4.0.0
          if 'mswin' != tag( 'host_os' )
            target.compiler_option( "-fPIC" )
          end

          target.mxx_all_defines.each { |d|
            target.compiler_option( "-D" + d )
          }

          target.mxx_all_include_paths.each { |p|
            target.compiler_option( "-I" + p )
          }

          if CPP_STD17 == cpp_std
            target.cpp_compiler_option( '-std=c++17' )
          elsif CPP_STD14 == cpp_std
            target.cpp_compiler_option( '-std=c++14' )
          elsif CPP_STD11 == cpp_std
            target.cpp_compiler_option( "-std=c++11" )
          end

          if CPP_STD11 <= cpp_std and
              THREADING_MULTI == target.mxx_threading_mode
            target.linker_option( '-pthread' )
          end
        end

        # Returns C compiler name.
        def c_compiler_name
          tag( [ C_COMPILER_NAME_TAG, COMPILER_NAME_TAG ], "clang" )
        end

        # Returns C++ compiler name.
        def cpp_compiler_name
          tag( [ CPP_COMPILER_NAME_TAG, COMPILER_NAME_TAG ], "clang++" )
        end

        # Returns linker name.
        def linker_name
          tag( LINKER_NAME_TAG, "clang++" )
        end

        # Returns librarian name.
        def librarian_name
          tag( LIBRARIAN_NAME_TAG, "ar" )
        end

        # Special implementation for support explicit library type linker option.
        def make_linker_include_lib_options( target, libs )
          current_lib_type = nil
          all_libs = libs.inject( '' ) { |r, l|
              # Insert new lib mode switch only if next library has different
              # type than current type.
              if nil == current_lib_type || current_lib_type != l.type
                r << switch_to_default_lib_mode_if_needed( current_lib_type )
                current_lib_type = l.type
                if BinaryLibrary::ANY != l.type &&
                    default_lib_linking_mode != l.type
                  r << lib_linking_mode_switch( l.type )
                end
              end

              r << '-l' << port_specific_lib_name_checker( l.name ) << ' '
          }
          all_libs << switch_to_default_lib_mode_if_needed( current_lib_type )
          enclose_linker_include_lib_options_into_brackes( all_libs )
        end


        # Return default library linking mode for current platform.
        #
        # Default mode BinaryLibrary::SHARED for Unix with GCC assumed.
        # But default mode can be specified in toolset tag 'lib_linking_mode'.
        def default_lib_linking_mode
          t = tag( 'lib_linking_mode', 'unknown' )
          if 'unknown' != t
            'static' == t ? BinaryLibrary::STATIC : BinaryLibrary::SHARED
          else
            BinaryLibrary::SHARED
          end
        end

        # Return command line switch for forcing specified library type.
        def lib_linking_mode_switch( linking_mode )
          "-Wl,#{linking_mode == BinaryLibrary::SHARED ? '-Bdynamic' : '-Bstatic'} "
        end

        # Return command line switch for closing group of libraries with same type.
        # Empty string returned if current_lib_type is nil of is current_lib_type
        # is default lib linking mode on this platform.
        def switch_to_default_lib_mode_if_needed( current_lib_type )
          if nil != current_lib_type &&
              BinaryLibrary::ANY != current_lib_type &&
              default_lib_linking_mode != current_lib_type
            lib_linking_mode_switch( default_lib_linking_mode )
          else
            ''
          end
        end
        
      private
        # Create toolset identification string.
        def make_toolset_id_string
          result = IO.popen( "#{c_compiler_name} -v 2>&1", :err => [:child, :out] ) do |io|
            target = 'generic'
            version = 'unknown'
            io.each_line do |line|
              if /^Target:\s*(?<trgt>\S+)/ =~ line
                target = trgt
              elsif /clang version (?<v>\S+)/ =~ line
                version = v
              end
            end
            version + '--' + target
          end
          "clang_#{result}"
        end
      end # class ClangFamily

      # Toolset implemetation for Clang compiler for Win32.
      class ClangMswinFamily < ClangFamily
      public
        # See description at MxxRu::Cpp::Toolset#setup_mandatory_options.
        def setup_mandatory_options( target )

          super( target )

          # All defines and all include_path should be applied
          # to resource compiler too.
          target.mxx_all_defines.each { |d|
            target.compiler_option( "-D" + d )
            target.mswin_rc_option( "/d" + d )
          }

          target.mxx_all_include_paths.each { |p|
            target.compiler_option( "-I" + p )
            target.mswin_rc_option( "/i" + p )
          }

          # Resource compiler specific options.
          target.mxx_all_mswin_rc_defines.each { |d|
            target.mswin_rc_option( "/d" + d )
          }
          target.mxx_all_mswin_rc_include_paths.each { |p|
            target.mswin_rc_option( "/i" + p )
          }
        end

        # Returns librarian name.
        def librarian_name
          tag( LIBRARIAN_NAME_TAG, "llvm-ar" )
        end

        # Returns resource compiler name.
        def rc_name
          return tag( RC_NAME_TAG, "rc" )
        end

        # See description at MxxRu::Cpp::Toolset#exe_file_name.
        def exe_file_name( source_name, target )
          return construct_target_name( source_name, NO_PREFIX, ".exe", target )
        end

        # See description at MxxRu::Cpp::Toolset#lib_file_name.
        def lib_file_name( source_name, target )
          return construct_target_name( source_name, NO_PREFIX, '.lib', target )
        end

        # See description at MxxRu::Cpp::Toolset#dll_file_name.
        def dll_file_name( source_name, target )
          return construct_target_name( source_name, NO_PREFIX, '.dll', target )
        end

        # See description at MxxRu::Cpp::Toolset#make_lib_command_lines.
        def make_lib_command_lines(
          lib_name,
          obj_files,
          librarian_options,
          target )

          result = "r #{librarian_options.join(' ')} " +
            "#{lib_name} #{obj_files.join(' ')}"

          return [ "#{librarian_name} #{result}" ]
        end

        # See description at MxxRu::Cpp::Toolset#clean_dll_specific_files.
        #
        # Delete import library if exists.
        def clean_dll_specific_files(
          a_dll_file,
          a_dll_info,
          a_target )

          # Delete import library if exists.
          if nil != a_dll_info.link_name
            implib_name = File.join( [ a_dll_info.link_path,
              lib_file_name( a_dll_info.link_name, a_target ) ] )
            MxxRu::Util::delete_file( implib_name )
          end
        end

        # See description at MxxRu::Cpp::Toolset#implib_link_name.
        def implib_link_name(
          dll_real_name,
          target )

          # It's required to pass import library name to linker on mswin platform
          if nil != target.mxx_implib_path
            return lib_link_name( target.mxx_target_name, target )
          end

          return nil
        end

        # See description at MxxRu::Cpp::Toolset#implib_link_path.
        def implib_link_path(
          dll_real_name,
          dll_real_path,
          target )
          return target.mxx_obj_placement.get_lib(
            target.mxx_implib_path, self, target )
        end

        # See description at MxxRu::Cpp::Toolset#make_dll_requirements.
        def make_dll_requirements(
          a_dll_name,
          a_dll_info,
          a_linker_lists,
          a_target )

          result = DllRequirements.new

          # Dependencies are exists only if import library is present.
          if nil != a_dll_info.link_name
            result.add_libs( [ a_dll_info.link_name ] )
            result.add_lib_paths( [ a_dll_info.link_path ] )
          end

          return result
        end

        # Return string containing port-specific linker option for DLL linking.
        #
        # All parameters are similar to make_dll_command_lines parameters.
        #
        # Return empty string in a base class.
        def port_specific_dll_link_options(
          a_dll_name, a_dll_info, a_linker_lists, a_target )
          # Build import library if it's required.
          if nil != a_dll_info.link_name
            full_lib_name = File.join( [ a_dll_info.link_path,
              lib_file_name( a_dll_info.link_name, a_target ) ] )

            return "-Wl,/IMPLIB:#{full_lib_name}"
          end

          return ""
        end

        # Checks library name for suffix '.lib' and return name without
        # that suffix.
        def port_specific_lib_name_checker(library_name)
          if /\.lib$/i =~ library_name
            MxxRu::Util::remove_file_ext(library_name)            
          else
            library_name
          end
        end

        # Return command line switch for forcing specified library type.
        def lib_linking_mode_switch( linking_mode )
          ""
        end

        def enclose_linker_include_lib_options_into_brackes( options )
          " #{options} "
        end

        # See description at MxxRu::Cpp::Toolset#mswin_res_file_name.
        def mswin_res_file_name( source_name )
          return source_name + ".res"
        end

        # See description at MxxRu::Cpp::Toolset#make_mswin_res_command_lines.
        def make_mswin_res_command_lines(
          res_name,
          rc_file,
          rc_options,
          target )

          return [ "#{rc_name} " +
            "#{rc_options.join(' ')} /r " +
            "/fo#{res_name} #{rc_file}" ]
        end

      end # class ClangMswinFamily

    end # module Toolsets

  end # module Cpp

end # module MxxRu

