#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/binary_library'
require 'mxx_ru/cpp/toolset'

module  MxxRu
  module  Cpp
    module Toolsets

      # Toolset implemetation for GCC compiler.
      class GccFamily < MxxRu::Cpp::Toolset
      public
        # Tag name, indicates gcc port beeing used.
        GCC_PORT_TAG  = "gcc_port"
        # GCC_PORT: GCC on Unix/Linux.
        GCC_PORT_UNIX = "unix"
        # GCC_PORT: MinGW on Win32.
        GCC_PORT_MINGW = "mingw"
        # GCC_PORT: Cygwin on Win32.
        GCC_PORT_CYGWIN = "cygwin"

        def initialize( a_name )
          super( a_name )
        end

        # Returns C compiler name.
        def c_compiler_name
          tag( [ C_COMPILER_NAME_TAG, COMPILER_NAME_TAG ], "gcc" )
        end

        # Returns C++ compiler name.
        def cpp_compiler_name
          tag( [ CPP_COMPILER_NAME_TAG, COMPILER_NAME_TAG ], "g++" )
        end

        # Returns linker name.
        def linker_name
          tag( LINKER_NAME_TAG, "g++" )
        end

        # Returns librarian name.
        def librarian_name
          tag( LIBRARIAN_NAME_TAG, "ar" )
        end

        # See description at MxxRu::Cpp::Toolset#setup_mandatory_options.
        def setup_mandatory_options( target )

          if RUNTIME_DEBUG == target.mxx_runtime_mode
            target.compiler_option( "-g" )
            target.linker_option( "-g" )
          elsif RUNTIME_RELEASE == target.mxx_runtime_mode
            target.define( "NDEBUG" )
            target.linker_option( "-s" )
            if OPTIM_SIZE == target.mxx_optimization
              target.compiler_option( "-Os" )
            else
              target.compiler_option( "-O2" )
            end
          end

          if RTTI_DISABLED == target.mxx_rtti_mode
            target.cpp_compiler_option( "-fno-rtti" )
          end

          if RTL_STATIC == target.mxx_rtl_mode
            target.linker_option( "-static-libgcc" )
          else
            target.linker_option( "-shared-libgcc" )
          end

          target.mxx_all_defines.each { |d|
            target.compiler_option( "-D" + d )
          }

          target.mxx_all_include_paths.each { |p|
            target.compiler_option( "-I" + p )
          }

          if CPP_STD17 == cpp_std
            target.cpp_compiler_option( '-std=c++17' )
          elsif CPP_STD14 == cpp_std
            target.cpp_compiler_option( '-std=c++14' )
          elsif CPP_STD11 == cpp_std
            target.cpp_compiler_option( '-std=c++11' )
          end

          if CPP_STD11 <= cpp_std and
              THREADING_MULTI == target.mxx_threading_mode
            target.linker_option( '-pthread' )
          end
        end

        # See description at MxxRu::Cpp::Toolset#obj_file_ext.
        def obj_file_ext
          return ".o"
        end

        # See description at MxxRu::Cpp::Toolset#make_c_obj_command_lines.
        def make_c_obj_command_lines(
          obj_name,
          source_name,
          compiler_options,
          target )

          cmd_line = "-c -o #{obj_name} " +
            "#{compiler_options.join(' ')} #{source_name}"

          return [ "#{c_compiler_name} #{cmd_line}" ]
        end

        # See description at MxxRu::Cpp::Toolset#make_c_obj_command_lines.
        def make_cpp_obj_command_lines(
          obj_name,
          source_name,
          compiler_options,
          target )

          cmd_line = "-c -o #{obj_name} " +
            "#{compiler_options.join(' ')} #{source_name}"

          return [ "#{cpp_compiler_name} #{cmd_line}" ]
        end

        # See description at MxxRu::Cpp::Toolset#lib_file_name.
        def lib_file_name( source_name, target )
          return construct_target_name( source_name, 'lib', '.a', target )
        end

        # See description at MxxRu::Cpp::Toolset#lib_link_name.
        def lib_link_name( source_name, target )
          return source_name
        end

        # See description at MxxRu::Cpp::Toolset#make_lib_command_lines.
        def make_lib_command_lines(
          lib_name,
          obj_files,
          librarian_options,
          target )

          result = "-r #{librarian_options.join(' ')} " +
            "#{lib_name} #{obj_files.join(' ')}"

          return [ "#{librarian_name} #{result}" ]
        end

        # See description at MxxRu::Cpp::Toolset#dll_file_name.
        def dll_file_name( source_name, target )
          return construct_target_name( source_name, 'lib', '.so', target )
        end

        # See description at MxxRu::Cpp::Toolset#implib_link_name.
        def implib_link_name(
          dll_real_name,
          target )

          # Target name should be passed to linker on UNIX platform
          return lib_link_name(
            File.basename( target.mxx_target_name ), target )
        end

        # See description at MxxRu::Cpp::Toolset#implib_link_path.
        def implib_link_path(
          dll_real_name,
          dll_real_path,
          target )

          # DLL name is the import library name on UNIX platform.
          return dll_real_path
        end

        # See description at MxxRu::Cpp::Toolset#make_dll_command_lines.
        def make_dll_command_lines(
          a_dll_name,
          a_dll_info,
          a_linker_lists,
          a_target )

          result = "-shared " +
            "#{a_linker_lists.linker_options.join(' ')} " +
            "-o #{a_dll_name} "

          a_linker_lists.lib_paths.each { |p| result << "-L#{p} " }

          result << "#{a_linker_lists.objs.join(' ')} "
          result << make_linker_include_lib_options(
              a_target, a_linker_lists.libs )

          result << port_specific_dll_link_options(
            a_dll_name, a_dll_info, a_linker_lists, a_target )

          return [ "#{linker_name} #{result}" ]
        end

        # See description at MxxRu::Cpp::Toolset#make_dll_requirements.
        def make_dll_requirements(
          a_dll_name,
          a_dll_info,
          a_linker_lists,
          a_target )

          result = DllRequirements.new

          # On UNIX all requirements are dependencies.

          # This library, as a first one.
          result.add_libs( [
            BinaryLibrary.new( a_dll_info.link_name, BinaryLibrary::SHARED ) ] )
          result.add_lib_paths( [ a_dll_info.link_path ] )

          # And all required libraries.
          a_target.mxx_required_prjs.each { |d|
            if Toolset::has_linkable_dependecies?( d )
              result.add_libs( d.mxx_required_libs )
              result.add_lib_paths( d.mxx_required_lib_paths )
            end
          }

          return result
        end

        # See description at MxxRu::Cpp::Toolset#exe_file_name.
        def exe_file_name( source_name, target )
          return construct_target_name( source_name, NO_PREFIX, NO_SUFFIX, target )
        end

        # See description at MxxRu::Cpp::Toolset#make_exe_command_lines.
        def make_exe_command_lines(
          a_exe_name,
          a_exe_info,
          a_linker_lists,
          a_target )

          result = "#{a_linker_lists.linker_options.join(' ')} " +
            "-o #{a_exe_name} "

          a_linker_lists.lib_paths.each { |p| result << "-L#{p} " }

          result << "#{a_linker_lists.objs.join(' ')} "
          result << make_linker_include_lib_options(
              a_target, a_linker_lists.libs )

          result << port_specific_exe_link_options(
            a_exe_name, a_exe_info, a_linker_lists, a_target )

          return [ "#{linker_name} #{result}" ]
        end

        # Make fragment of linker command line which contains names
        # of included libraries.
        #
        # [_target_] target for that this fragment will be prepared.
        # [_libs_] list of BinaryLibrary with required libraries.
        def make_linker_include_lib_options( target, libs )
          all_libs = libs.inject( '' ) { |r, l|
              r << '-l' << port_specific_lib_name_checker( l.name ) << ' '
          }
          enclose_linker_include_lib_options_into_brackes( all_libs )
        end

        # Enclose library list into brackets if necessary.
        #
        # For example, on Linux -Wl,--start-group and -Wl,--end-group
        # should be used around library list.
        #
        def enclose_linker_include_lib_options_into_brackes( options )
          "-Wl,--start-group #{options} -Wl,--end-group "
        end

        # Return string containing port-specific linker option for DLL linking.
        #
        # All parameters are similar to make_dll_command_lines parameters.
        #
        # Return empty string in a base class.
        def port_specific_dll_link_options(
          a_dll_name, a_dll_info, a_linker_lists, a_target )
          return ""
        end

        # Return string containing port-specific linker option for EXE linking.
        #
        # All parameters are similar to make_exe_command_lines parameters.
        #
        # Return empty string in a base class.
        def port_specific_exe_link_options(
          a_exe_name, a_exe_info, a_linker_lists, a_target )
          return ""
        end

        # Return correct name of library to be given for linker.
        def port_specific_lib_name_checker(library_name)
          library_name
        end

        # Create toolset identification string.
        def make_toolset_id_string
          result = IO.popen( "#{c_compiler_name} -v 2>&1", :err => [:child, :out] ) do |io|
            target = 'generic'
            version = 'unknown'
            io.each_line do |line|
              if /^Target:\s*(?<trgt>\S+)/ =~ line
                target = trgt
              elsif /^gcc version (?<v>\S+)/ =~ line
                version = v
              end
            end
            version + '--' + target
          end
          "gcc_#{result}"
        end
      end # class GccFamily

      # For compatibility with previous versions.
      Gcc_family = GccFamily

      # Toolset implemetation for GCC compiler for Win32.
      # This class is a base class for both MinGW and Cygwin
      class GccMswinFamily < GccFamily
      public

        # See description at MxxRu::Cpp::Toolset#exe_file_name.
        def exe_file_name( source_name, target )
          return construct_target_name( source_name, NO_PREFIX, ".exe", target )
        end

        # See description at MxxRu::Cpp::Toolset#clean_dll_specific_files.
        #
        # Delete import library if exists.
        def clean_dll_specific_files(
          a_dll_file,
          a_dll_info,
          a_target )

          # Delete import library if exists.
          if nil != a_dll_info.link_name
            implib_name = File.join( [ a_dll_info.link_path,
              lib_file_name( a_dll_info.link_name, a_target ) ] )
            MxxRu::Util::delete_file( implib_name )
          end
        end

        # See description at MxxRu::Cpp::Toolset#implib_link_name.
        def implib_link_name(
          dll_real_name,
          target )

          # It's required to pass import library name to linker on mswin platform
          if nil != target.mxx_implib_path
            return lib_link_name( target.mxx_target_name, target )
          end

          return nil
        end

        # See description at MxxRu::Cpp::Toolset#implib_link_path.
        def implib_link_path(
          dll_real_name,
          dll_real_path,
          target )
          return target.mxx_obj_placement.get_lib(
            target.mxx_implib_path, self, target )
        end

        # See description at MxxRu::Cpp::Toolset#make_dll_requirements.
        def make_dll_requirements(
          a_dll_name,
          a_dll_info,
          a_linker_lists,
          a_target )

          result = DllRequirements.new

          # Dependencies are exists only if import library is present.
          if nil != a_dll_info.link_name
            result.add_libs( [ a_dll_info.link_name ] )
            result.add_lib_paths( [ a_dll_info.link_path ] )
          end

          return result
        end

        # Return string containing port-specific linker option for DLL linking.
        #
        # All parameters are similar to make_dll_command_lines parameters.
        #
        # Return empty string in a base class.
        def port_specific_dll_link_options(
          a_dll_name, a_dll_info, a_linker_lists, a_target )
          # Build import library if it's required.
          if nil != a_dll_info.link_name
            full_lib_name = File.join( [ a_dll_info.link_path,
              lib_file_name( a_dll_info.link_name, a_target ) ] )

            return "-Wl,--out-implib=#{full_lib_name}" +
              ",--export-all-symbols"
          end

          return ""
        end

        # Checks library name for suffix '.lib' and return name without
        # that suffix.
        def port_specific_lib_name_checker(library_name)
          if /\.lib$/i =~ library_name
            MxxRu::Util::remove_file_ext(library_name)            
          else
            library_name
          end
        end

      end # class GccMswinFamily

      # For compatibility with previous versions.
      Gcc_mswin_family = GccMswinFamily

    end # module Toolsets

  end # module Cpp

end # module MxxRu

