#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/cpp/toolsets/gcc_family'

module  MxxRu
  module  Cpp
    module Toolsets

      # Toolset implemetation for GCC compiler for Win32.
      # This class is a base class for both MinGW and Cygwin
      class GccMingw < GccMswinFamily
      public

        def initialize( a_name = "gcc" )
          super( a_name )

          setup_tag( GCC_PORT_TAG, GCC_PORT_MINGW )
          setup_tag( "host_os", "mswin" )
          setup_tag( "target_os", "mswin" )
        end

        # Returns resource compiler name.
        def rc_name
          tag( RC_NAME_TAG, "windres" )
        end

        # See description at MxxRu::Cpp::Toolset#setup_mandatory_options.
        #
        # Setups resource compiler parameters.
        #
        # Required options are set up based on screen_mode.
        def setup_mandatory_options( target )
          super( target )

          if SCREEN_WINDOW == target.mxx_screen_mode
            target.linker_option( "-mwindows" )
          else
            target.linker_option( "-mconsole" )
          end

          if THREADING_MULTI == target.mxx_threading_mode
            target.compiler_option( "-mthreads" )
          end

          # All defines and all include_path should be distributed to resource compiler.
          target.mxx_all_defines.each { |d|
            target.mswin_rc_option( "-D " + d )
          }

          target.mxx_all_include_paths.each { |p|
            target.mswin_rc_option( "--include-dir=" + p )
          }

          # Personal resource compiler options.
          target.mxx_all_mswin_rc_defines.each { |d|
            target.mswin_rc_option( "-D " + d )
          }
          target.mxx_all_mswin_rc_include_paths.each { |p|
            target.mswin_rc_option( "--include-dir=" + p )
          }

        end

        # See description at MxxRu::Cpp::Toolset#mswin_res_file_name.
        def mswin_res_file_name( source_name )
          return source_name + ".res.o"
        end

        # See description at MxxRu::Cpp::Toolset#make_mswin_res_command_lines.
        def make_mswin_res_command_lines(
          res_name,
          rc_file,
          rc_options,
          target )

          return [ "#{rc_name} " +
            "#{rc_options.join(' ')} " +
            "-o #{res_name} #{rc_file}" ]
        end

        # See description at
        # MxxRu::Cpp::Toolsets::GccFamily#port_specific_dll_link_options.
        #
        # If mswin-res file exists than it added to linker options.
        def port_specific_dll_link_options(
          a_dll_name, a_dll_info, a_linker_lists, a_target )

          r = super( a_dll_name, a_dll_info, a_linker_lists, a_target )
          if a_linker_lists.resources.size
            r = a_linker_lists.resources.join( " " ) + " " + r
          end

          return r
        end

        # See description at
        # MxxRu::Cpp::Toolsets::GccFamily#port_specific_exe_link_options.
        #
        # If mswin-res file exists than it added to linker options.
        def port_specific_exe_link_options(
          a_exe_name, a_exe_info, a_linker_lists, a_target )

          r = super( a_exe_name, a_exe_info, a_linker_lists, a_target )
          if a_linker_lists.resources.size
            r = a_linker_lists.resources.join( " " ) + " " + r
          end

          return r
        end

      end # class GccMingw

      # For compatibility with previous versions.
      Gcc_mingw = GccMingw

    end # module Toolsets

  end # module Cpp

end # module MxxRu

MxxRu::Cpp::setup_toolset( MxxRu::Cpp::Toolsets::GccMingw.new )

