#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/binary_library'
require 'mxx_ru/cpp/toolsets/gcc_family'

module MxxRu
module Cpp
module Toolsets

# Base class for Unix GCC ports.
#
# Under Unix library name in '-l' linker option can be either static or shared.
# Therefore on Unix MxxRu must do special action to handle Target#lib_static
# and Target#lib_shared descriptions. This action implemented in this class.
class GccUnixFamily < GccFamily
  def initialize( name )
    super( name )
  end

  # This implementation adds -fPIC option for GccUnix compilers.
  # In 1.5.6 version -fPIC existed in GccFamily toolset's method but 
  # MinGW doesn't support that option.
  def setup_mandatory_options( target )
    super( target )
    target.compiler_option( "-fPIC" )
  end 

  # Special implementation for support explicit library type linker option.
  def make_linker_include_lib_options( target, libs )
    current_lib_type = nil
    all_libs = libs.inject( '' ) { |r, l|
        # Insert new lib mode switch only if next library has different
        # type than current type.
        if nil == current_lib_type || current_lib_type != l.type
          r << switch_to_default_lib_mode_if_needed( current_lib_type )
          current_lib_type = l.type
          if BinaryLibrary::ANY != l.type &&
              default_lib_linking_mode != l.type
            r << lib_linking_mode_switch( l.type )
          end
        end

        r << '-l' << port_specific_lib_name_checker( l.name ) << ' '
    }
    all_libs << switch_to_default_lib_mode_if_needed( current_lib_type )
    enclose_linker_include_lib_options_into_brackes( all_libs )
  end

  # Return default library linking mode for current platform.
  #
  # Default mode BinaryLibrary::SHARED for Unix with GCC assumed.
  # But default mode can be specified in toolset tag 'lib_linking_mode'.
  def default_lib_linking_mode
    t = tag( 'lib_linking_mode', 'unknown' )
    if 'unknown' != t
      'static' == t ? BinaryLibrary::STATIC : BinaryLibrary::SHARED
    else
      BinaryLibrary::SHARED
    end
  end

  # Return command line switch for forcing specified library type.
  def lib_linking_mode_switch( linking_mode )
    "-Wl,#{linking_mode == BinaryLibrary::SHARED ? '-Bdynamic' : '-Bstatic'} "
  end

  # Return command line switch for closing group of libraries with same type.
  # Empty string returned if current_lib_type is nil of is current_lib_type
  # is default lib linking mode on this platform.
  def switch_to_default_lib_mode_if_needed( current_lib_type )
    if nil != current_lib_type &&
        BinaryLibrary::ANY != current_lib_type &&
        default_lib_linking_mode != current_lib_type
      lib_linking_mode_switch( default_lib_linking_mode )
    else
      ''
    end
  end

end

end # module Toolsets
end # module Cpp
end # module MxxRu

