#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'ostruct'
require 'set'

require 'mxx_ru/cpp/target'
require 'mxx_ru/cpp/toolsets/vc_family'

module  MxxRu
  module  Cpp

    # Expand class Target to store source and actual manifest descriptions.
    class Target < MxxRu::BinaryTarget
      # Get source manifest description. If target has no manifest
      # then default manifest will be returned.
      def vc8_source_manifest
        if nil != @vc8_source_manifest
          @vc8_source_manifest
        else
          MxxRu::Cpp::Toolsets::Vc8Family::default_manifest
        end
      end

      # Setter for source manifest description.
      def vc8_source_manifest=( manifest )
        @vc8_source_manifest = manifest
      end

      attr_accessor :vc8_actual_manifest

    end

    module Toolsets

      # Toolset implementation for Visual C++ 2005 (8.0) and above.
      #
      class Vc8Family < MxxRu::Cpp::Toolsets::VcFamily

        # Tag name for a custom manifest tool name.
        #
        # Since v.1.4.0
        MT_NAME_TAG = 'mt_name'

        # Actual manifest description.
        class ActualManifest
          attr_reader :manifest_file
          attr_reader :digital_resource_id

          def initialize(
              source_manifest,
              manifest_file,
              digital_resource_id )
            @source_manifest = source_manifest
            @manifest_file = manifest_file
            @digital_resource_id = digital_resource_id
          end

          # Is autogeneration of manifest must be used?
          def autogen?
            @source_manifest.autogen
          end

          # Is manifest tool (mt.exe) must be used?
          def mt?
            @source_manifest.mt
          end

          # Is autogenerated manifest file should be kept after
          # embeding into executable?
          # By default autogenerated manifest file will be deleted.
          def keep_manifest_file?
            ( mt? and autogen? ) ? @source_manifest.keep_manifest_file : true
          end
        end # class ActualManifest

        # For compatibility with previous versions.
        Actual_manifest = ActualManifest

        @@default_manifest = OpenStruct.new(
          :autogen => true,
          :mt => true,
          :manifest_file => :default,
          :resource_id => :default,
          :keep_manifest_file => false )

        # Set of avaliable values for :resource_id key-value pair
        # in manifest description.
        @@resource_id_names = Set.new( [
          :default,
          :process_manifest,
          :isolationaware_manifest,
          :isolationaware_nostaticimport_manifest ] )

      public
        def initialize( a_name = "vc" )
          super( a_name )
        end

        # Check some VC8.0 constraints before calling superclass
        # implementation
        #
        # VC8.0 does not support Single-Thread Run-Time.
        def setup_mandatory_options( target )
          if THREADING_SINGLE == target.mxx_threading_mode
            raise MxxRu::UnsupportedModeEx.new(
              "Visual C++ 2005 and above doesn't support " +
              "single-threaded static RTL" )
          elsif THREADING_DEFAULT == target.mxx_threading_mode
            # Because VC8.0 doesn't support Single Thread Run-Time then
            # by default we must use Mutli-Threading Run-Time. 
            target.threading_mode( THREADING_MULTI )
          end

          super( target )

          setup_manifest( target )
          setup_manifest_params( target )
        end

        # Remove autogenerated manifest file if any.
        def clean_dll_specific_files( dll_file, dll_info, target )
          super( dll_file, dll_info, target )
          clean_autogenerated_manifest( target )
        end

        # Remove autogenerated manifest file if any.
        def clean_exe_specific_files( exe_file, exe_info, target )
          super( exe_file, exe_info, target )
          clean_autogenerated_manifest( target )
        end

        # Append command line for embeding manifest file to result
        # of super class implementation.
        def make_dll_command_lines(
            dll_name,
            dll_info,
            linker_lists,
            target )

          append_mt_command_lines(
            super( dll_name, dll_info, linker_lists, target ),
            dll_name,
            target )
        end

        # Append command line for embeding manifest file to result
        # of super class implementation.
        def make_exe_command_lines(
            exe_name,
            exe_info,
            linker_lists,
            target )
          append_mt_command_lines(
            super( exe_name, exe_info, linker_lists, target ),
            exe_name,
            target )
        end

        # Make manifest description.
        #
        # Examples:
        #  # Autogenerate manifest for target. Do not embed into executable.
        #  # Manifest stored in <target>.manifest file.
        #  MxxRu::Cpp::Toolsets::Vc8::manifest(
        #    :target => self,
        #    :autogen => :to_default_file )
        # 
        #  # Autogenerate manifest for target. Embed into executable.
        #  # Autogenerated manifest will be deleted after embeding.
        #  MxxRu::Cpp::Toolsets::Vc8::manifest(
        #    :target => self,
        #    :autogen => :to_default_file,
        #    :mt => {} )
        # 
        #  # Autogenerate manifest for target. Embed into executable.
        #  # Autogenerated manifest will be kept after embeding.
        #  MxxRu::Cpp::Toolsets::Vc8::manifest(
        #    :target => self,
        #    :autogen => :to_default_file,
        #    :mt => { :keep_manifest_file => true } )
        # 
        #  # Autogenerate manifest for target. Embed into executable as
        #  # resource with id CREATEPROCESS_MANIFEST_RESOURCE_ID id.
        #  MxxRu::Cpp::Toolsets::Vc8::manifest(
        #    :target => self,
        #    :autogen => :to_default_file,
        #    :mt => { :resource_id => :process_manifest } )
        # 
        #  # Autogenerate manifest for target. Store manifest into
        #  # file 'hello_world.exe.manifest'. Embed info executable.
        #  MxxRu::Cpp::Toolsets::Vc8::manifest(
        #    :target => self,
        #    :autogen => 'hello_world.exe.manifest',
        #    :mt => {} )
        # 
        #  # Embed manifest from file 'hello_world.exe.manifest' into 
        #  # target executable.
        #  MxxRu::Cpp::Toolsets::Vc8::manifest(
        #    :target => self,
        #    :mt => { :manifest => 'hello_world.exe.manifest' } )
        #  # Embed manifest from file 'hello_world.dll.manifest' into 
        #  # target executable as resource with
        #  # ISOLATIONAWARE_NOSTATICIMPORT_MANIFEST_RESOURCE_ID id.
        #  MxxRu::Cpp::Toolsets::Vc8::manifest(
        #    :target => self,
        #    :mt => { :manifest => 'hello_world.exe.manifest',
        #      :resource_id => :isolationaware_nonstaticimport } )
        # 
        #  # Autogenerate manifest for all targets in build. Store manifest
        #  # in default files. Do not embed manifest into executable.
        #  MxxRu::Cpp::Toolsets::Vc8::manifest(
        #    :autogen => :to_default_file )
        #    
        #  # Autogenerate manifest for all targets in build. Store manifest
        #  # in default files. Embed manifest into executable.
        #  MxxRu::Cpp::Toolsets::Vc8::manifest(
        #    :autogen => :to_default_file,
        #    :mt => {} )
        # 
        #  # Drop default manifest for build. Only explicitly specified
        #  # manifest will be generated.
        #  MxxRu::Cpp::Toolsets::Vc8::manifest( nil )
        #
        def self.manifest( params )
          if nil == params
            # No default manifest.
            @@default_manifest = nil
          else
            check_manifest_params( params )

            d = OpenStruct.new( :keep_manifest_file => true )
            
            if params.key?( :autogen )
              d.autogen = true
              d.manifest_file = ( params[ :autogen ] == :to_default_file ?
                :default : params[ :autogen ] )
              d.keep_manifest_file = false
            else
              d.autogen = false
            end

            if params.key?( :mt ) && nil != params[ :mt ]
              d.mt = true
              mt_params = params[ :mt ]
              d.manifest_file = mt_params[ :manifest ] if
                mt_params.key?( :manifest )
              d.resource_id = ( mt_params.key?( :resource_id ) ?
                mt_params[ :resource_id ] : :default )

              if mt_params.key?( :keep_manifest_file ) and
                params.key?( :autogen )
                d.keep_manifest_file = mt_params[ :keep_manifest_file ]
              end
            else
              d.mt = false
            end

            if params.key?( :target )
              # Source manifest must be set for specified target.
              params[ :target ].vc8_source_manifest = d
            else
              # This is new default manifest.
              @@default_manifest = d
            end
          end
        end

        # Get default manifest.
        def self.default_manifest
          @@default_manifest
        end

      protected
        # Set actual manifest for target if source manifest defined
        # (or default manifest exists).
        def setup_manifest( target )
          # Manifest can be set only for EXE or DLL.
          if ExeTargetType::TYPE == target.target_type.name or
            DllTargetType::TYPE == target.target_type.name
            desc = target.vc8_source_manifest

            if nil != desc
              # Actual manifest must be created for target.
              manifest_file =
                if :default == desc.manifest_file
                  # We must create name for manifest.
                  if ExeTargetType::TYPE == target.target_type.name
                    full_exe_name( target )
                  else
                    full_dll_name( target )
                  end + ".manifest"
                else
                  # Manifest filename already specified.
                  desc.manifest_file
                end

              # If mt.exe will be used then we must determine
              # digital resource_id for manifest resource.
              digital_resource_id =
                if desc.mt
                  case desc.resource_id
                    when :default
                      # Resource ID must be caclulated based on
                      # target type.
                      ExeTargetType::TYPE == target.target_type.name ? 1 : 2
                    when :process_manifest
                      1
                    when :isolationaware_manifest
                      2
                    else
                      # :isolationaware_nostaticimport_manifest
                      3
                  end
                else
                  nil
                end

              target.vc8_actual_manifest = ActualManifest.new(
                  desc,
                  manifest_file,
                  digital_resource_id )
            end
          end
        end

        # Setup linker params for processing target manifest.
        def setup_manifest_params( target )
          m = target.vc8_actual_manifest
          if nil != m and m.autogen?
            # Linker must generate manifest.
            target.linker_option( '/MANIFEST' )
            target.linker_option( '/MANIFESTFILE:' + m.manifest_file )
          else
            # Linker should not generate manifest.
            target.linker_option( '/MANIFEST:NO' )
          end
        end

        # Remove autogenerated manifest file if any.
        def clean_autogenerated_manifest( target )
          m = target.vc8_actual_manifest
          MxxRu::Util::delete_file( m.manifest_file ) if
            m != nil and m.autogen?
        end

        # Append command line for embeding manifest file to end of
        # specified command line list.
        def append_mt_command_lines(
            cmd_lines,
            target_name,
            target )

          m = target.vc8_actual_manifest
          if nil != m and m.mt?
            cmd_lines << "#{mt_name} /nologo " +
              "/outputresource:\"#{target_name};\#" +
                  "#{m.digital_resource_id}\" " +
              "/manifest \"#{m.manifest_file}\""

            if !m.keep_manifest_file?
              # del utility doesn't work with unix slashes
              cmd_lines << "del \"#{MxxRu::Util::native_pathname(m.manifest_file)}\""
            end
          end

          cmd_lines
        end

        # Name of manifest tool executable
        def mt_name
          tag( MT_NAME_TAG, "mt" )
        end

        # Checking manifest params validity.
        def self.check_manifest_params( params )
          if nil != params
            if params.key?( :mt )
              mt_hash = params[ :mt ]

              # :resource_id must have one of avaliable values.
              if mt_hash.key?( :resource_id ) and
                  !@@resource_id_names.member?( mt_hash[ :resource_id ] )
                raise InvalidValueEx.new(
                  "invalid value for :resource_id key: " +
                    "'#{mt_hash[:resource_id]}'" ) 
              end

              if !mt_hash.key?( :manifest ) and
                !params.key?( :autogen )

                raise InvalidValueEx.new(
                  "no manifest file name specified" )
              end

              # :keep_manifest_file should be Bool.
              if mt_hash.key?( :keep_manifest_file ) and
                true != mt_hash[ :keep_manifest_file ] and
                false != mt_hash[ :keep_manifest_file ]
                raise InvalidValueEx.new(
                  "value for :keep_manifest_file should be 'true' " +
                  "or 'false'" )
              end
            end

            if params.key?( :autogen )
              v = params[ :autogen ]
              if !( v == :to_default_file or
                v.kind_of?( String ) )
                raise InvalidValueEx.new(
                  "invalid value for :autogen (must be " +
                  ":to_default_file or String instance)" )
              end
            end
          end
        end

      end # class Vc8Family

    end # module Toolsets

  end # module Cpp

end # module MxxRu

