#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

module  MxxRu

  # Base class for all exceptions.
  class Ex < Exception
  end

  # Exception, thrown on execution of 'pure abstract method'.
  class AbstractMethodEx < Ex
    def initialize( method_name )
      super( "Pure virtual method call: " + method_name )
    end
  end

  # For compatibility with previous versions.
  Abstract_method_ex = AbstractMethodEx

  # Exception, thrown if returned object type is not expected.
  class InvalidCastEx < Ex
    def initialize( msg )
      super( msg )
    end
  end

  # For compatibility with previous versions.
  Invalid_cast_ex = InvalidCastEx

  # Exception, thrown if build commands fail.
  class BuildEx < Ex
    def initialize( cmd_line, desc )
      super( "Build error: '#{cmd_line}' returns '#{desc}'" )
    end
  end

  # For compatibility with previous versions.
  Build_ex = BuildEx

  # Exception, thrown if unknown target type is detected.
  class UnsupportedTargetTypeEx < Ex
    def initialize( target_class, target_type )
      super( "Unsupported target type '#{target_type}'" +
        " in '#{target_class}'" )
    end
  end

  # For compatibility with previous versions.
  Unsupported_target_type_ex = UnsupportedTargetTypeEx 

  # Exception, thrown if no target object is found for given alias.
  class TargetUnknownEx < Ex
    def initialize( target_name )
      super( "Target unknown for: " + target_name )
    end
  end

  # For compatibility with previous versions.
  Target_unknown_ex = TargetUnknownEx

  # Exception, thrown if value conflict for global options such as 
  # runtime_mode, rtl_mode, etc is detected
  class GlobalOptionConflictEx < Ex
    # [_a_name_] Option name.
    # [_a_current_] Current option value.
    # [_a_current_owner_] Who was set current value.
    # [_a_new_] New option value.
    # [_a_new_owner_] Who tries to set current value.
    def initialize(
      a_name,
      a_current,
      a_current_owner,
      a_new,
      a_new_owner )

      super( "Global option '#{a_name}' conflict detected! " +
        "Current value '#{a_current}' set by '#{a_current_owner}'. " +
        "New value '#{a_new}' required by '#{a_new_owner}'" )
    end
  end

  # For compatibility with previous versions.
  Global_option_conflict_ex = GlobalOptionConflictEx

  # Exception, thrown if anyone tries to set invalid value to global options
  # such as runtime_mode, rtl_mode, etc.
  class GlobalOptionDisallowedValueEx < Ex
    # [_a_name_] Option name.
    # [_a_available_values_] Available values list. Should be Array of String.
    # [_a_new_] New option value.
    # [_a_new_owner_] Who tries to set new value.
    def initialize(
      a_name,
      a_available_values,
      a_new,
      a_new_owner )

      super( "Value '#{new}' required by '#{a_new_owner}' " +
        "disallowed for global option '#{a_name}'! " +
        "Allowed values are: #{a_available_values.join(' ')}" )
    end
  end

  # For compatibility with previous versions.
  Global_option_disallowed_value_ex = GlobalOptionDisallowedValueEx

  # Exception, thrown if some mode is not supported by the toolset.
  class UnsupportedModeEx < Ex
    def initialize( msg )
      super( msg )
    end
  end

  # For compatibility with previous versions.
  Unsupported_mode_ex = UnsupportedModeEx

  # Exception, thrown if some dependency file is missing.
  class MissingDependFileEx < Ex
    def initialize( missing_file, depend_for )
      super( "File '#{missing_file}' required by '#{depend_for}' " +
        "missing" )
    end
  end

  # For compatibility with previous versions.
  Missing_depend_file_ex = MissingDependFileEx

  # Exception, thrown if some value is invalid in current context.
  class InvalidValueEx < Ex
    def initialize( msg )
      super( msg )
    end
  end

  # For compatibility with previous versions.
  Invalid_value_ex = InvalidValueEx

end

load File.dirname( __FILE__ ) + '/compat.rb'
