#--
# Copyright (c) 2006-2016, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'rake'
require 'rake/tasklib'

require 'uri'
require 'singleton'

require 'digest'

require 'json'

require 'mxx_ru/helpers'

module MxxRu

module Externals

EXTERNALS_STORAGE_DIR = '.externals'

class Registry
  include Singleton

  INITIAL_MODE = 0
  NEW_RULES_MODE = 1
  OLD_RULES_MODE = 2

  def initialize
    @mode = INITIAL_MODE
    @externals = {}
  end

  def turn_new_rules_on
    @mode = NEW_RULES_MODE
  end

  def turn_old_rules_on
    @mode = OLD_RULES_MODE
  end

  def Registry.handle_external(*args)
    Registry.instance.handle_external(*args)
  end

  def handle_external(name, obj)
    if INITIAL_MODE == @mode
      raise "Use mxxruexternals command for handling externals.\n" +
          "For example:\n\n" +
          "mxxruexternals -f my_externals.rb\n"
    elsif NEW_RULES_MODE == @mode
      @externals[ name ] = obj.make_hash
      obj.define_rules :new_rules
    elsif OLD_RULES_MODE == @mode
      fresh_version = @externals.fetch(name, nil)
      if fresh_version
        puts "[Note] #{name} previous rules found"
        old_hash = obj.make_hash
        if old_hash == fresh_version
          puts "[Note] #{name} previous rules the same with current"
        else
          puts "[Note] #{name} previous rules differ with current"
          puts "[Note] #{name} WILL BE REGET"
          obj.define_rules :old_rules
        end
      else
        puts "[Note] #{name} not present in the current version"
        puts "[Note] #{name} WILL BE REMOVED"
        obj.define_rules :old_rules
      end
    end
  end
end

module Impl

module OptionsHolder
  def option(*values)
    make_options.push(*values)
  end

protected

  def push_options_to(receiver)
    receiver.push(*@options)
  end

private

  def make_options
    @options = [] unless @options
    @options
  end

  def options_to_map
    { options: @options }
  end
end # module OptionsHolder

class CurlDownloder
  include OptionsHolder

  def CurlDownloder.check_presence
    MxxRu::Helpers.external_tool_version_probe('curl --version', /^curl\s(\S+)\s/)
  end

  def CurlDownloder.downloader_id
    :curl
  end

  def initialize(options)
    @options = options
  end

  def make_download_sh_args(uri, result_name)
    push_options_to(['curl', '-L']).push('-o', result_name, uri)
  end
end

class WgetDownloder
  include OptionsHolder

  def WgetDownloder.check_presence
    MxxRu::Helpers.external_tool_version_probe('wget --version', /^GNU Wget\s(\S+)\s/)
  end

  def WgetDownloder.downloader_id
    :wget
  end

  def initialize(options)
    @options = options
  end

  def make_download_sh_args(uri, result_name)
    push_options_to(['wget']).push('-O', result_name, uri)
  end
end

module WebDownloaderOptions
  def downloader_option(downloader_id, *values)
    downloader_option_storage()[downloader_id].push(*values)
  end

  def downloader_options_for(downloader_id)
    downloader_option_storage()[downloader_id]
  end

  def all_downloader_options
    downloader_option_storage
  end

private
  def downloader_option_storage
    @downloader_options = Hash.new{|h,k| h[k] = []} unless @downloader_options
    @downloader_options
  end
end

class WebDownloader
  include Singleton
  SUPPORTED_DOWNLOADERS = [WgetDownloder, CurlDownloder]
  
  @@preffered_downloaders = []
  @@downloader_options = Hash.new{|h,k| h[k] = []}

  def initialize
    @downloader_type, @downloader_version = *detect_downloader_type

    if Rake.verbose
      puts "#{@downloader_type.downloader_id} v.#{@downloader_version} found"
    end
  end

  def make_downloader(options_holder)
    id = @downloader_type.downloader_id
    @downloader_type.new(
      merge_options(
        options_holder.downloader_options_for(id),
        @@downloader_options[id]))
  end

  def WebDownloader.preffered_downloader(downloader_id)
    found = SUPPORTED_DOWNLOADERS.find do |klass|
      downloader_id == klass.downloader_id ? klass : nil
    end
    raise "Unsupported downloader: #{downloader_id}" unless found

    @@preffered_downloaders.push(found)
  end

  def WebDownloader.downloader_option(downloader_id, *values)
    @@downloader_options[downloader_id].push(*values)
  end

private
  def detect_downloader_type
    r = search_downloader_in(@@preffered_downloaders)
    r = search_downloader_in(SUPPORTED_DOWNLOADERS) unless r
    raise "No web downloader found (curl, wget)" unless r
    r
  end

  def search_downloader_in(sequence)
    version = nil
    downloader = sequence.find {|klass| version = klass.check_presence}
    downloader ? [downloader, version] : nil
  end

  def merge_options(head, tail)
    r = head.dup
    r.push(*tail)
  end
end # class WebDownloader

module ExternalBasics

  attr_reader :url

  attr_reader :paths

  attr_reader :name

  def url(value)
    @url = value
  end

  alias :url= :url

  def map_dir(value)
    raise "map_dir() value must be a Hash" unless value.is_a?(Hash)

    @paths.merge!(value)
  end

  def map(value)
    STDERR.print("#{caller(1)[0]}: 'map_dir' must be used instead of 'map'," +
      " 'map' is deprecated\n")

    map_dir(value)
  end

  def map_file(value)
    raise "map_file() value must be a Hash" unless value.is_a?(Hash)

    @files.merge!(value)
  end

private

  def defaults(name)
    @name = name
    @install_task_name = "#{@name}_#{object_id}"
    @remove_task_name = "#{@install_task_name}:remove"
    @reget_task_name = "#{@install_task_name}:reget"
    @paths = {}
    @files = {}
    @verbose = Rake.verbose
  end

  def basics_to_map
    { url: @url, paths: @paths, files: @files }
  end

  def fileop_options
    { :verbose => @verbose }
  end

  def rm_dir_if_exists(dir_name)
    rm_r(dir_name, fileop_options) if Dir.exists?(dir_name)
  end

  def rm_directories(dirs)
    dirs.each { |d| rm_dir_if_exists(d) }
  end

  def rm_file_if_exists(file_name)
    rm(file_name, fileop_options) if File.exists?(file_name)
  end

  def rm_files(files)
    files.each { |f| rm_file_if_exists(f) }
  end

  def ensure_basics_correctness(name)
    raise "#{name}: url not specified" unless @url
    raise "#{name}: path or file map must be specified" if @paths.empty? and @files.empty?
  end

  def make_name(prefix, suffix)
    prefix.to_s + suffix
  end

  def subdir_name(subdir)
    File.join(EXTERNALS_STORAGE_DIR, subdir)
  end

  def make_tmp_name(*prefixes)
    "#{prefixes.join('.')}.#{Process.pid}.#{Thread.current.object_id}"
  end

  # Note: argument old_or_new can receive values :new_rules or :old_rules.
  def define(old_or_new, &exporter)
    ensure_basics_correctness(@name)

    ext_prj_sources_dir, dir_subtargets, file_subtargets =
        define_basic_task(old_or_new, &exporter)

    define_remove_task(
        old_or_new,
        ext_prj_sources_dir,
        dir_subtargets,
        file_subtargets)

    if :new_rules == old_or_new
      define_reget_task(ext_prj_sources_dir, dir_subtargets, file_subtargets)
    end
  end

  def define_basic_task(old_or_new, &exporter)
    ext_prj_sources_dir = subdir_name(@name.to_s)

    directory EXTERNALS_STORAGE_DIR

    dir_subtargets = []
    @paths.each do |src, dst|
      src, actual_dst, dir_dep = handle_dir_map_pair(src, dst)

      define_only_if_new_rules(old_or_new) do
        directory dir_dep

        directory actual_dst => [ext_prj_sources_dir, dir_dep] do
          cp_r(File.join(ext_prj_sources_dir, src), actual_dst, fileop_options)
        end
      end

      dir_subtargets << actual_dst
    end

    file_subtargets = []
    @files.each do |src, dst|
      dst, dir_dep = handle_file_map_pair(src, dst)

      define_only_if_new_rules(old_or_new) do
        directory dir_dep

        file dst => [ext_prj_sources_dir, dir_dep] do
          cp(File.join(ext_prj_sources_dir, src), dst, fileop_options)
        end
      end

      file_subtargets << dst
    end

    define_only_if_new_rules(old_or_new) do
      directory ext_prj_sources_dir => [EXTERNALS_STORAGE_DIR] do
        # All old destinations must be removed because source code
        # will be exported again.
        rm_directories(dir_subtargets)
        rm_files(file_subtargets)

        ext_prj_sources_tmp_dir = subdir_name(make_tmp_name(@name))
        exporter.call(ext_prj_sources_tmp_dir)
        mv(ext_prj_sources_tmp_dir, ext_prj_sources_dir, fileop_options)
      end

      task @install_task_name =>
        [:mxxru_cleanup_previous, dir_subtargets, file_subtargets].flatten!

      task install: @install_task_name
    end

    return ext_prj_sources_dir, dir_subtargets, file_subtargets
  end

  def define_remove_task(
    old_or_new, ext_prj_sources_dir, dir_subtargets, file_subtargets)

    task @remove_task_name do
      rm_files(file_subtargets)
      rm_directories(dir_subtargets)
      rm_dir_if_exists(ext_prj_sources_dir)
    end

    if :new_rules == old_or_new
      task :remove => @remove_task_name
    else
      task :mxxru_cleanup_previous => @remove_task_name
    end
  end

  def define_reget_task(ext_prj_sources_dir, dir_subtargets, file_subtargets)
    task @reget_task_name => [@remove_task_name, @install_task_name]

    task :reget => @reget_task_name
  end

  # Check the presence of start in src and make actual_dst and dir_dep
  # in dependence of the result. If start is present then it is removed
  # from actual_src.
  #
  # Returns triplet: [actual_src, actual_dst, dir_dep]
  #
  def handle_dir_map_pair(src, dst)
    src_parts = File.split(src)
    if '*' == src_parts.last
      # Since v.1.6.9: source path must be copied with different name.
      # Source name must not contain '*' at the end.
      src = src_parts.first
      # The first part original dst must be used as destination.
      [src, dst, File.split(dst).first]
    else
      # Last name of src must be added to dst to create actual destination.
      [src, File.join(dst, src_parts.last), dst]
    end
  end

  # Check the presence of start in dst and make actual_dst and dir_dep
  # in dependence of the result.
  #
  # Returns pair: [actual_dst, dir_dep]
  #
  def handle_file_map_pair(src, dst)
    dst_parts = File.split(dst)
    if '*' == dst_parts.last
      # Since v.1.6.9: original name of file must be preserved.
      dst = File.join(dst_parts.first, File.split(src).last)
    end
    [dst, dst_parts.first]
  end

  # Helper function for doing some actions only if old_or_new is :new_rules.
  def define_only_if_new_rules(old_or_new, &block)
    if :new_rules == old_or_new
      block.call
    end
  end
end

end # module Impl

class Svn < Rake::TaskLib
  include Impl::ExternalBasics
  include Impl::OptionsHolder

  def initialize(name)
    defaults(name)
    @with_externals = false

    yield self if block_given?

    Registry::handle_external(@name, self)
  end

  attr_reader :rev
  def rev(v); @rev = v; end
  alias :rev= :rev

  def with_externals
    @with_externals = true
  end

  def make_hash
    options_to_map.merge!( basics_to_map ).merge!(
      { :rev => @rev, :with_externals => @with_externals } )
  end

  def define_rules(old_or_new)
    define(old_or_new) do |tmp_dir|
      sh_args = [ 'svn', 'export' ]

      sh_args << '-r' << @rev.to_s if @rev
      sh_args << '--ignore-externals' unless @with_externals
      push_options_to(sh_args)

      sh_args << @url << tmp_dir

      sh *sh_args
    end
  end

end

class Git < Rake::TaskLib
  include Impl::ExternalBasics
  include Impl::OptionsHolder

  def initialize(name)
    defaults(name)
    @unlimited_depth = false
    @recursive = false

    yield self if block_given?

    raise "#{name}: tag and commit cannot be specified together" if @tag and @commit

    Registry::handle_external(@name, self)
  end

  attr_reader :tag
  def tag(v); @tag = v; end
  alias :tag= :tag

  def unlimited_depth
    @unlimited_depth = true
  end
  
  def recursive
    @recursive = true
  end

  attr_reader :commit
  def commit(v); @commit = v; end
  alias :commit= :commit

  def make_hash
    options_to_map.merge!( basics_to_map ).merge!(
      { :recursive => @recursive,
        :unlimited_depth => @unlimited_depth,
        :tag => @tag,
        :commit => @commit
      } )
  end

  def define_rules(old_or_new)
    define(old_or_new) do |tmp_dir|
      if @commit.nil?
        # No specific commit.
        # Simple clone (probably with --depth 1) is enough.
        sh_args = ['git', 'clone']

        sh_args << '--branch' << @tag.to_s if @tag
        sh_args << '--depth' << '1' unless @unlimited_depth
        sh_args << '--recursive' if @recursive
        push_options_to(sh_args)

        sh_args << @url << tmp_dir

        sh *sh_args
      else
        # To extract specific commit is necessary to do
        # at least two commands: clone (without limitiation of depth) and
        # checkout for specific commit.

        # Do `clone repo tmp_dir` 
        sh *(push_options_to(['git', 'clone']).push(@url).push(tmp_dir))
        # Do `checkout commit` inside tmp_dir
        cd tmp_dir do
          sh 'git', 'checkout', @commit
        end
      end
    end
  end

end

class Hg < Rake::TaskLib
  include Impl::ExternalBasics
  include Impl::OptionsHolder

  def initialize(name)
    defaults(name)

    yield self if block_given?

    Registry::handle_external(@name, self)
  end

  attr_reader :tag
  def tag(v); @tag = v; end
  alias :tag= :tag

  def make_hash
    options_to_map.merge!( basics_to_map ).merge!( { :tag => @tag } )
  end

  def define_rules(old_or_new)
    define(old_or_new) do |tmp_dir|
      sh_args = ['hg', 'clone']

      sh_args << '--updaterev' << @tag.to_s if @tag
      push_options_to(sh_args)

      sh_args << @url << tmp_dir

      sh *sh_args
    end
  end

end # class Hg

# Implementation of externals represented as downloadable
# archive (like tar.gz, zip, rar or 7z).
#
class ArchiveAsExternals < Rake::TaskLib
  include Impl::ExternalBasics
  include Impl::WebDownloaderOptions

  @@archive_handlers = {
    :tar => proc {
        ['tar', 'x'].push(*@unpacker_options).push('-f').push(archive_name)
      },
    :zip => proc {
        ['unzip'].push(*@unpacker_options).push(archive_name)
      },
    :sevenzip => proc {
        ['7z', 'x'].push(*@unpacker_options).push(archive_name)
      }
  }

  @@archive_extensions = {
    '.tar' => :tar,
    '.tar.gz' => :tar,
    '.tgz' => :tar,
    '.taz' => :tar,
    '.tar.Z' => :tar,
    '.taZ' => :tar,
    '.tar.bz2' => :tar,
    '.tz2' => :tar,
    '.tbz2' => :tar,
    '.tar.lz' => :tar,
    '.tar.lzma' => :tar,
    '.tlz' => :tar,
    '.tar.lzo' => :tar,
    '.tar.xz' => :tar,
    '.zip' => :zip,
    '.7z' => :sevenzip
  }

  def initialize(name)
    defaults(name)
    @unpacker_options = []

    yield self if block_given?

    parse_uri
    detect_archive_type

    Registry::handle_external(@name, self)
  end

  def md5(v)
    @md5 = v
  end

  def sha1(v)
    @sha1 = v
  end

  def sha256(v)
    @sha256 = v
  end

  def sha512(v)
    @sha512 = v
  end

  def unpacker_option(v)
    @unpacker_options.push(v)
  end

  def make_hash
    all_downloader_options.merge( basics_to_map ).merge!(
      { :md5 => @md5,
        :sha1 => @sha1,
        :sha256 => @sha256,
        :sha512 => @sha512,
        :unpacker_options => @unpacker_options
      } )
  end

  def define_rules(old_or_new)
    define(old_or_new) do |tmp_dir|
      # Temporary directory must be removed in a case of any error.
      successful = false
      begin
        mkdir(tmp_dir, fileop_options) if !Dir.exists?(tmp_dir)

        cd tmp_dir do
          download_archive
          unpack_archive
        end

        successful = true
      ensure
        rm_dir_if_exists(tmp_dir) unless successful
      end
    end
  end

private
  def parse_uri
    @parsed_uri = URI::split(@url)
    @archive_name = File.split(uri_path).last
  end

  def uri_scheme
    @parsed_uri[0]
  end

  def uri_path
    @parsed_uri[5] # See doc for URI::split for description
                   # of items in result vector.
  end

  attr_reader :archive_name

  def detect_archive_type
    ext = @@archive_extensions.keys.find {|k| @archive_name.end_with?(k)}
    raise "#{name} unable to detect archive type for " +
        "#{@archive_name}" unless ext

    @archive_type = @@archive_extensions[ext]
  end

  def download_archive
    if 'file' == uri_scheme
      download_archive_via_cp
    else
      download_archive_from_web
    end
    check_digest_if_necessary(:MD5, @md5)
    check_digest_if_necessary(:SHA1, @sha1)
    check_digest_if_necessary(:SHA256, @sha256)
    check_digest_if_necessary(:SHA512, @sha512)
  end

  def download_archive_from_web
    web_downloader = Impl::WebDownloader.instance.make_downloader(self)
    sh *(web_downloader.make_download_sh_args(@url, archive_name))
  end

  def download_archive_via_cp
    # On Windows path can look like '/d:/...'.
    # The leading slash must be removed.
    path = uri_path.gsub(/^(\/)(\w:)(.+)$/, '\2\3')
    cp(path, archive_name, fileop_options)
  end

  def unpack_archive
    sh(*make_unpacker_args)
    rm(archive_name, fileop_options)
    names = Dir['*']
    if 1 == names.size and Dir.exists?(names.first)
      # This is the only folder in the distributive.
      # Move all its contents one level up.
      cd names.first do
        # Files like '.gitignore' won't be found without FNM_DOTMATCH.
        # But in that case '.' and '..' will be found too.
        Dir.glob('*', File::FNM_DOTMATCH).each do |n|
          mv(n, '..', fileop_options) if n != '.' && n != '..'
        end
      end
      # This single directory no more needed.
      rm_r(names.first, fileop_options)
    end
  end

  def check_digest_if_necessary(digest_id, expected)
    return unless expected

    digest_klass = Digest(digest_id)
    puts "#{archive_name} checking #{digest_id} checksum..." if @verbose
    chsum = digest_klass.file(archive_name).hexdigest
    if chsum != expected
      raise "#{archive_name}: #{digest_id} checksum missmatch. " +
        "actual: #{chsum}, expected: #{expected}"
    end
  end

  def make_unpacker_args
    handler = @@archive_handlers[@archive_type]
    self.instance_eval &handler
  end
end # class ArchiveAsExternals

def Externals::preffered_downloader(downloader_id)
  Impl::WebDownloader.preffered_downloader(downloader_id)
end

def Externals::downloader_option(downloader_id, *values)
  Impl::WebDownloader.downloader_option(downloader_id, *values)
end

end # module Externals

def MxxRu.svn_externals(name, &block)
  Externals::Svn.new(name, &block)
end

def MxxRu.git_externals(name, &block)
  Externals::Git.new(name, &block)
end

def MxxRu.hg_externals(name, &block)
  Externals::Hg.new(name, &block)
end

def MxxRu.arch_externals(name, &block)
  Externals::ArchiveAsExternals.new(name, &block)
end

end # module MxxRu

desc "Remove all external projects"
task :remove do
  d = MxxRu::Externals::EXTERNALS_STORAGE_DIR
  rm_r(d, :verbose => Rake.verbose) if Dir.exists?(d)
end

desc "Remove and get again all external projects"
task :reget

local_rc_file = File.join(Dir.getwd, '.externalsrc')
if File.exists?(local_rc_file)
  load local_rc_file
end

# vim:ts=2:sts=2:sw=2:expandtab

