#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2016, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

# Since v.1.6.11
#
# Implementation of stabs generation for MxxRu::externals file

require 'erb'
require 'optparse'

require 'mxx_ru/generators/impl/std_receiver'

module MxxRu

module Generators

module Externals

# Class for storing command-line arguments as options.
#
# Usage:
#   options = Options.parse( args, banner )
#
class Options
  # Name of output file (-o, --output-file). nil if missing.
  attr_accessor :output_file

  # Parsing command-line arguments and returning Options instance.
  #
  # Calls exit(1) if --help present in _args_.
  #
  def Options.parse( args, banner )
    parser = OptionParser.new

    result = Options.new

    parser.banner = banner

    parser.on( '-o', '--output-file FILE', 'Output file name' ) do |p|
      result.output_file = p
    end

    parser.on_tail( '-h', '--help', 'Show this message' ) do
      puts parser
      exit(1)
    end

    parser.parse!( args )
    result
  end
end # class Options

# Class to be used in ERb template generation.
#
# Usage:
#   template_params = TemplateParams.new( target_type, options )
#   template = ERb.new( IO.read( some_template_file ) )
#   result = template.generate( template.get_binding )
#
class TemplateParams
  # For a case when output_file is undetectable.
  DEFAULT_FILE = 'externals.rb'

  # Name of externals file
  attr_reader :externals_file

  def initialize( options )
    @externals_file = try_detect_externals_file( options )
  end

  # Returns binding to use in ERb generation.
  def get_binding
    binding
  end

private
  # Try to setup name of externals file.
  #
  # If --output-file specified then its value used.
  # Otherwise value 'externals.rb' is used.
  #
  def try_detect_externals_file( options )
    options.output_file ?
        File.basename( options.output_file ) :
        DEFAULT_FILE
  end

end # class TemplateParams

# Main class for code generation of binary unit test projects.
#
# Usage:
#   receiver = StdReceiver.new
#   generator = Generator.new( args, receiver )
#   generator.run
#
class Generator
  def initialize( args, receiver )
    @args = args
    @receiver = receiver
  end

  def run
    options = Options.parse( @args,
        "Stubs for C++ build root file generator\n\n" +
        "Usage:\n" +
        "mxxrugen [<mxxrugen-options>] cpp-build-root [<options>]\n\n" )
    result = do_generation( options )
    @receiver.receive( result, options.output_file )
  end

private
  # Performs main generation actions.
  #
  # Returns generation result as String.
  def do_generation( options )
    template = IO.read( File.join( File.dirname( __FILE__ ), 'template.erb' ) )
    generator = ERB.new( template )

    params = TemplateParams.new( options )
    generator.result( params.get_binding ).gsub( /\n\n\n+/, "\n\n" )
  end
end

# Helper method for generation.
#
# Usage:
#   # For using StdReceiver.
#   generate_for( ARGV )
#
#   # For using custom Receiver.
#   generate_for( ARGV, CustomReceiver.new )
#
def Externals.generate_for( args, receiver = nil )
  generator = Generator.new( args,
      receiver ? receiver : MxxRu::Generators::Impl::StdReceiver.new )
  generator.run
end

end # module Externals

end # module Generators

end # module MxxRu

MxxRu::Generators::Externals::generate_for( ARGV )

