#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

# Since v.1.4.0
#
# Implementation of project's stabs generation for C++ projects.

require 'erb'
require 'optparse'

require 'mxx_ru/generators/impl/std_receiver'

module MxxRu

module Generators

module Impl

module Cpp

# Types of C/C++ projects.

# Project type is static library.
LIB = :lib
# Project type is library collection.
LIB_COLLECTION = :lib_collection
# Project type is dinamic library.
DLL = :dll
# Project type is executable.
EXE = :exe
# Project type is composite.
COMPOSITE = :composite

# Class for storing command-line arguments as options.
#
# Usage:
#   options = Options.parse( args, banner, { :implib_path => false } )
#
class Options
  # Name of target (-t, --target).
  attr_accessor :target_name
  # Project path (name of project directory in cmd-line).
  attr_accessor :project_path
  # Name of import library path (--implib-path).
  attr_accessor :implib_path
  # Name of output file (-o, --output-file). nil if missing.
  attr_accessor :output_file

  # Parsing command-line arguments and returning Options instance.
  #
  # These keys are supported in _params_:
  # [_implib_path_] true or false. Enables/disables argument --implib-path.
  #
  # Calls exit(1) if --help present in _args_.
  #
  def Options.parse( args, banner, params )
    parser = OptionParser.new

    result = Options.new

    parser.banner = banner

    parser.on( '-t', '--target NAME', 'Target name' ) do |p|
      result.target_name = p
    end

    parser.on( '-o', '--output-file FILE', 'Output file name' ) do |p|
      result.output_file = p
    end

    if true == params.fetch( :implib_path, false ) 
      parser.on( '--implib-path PATH', 'Import library path name' ) do |p|
        result.implib_path = p
      end
    end

    parser.on_tail( '-h', '--help', 'Show this message' ) do
      puts parser
      exit(1)
    end

    parser.order!( args ) do |noarg|
      # Any non-options is considered as project path name.
      result.project_path = noarg
    end

    result
  end
end # class Options

# Class to be used in ERb template generation.
#
# Usage:
#   template_params = TemplateParams.new( target_type, options )
#   template = ERb.new( IO.read( some_template_file ) )
#   result = template.generate( template.get_binding )
#
class TemplateParams
  # Dependencies between target_type and setup_target function name.
  @@setup_target_functions = {
      LIB => 'lib_target',
      LIB_COLLECTION => 'lib_collection_target',
      DLL => 'dll_target',
      EXE => 'exe_target',
      COMPOSITE => 'composite_target'
  }

  # target_type checker.
  #
  # Raises exception if target_type has unsupported value.
  def TemplateParams.check_target_type( target_type )
    if nil == @@setup_target_functions.fetch( target_type, nil )
      raise InvalidValueEx.new( "unsupported target_type: #{target_type}" )
    end
  end

  # For a case when target_name is undetectable.
  YOUR_TARGET_NAME = 'your target name'
  # For a case when implib_path is undetectable.
  YOUR_IMPLIB_PATH = 'your import library path'

  # Type of target.
  attr_reader :target_type
  # Name of target.
  attr_reader :target_name
  # Name of import library path.
  attr_reader :implib_path

  # Param target_type must be present in @@setup_target_functions.
  def initialize( target_type, options )
    TemplateParams.check_target_type( target_type )

    @target_type = target_type
    @options = options

    try_setup_target_name
    try_setup_implib_name
  end

  # Name of setup target functions (like lib_target, dll_target, etc).
  def setup_target_function
    @@setup_target_functions.fetch( @target_type )
  end

  # Returns true if implib_path need to be defined for this type of target.
  # Or, in case of EXE target, if implib_path defined in options.
  # (Under Windows even EXEs may have import libraries).
  def has_or_require_implib
    DLL == @target_type || nil != @options.implib_path
  end

  # Returns binding to use in ERb generation.
  def get_binding
    binding
  end

private
  # Try to setup name of target from options.
  #
  # If --target specified target name is gotten from it.
  # Otherwise if project path specified then name is constructed from it.
  # Otherwise value 'your target name' is used.
  #
  def try_setup_target_name
    @target_name = if @options.target_name
        @options.target_name
      elsif @options.project_path
        @options.project_path.gsub( /[\\\/]/, '.' ) 
      else
        YOUR_TARGET_NAME
      end
  end

  # Try to setup name of implib path from options.
  #
  # If --implib-path specified then its value used.
  # Otherwise value 'your import library path' is used.
  #
  def try_setup_implib_name
    @implib_path = @options.implib_path ?
        @options.implib_path :
        YOUR_IMPLIB_PATH
  end

end # class TemplateParams

# Main class for code generation of C/C++ projects.
#
# Usage:
#   receiver = StdReceiver.new
#   generator = Generator.new( target_type, args, receiver )
#   generator.run
#
class Generator
  def initialize( target_type, args, receiver )
    @target_type = target_type
    @args = args
    @receiver = receiver
  end

  def run
    options = Options.parse( @args,
        "Stubs for C/C++ projects generator\n" +
        target_specific_banner_line + "\n",
        :implib_path => ( EXE == @target_type || DLL == @target_type ) )
    result = do_generation( options )
    @receiver.receive( result, options.output_file )
  end

private
  # Target specific parts of banner.
  TARGET_SPECIFIC_BANNERS = {
    LIB => [ 'Lib target generator', 'cpp-lib' ],
    LIB_COLLECTION =>
        [ 'LibCollection target generator', 'cpp-lib-collection' ],
    DLL => [ 'DLL target generator', 'cpp-dll' ],
    EXE => [ 'EXE target generator', 'cpp-exe' ],
    COMPOSITE => [ 'Composite target generator', 'cpp-composite' ]
  }

  # Returns part of banner which depends from target type.
  def target_specific_banner_line
    parts = TARGET_SPECIFIC_BANNERS.fetch( @target_type, nil )
    if parts
      "#{parts[0]}\n\nUsage:\nmxxrugen [<mxxrugen-options>] " +
          "#{parts[1]} [<options>]\n"
    else
      ''
    end
  end

  # Performs main generation actions.
  #
  # Returns generation result as String.
  def do_generation( options )
    template = IO.read( File.join( File.dirname( __FILE__ ), 'template.erb' ) )
    generator = ERB.new( template )

    params = TemplateParams.new( @target_type, options )
    generator.result( params.get_binding ).gsub( /\n\n\n+/, "\n\n" )
  end
end

# Helper method for generation.
#
# Usage:
#   # For using StdReceiver.
#   generate_for( EXE, ARGV )
#
#   # For using custom Receiver.
#   generate_for( EXE, ARGV, CustomReceiver.new )
#
def Cpp.generate_for( target_type, args, receiver = nil )
  generator = Generator.new( target_type, args,
      receiver ? receiver : StdReceiver.new )
  generator.run
end

end # module Cpp

end # module Impl

end # module Generators

end # module MxxRu

