#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

# Since v.1.4.0
#
# Implementation of standard object that send result of code generation
# to standard output stream or into specified output file.

require 'fileutils'

require 'mxx_ru/util'

module MxxRu

module Generators

module Impl

# Objects of this class will send result of code generation to standard
# output stream or into specified output file.
#
# Usage:
#   receiver = StdReceiver.new
#   # Send result to $stdout.
#   receiver.receive( result, nil )
#   # Send result to file.
#   receiver.receive( result, 'some_file' )
#
class StdReceiver

  # Sends _result_ to output file if _output_file_name_ isn't nil.
  #
  # If _result_ is going to output file then:
  # 1. Temporary file will be created.
  # 2. Result will be written to temporary file.
  # 3. Output file will be deleted if exists.
  # 4. Temporary file will be moved to output file.
  def receive( result, output_file_name )
    if output_file_name
      tmp_file_name = MxxRu::Util::TmpFiles.instance.create( result )
      FileUtils.rm( output_file_name, :force => true )
      FileUtils.cp( tmp_file_name, output_file_name )
    else
      $stdout.write( result )
    end
  end

end # class StdReceiver

end # module Impl

end # module Generators

end # module MxxRu

