#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/abstract_target.rb'

module MxxRu

# A Generator, which allows to define simple make rules in a project.
#
# For example, if it's needed to get from ddl file two other files, 
# cpp and hpp, using typegen program, it should be done like that:
#  generator( MxxRu::MakestyleGenerator.new(
#    [ "myfile.cpp", "myfile.hpp" ],
#    [ "myfile.ddl" ],
#    [ "typegen --from myfile.ddl --cpp myfile.cpp --header myfile.hpp" ] ) )
#
# NOTE: For C++ projects current value of sources_root isn't taken into account.
class MakestyleGenerator < AbstractGenerator
  # File names list, which should be built by given generator.
  attr_reader :target_files

  # File names list, result files depends from.
  attr_reader :dependencies

  # Command list, which should be ran in a sequence to get result files.
  attr_reader :build_cmds

  # Command list, which should be ran in a sequence to remove result files.
  #
  # If this list is empty, removal of result files would occur on clean
  # operation.
  attr_reader :clean_cmds

  # Constructor.
  #
  # [_a_target_files_] Result files list.
  # [_a_dependencies_] Dependencies list.
  # [_a_build_cmds_] Command list, which should be ran in a sequence to get
  #                  result files.
  # [_a_clean_cmds_] Command list, which should be ran in a sequence to remove
  #                  result files.
  #
  # NOTE: Single values may be defined as a strings instead of vectors. For
  # example:
  #  MxxRu::MakestyleGenerator.new( "myfile.cpp", "myfile.ddl",
  #    [ "typegen --from myfile.ddl --to myfile.cpp" ] )
  # It's the same as:
  #  MxxRu::MakestyleGenerator.new( [ "myfile.cpp" ], [ "myfile.ddl" ],
  #    "typegen --from myfile.ddl --to myfile.cpp" )
  #
  def initialize(
    a_target_files,
    a_dependencies,
    a_build_cmds,
    a_clean_cmds = Array.new )

    @target_files = make_array( a_target_files )
    @dependencies = make_array( a_dependencies )
    @build_cmds = make_array( a_build_cmds )
    @clean_cmds = make_array( a_clean_cmds )
  end

  # Perform the build of result files.
  #
  # NOTE: _target_ argument is ignored.
  def build( target )
    need_build = false

    # Checking all result files. If one of them is obsolete, performing rebuild.
    index = 0
    while !need_build && index != @target_files.size
      if TargetState::EXISTS != TargetState.detect(
        @target_files[ index ], @dependencies ).state
        need_build = true
      else
        index += 1
      end
    end

    if need_build
      # It's required to run all commands to build target.
      AbstractTarget.run(
          @build_cmds,
          [],
          "building #{@target_files.join(', ')}" )
    end
  end

  # Perform the cleanup of result files.
  #
  # NOTE: _target_ argument is ignored.
  def clean( target )
    # If cleanup commands weren't set, deleting result files.
    # Otherwise, it's required to run all commands in a sequence,
    # ignoring return codes.
    if 0 != @clean_cmds.size
      @clean_cmds.each do |c|
        system( c )
      end
    else
      @target_files.each do |f|
        Util.delete_file( f )
      end
    end
  end

protected
  # Conversion of single object into a vector, which contains that object.
  # If source object is a vector, it's returned without any operations on it.
  def make_array( a_object )
    if a_object.instance_of? Array
      return a_object
    end

    return [ a_object ]
  end
end

# For compatibility with previous versions.
Makestyle_generator = MakestyleGenerator
end
