#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'mxx_ru/ex'
require 'mxx_ru/abstract_target'

module MxxRu

# The class of a target, which is binary unit-test application, which create
# one or many text files.
#
# Given target performs a build of application, and then runs it and compares
# results with files which contains correct results.
#
# The basic idea consists in the presence of project file, which controls the
# build of unit-test application. In that project file, target object is
# created, inherited from MxxRu::BinaryTarget.  To start unit-test it's
# necessary to create one more file, where target object of
# MxxRu::TextfileUnittestTarget class is created. 
# For example:
#
# Unit-test application build file:
#  MxxRu::setup_target(
#    MxxRu::Cpp::ExeTarget.new( "test/pack/prj.rb" ) {
#      ...
#    }
#  )
#
# File to run unit-test:
#  MxxRu::setup_target(
#    MxxRu::TextfileUnittestTarget.new( 
#      "test/pack/pack.ut.rb",
#      "test/pack/prj.rb" ) {
# 
#      launch( "--loops 16 --out test/pack/out/16.txt",
#        [ pair( "test/pack/out/16.txt",
#            "test/pack/out/etalon/16.txt" ) ] )
# 
#      launch( "--loops 32 --out test/pack/out/32.txt",
#        [ pair( "test/pack/out/32.txt",
#            "test/pack/out/etalon/32.txt" ) ] )
#    }
#  )
#
# File which is using that unit-test.
#  MxxRu::setup_target(
#    MxxRu::Cpp::CompositeTarget.new( MxxRu::BUILD_ROOT ) {
#      required_prj( "some/project/prj.rb" )
#      required_prj( "test/pack/prj.ut.rb" )
#    }
#  )
#
class TextfileUnittestTarget < AbstractTarget

  # Exception, thrown if one subordinated project generated more
  # than one result name.
  class MustBeOneTargetNameEx < Ex
    # [_a_names_] Names generated.
    def initialize( a_names )
      super( "Only one name expected in [#{a_names.join(';')}]" )
    end
  end

  # For compatibility with previous versions.
  Must_be_one_target_name_ex = MustBeOneTargetNameEx

  # Exception, thrown if files are not equal.
  class FileMismatchEx < BuildEx
    # [_a_cmdline_] Command string of test application.
    # [_a_produced_] Name of file generated.
    # [_a_etalon_] Name of file containing correct results.
    # [_a_line_num_] String number where first mismatch was detected.
    # [_a_produced_line_] String of generated file. Nil if EOF is reached.
    # [_a_etalon_line_] String of correct file. Nil if EOF is reached.
    def initialize( a_cmdline,
      a_produced, a_etalon,
      a_line_num,
      a_produced_line, a_etalon_line )
      a_produced_line = "<EOF>" if !a_produced_line
      a_etalon_line = "<EOF>" if !a_etalon_line

      super( a_cmdline, "Mismatch found in line #{a_line_num}. " +
        "Line in '#{a_produced}' is '#{a_produced_line}'. " +
        "Line in '#{a_etalon}' is '#{a_etalon_line}'" )
    end
  end

  # For compatibility with previous versions.
  File_mismatch_ex = FileMismatchEx

  # Description of single pair of values compared.
  class Pair
    # Name of file generated.
    attr_reader :produced
    # Name of correct file.
    attr_reader :etalon

    # Constructor.
    def initialize( a_produced, a_etalon )
      @produced = a_produced
      @etalon = a_etalon
    end
  end

  # A single run of test application description.
  class Launch
    # Command line parameters.
    # Empty string if no parameters.
    attr_reader :cmdline
    # Files list to compare after application executed.
    # Should be Array of Pair.
    attr_reader :pairs

    # Constructor.
    def initialize( a_cmdline, a_pairs )
      @cmdline = a_cmdline
      @pairs = a_pairs
    end
  end

  # Target, responsible for build of unit-test application.
  attr_reader :mxx_target_project
  # True if build method was already executed.
  attr_reader :mxx_build_state
  # Description of test application runs.
  # Array of Launch.
  attr_reader :mxx_launches

  # Constructor.
  #
  # [_a_alias_] Self alias.
  # [_a_target_project_] Project, responsible for build of unit-test application.
  def initialize( a_alias, a_target_project, &block )
    super( a_alias )

    @mxx_build_state = nil

    @mxx_target_project = required_prj( a_target_project )

    @mxx_launches = Array.new

    instance_eval( &block ) if block
  end

  # Create description of file pairs to compare.
  #
  # [_a_produced_] Name of generated file.
  # [_a_etalon_] Name of correct file.
  def pair( a_produced, a_etalon )
    return Pair.new( a_produced, a_etalon )
  end

  # Add description of sequental run of test application.
  #
  # [_a_cmdline_] Command line parameters. May be empty string of nil 
  #               if no parameters required.
  # [_a_pairs_] A list of pairs of file names to compare. Should be Array of
  #             values returned by pair() method.
  #
  # Comparation will be performed in order they are defined in _a_pairs_.
  # Comparation is interrupted on first mismatch.
  def launch( a_cmdline, a_pairs )
    a_cmdline = "" if !a_cmdline
    @mxx_launches << Launch.new( a_cmdline, a_pairs )
  end

  # Build subordinated project and run it.
  #
  # If unit-test application returns exit code not equal to 0, exception is
  # thrown.
  def build
    if !@mxx_build_state
      @mxx_target_project.build

      # Determining the name of application, which should be run.
      full_names = @mxx_target_project.mxx_full_targets_names
      if 1 != full_names.size
        raise MustBeOneTargetNameEx.new( full_names )
      end

      # No execution of commands in dry-run mode.
      if !MxxRu::Util::Mode.instance.is_dry_run
        puts "running unit test: #{full_names[0]}..."
        do_all_launches( full_names[ 0 ] )
      end

      @mxx_build_state = TargetState.new( TargetState::REBUILT )
    end

    return @mxx_build_state
  end

  # Cleanup the project.
  #
  # Executes clean method from subordinated project.
  # Deletes all files, defined as result files of unit-test application.
  def clean
    @mxx_target_project.clean

    @mxx_launches.each { |launch|
      launch.pairs.each { |pair|
        Util.delete_file( pair.produced )
      }
    }
  end

  # Reset build status.
  def reset
    @mxx_target_project.reset
    @mxx_build_state = nil
  end

protected
  # Perform all defined runs of test application.
  # This method should not be executed in dry-run mode.
  def do_all_launches( a_file_to_exec )
    # On mswin32 platform if file name is in unix format and
    # redirection of input/output is set, some problems with
    # execution would occur.
    native_file_name = Util::native_pathname( a_file_to_exec )

    @mxx_launches.each { |launch|
      cmdline = native_file_name + " " + launch.cmdline 
      puts "\tlaunching '#{cmdline}'..."
      if !system( cmdline )
        puts "\n\nunit test '#{cmdline}' FAILED! #{$?}"
        raise BuildEx.new( cmdline, $? )
      end

      # Execution of application was successful.
      # Let's compare files.
      launch.pairs.each { |pair|
        puts "\t\tcomparing '#{pair.produced}' and '#{pair.etalon}'"
        do_file_compare( cmdline, pair.produced, pair.etalon )
      }
    }
  end

  # Perform comparation of two files.
  # If they are not identical, FileMismatchEx exception is thrown.
  # If some system exceptions would occur (absence of file for example),
  # BuildEx ewxception will be thrown.
  def do_file_compare( a_cmdline, a_produced, a_etalon )
    begin
      produced_file = File.new( a_produced, "r" )
      etalon_file = File.new( a_etalon, "r" )

      line_num = 1
      while true
        l1 = produced_file.gets
		  l1.chop! if l1
        l2 = etalon_file.gets
		  l2.chop! if l2

        if l1 != l2
          raise FileMismatchEx.new( a_cmdline,
            a_produced, a_etalon, line_num,
            l1, l2 )
        else
          line_num += 1
          if !l1
            break
          end
        end
      end

    rescue Exception => ex
      raise BuildEx.new( a_cmdline,
        "Error during comparing files " +
        "'#{a_produced}', '#{a_etalon}': #{ex}" )
    end
  end
end

# For compatibility with previous versions.
Textfile_unittest_target = TextfileUnittestTarget

end # module MxxRu
