#--
# Copyright (c) 1996-2004, Yauheni Akhotnikau
# Copyright (c) 2004-2006, JSC Intervale
# Copyright (c) 2006-2015, The Mxx_ru Project
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#++

require 'rbconfig'
require 'singleton'
require 'pathname'

require 'mxx_ru/ex'
require 'mxx_ru/cmd_line_option_processor'

module MxxRu

  module Util
    # Regular expression to divide file name and it's extension.
    MATCH_FILE_EXT_REGEX = Regexp.new( "^(.+)\\.([^\\./]+)$" )

    # Ruby command line parameters, handled by MxxRu.
    MXXARG_CLEAN = '--mxx-clean'
    MXXARG_REBUILD = '--mxx-rebuild'
    MXXARG_SHOW_CMD = '--mxx-show-cmd'
    MXXARG_KEEP_TMPS = '--mxx-keep-tmps'
    MXXARG_SHOW_TMPS = '--mxx-show-tmps'
    MXXARG_DRY_RUN = '--mxx-dry-run'
    MXXARG_BRIEF_DESC = '--mxx-brief-show'
    MXXARG_BRIEF_DESC_DISABLED = '--mxx-brief-hide'

    # OS name, script is run on.
    @@host_os = nil

    # Regular expression to parse entries in Kernel#caller result.
    CALLER_ENTRY_REGEX = /^(.+):\d+(:in\s.+){0,1}$/

    # Class, which detects MxxRu current mode.
    class Mode
      include Singleton

      # Option parser for common Mxx_ru options.
      #
      # Since v.1.4.0
      class OptionParser
        include Singleton

        # Exception for case when '--mxx-clean' and '--mxx-rebuild' used
        # together.
        class CleanAndRebuildUsedTogetherEx < Ex
          def initialize
            super "Mode #{MXXARG_CLEAN} and #{MXXARG_REBUILD} cannot be " +
                "used together"
          end
        end

        attr_reader :is_brief_desc
        attr_reader :is_brief_desc_disabled
        attr_reader :is_clean
        attr_reader :is_dry_run
        attr_reader :is_keep_tmps
        attr_reader :is_rebuild
        attr_reader :is_show_cmd
        attr_reader :is_show_tmps

        def initialize
          @is_brief_desc = false
          @is_brief_desc_disabled = false
          @is_clean = false
          @is_dry_run = false
          @is_keep_tmps = false
          @is_rebuild = false
          @is_show_cmd = false
          @is_show_tmps = false
        end

        def prepare( parser )
          parser.separator ''
          parser.separator 'Common Mxx_ru options:'

          parser.on( MXXARG_CLEAN, 'Clean up project' ) do
            @is_clean = true
            check_clean_and_rebuild_correctness
          end

          parser.on( MXXARG_REBUILD,
              'Clean up and then build project again' ) do
            @is_rebuild = true
            check_clean_and_rebuild_correctness
          end

          parser.on( MXXARG_SHOW_CMD,
              'Show commands during build process' ) do
            @is_show_cmd = true
          end

          parser.on( MXXARG_KEEP_TMPS,
              'Keep temporary files after build finihed' ) do
            @is_keep_tmps = true
          end

          parser.on( MXXARG_SHOW_TMPS,
              'Show content of temporary files during build process' ) do
            @is_show_tmps = true
          end

          parser.on( MXXARG_BRIEF_DESC,
              'Enable displaying short description of build steps' ) do
            @is_brief_desc = true
          end

          parser.on( MXXARG_BRIEF_DESC_DISABLED,
              'Disable displaying short description of build steps' ) do
            @is_brief_desc_disabled = true
          end

          parser.on( MXXARG_DRY_RUN,
              'Build imitation mode, no actual actions perfomed' ) do
            @is_dry_run = true
          end
        end

      private
        # Throws exception if --mxx-clean and --mxx-rebuild specified
        # at the same time.
        def check_clean_and_rebuild_correctness
          raise CleanAndRebuildUsedTogetherEx.new if @is_clean && @is_rebuild
        end
      end # class OptionParser

      # Method for generation methods like is_clean, is_rebuild, etc.
      def Mode.generate_option_reader( option )
        class_eval do
          define_method( option ) do
            OptionParser.instance.send( option )
          end
        end
      end

      generate_option_reader :is_clean
      generate_option_reader :is_rebuild
      generate_option_reader :is_show_cmd
      generate_option_reader :is_keep_tmps
      generate_option_reader :is_show_tmps
      attr_reader :is_brief_desc

      # Constructor checks for special arguments in command line.
      def initialize
        CmdLineOptionProcessor.instance.parse

        @is_brief_desc = OptionParser.instance.is_brief_desc
        @is_brief_desc = false if OptionParser.instance.is_brief_desc_disabled

        @is_dry_run = OptionParser.instance.is_dry_run

        @is_manual_dry_run = false
      end

      # Is dry-run mode set?
      # Dry-run mode is set, if --mxx-dry-run command line parameter is set,
      # or if it's manually set.
      def is_dry_run
        return ( @is_dry_run || @is_manual_dry_run )
      end

      # On/Off manual dry-run mode.
      # In some cases it's efficient to turn it on for some time.
      # For example, it's used to handle subprojects with --mxx-cpp-1 option
      # set.
      def manual_dry_run( a_is_manual_dry_run )
        @is_manual_dry_run = a_is_manual_dry_run
      end

      # Try to manually enable 'show-brief' mode. Attempt ignored if
      # '--mxx-show-brief-disabled' was specified.
      def try_enable_show_brief
        @is_brief_desc = true unless OptionParser.instance.is_brief_desc_disabled
      end

    end

    # Returns file name without last extension.
    def Util.remove_file_ext( a_name )
      match_result = MATCH_FILE_EXT_REGEX.match( a_name )
      if nil == match_result
        # No extension was present.
        return String.new( a_name )
      end

      return match_result[ 1 ]
    end

    # Returns filename with new last extension.
    def Util.change_file_ext( name, ext )
      remove_file_ext( name ) + ext
    end

    # Deleting file name without exception.
    def Util.delete_file( a_name )
      if FileTest.exist?( a_name )
        File.delete( a_name )
      end
    end

    # Class, responsible for creation and removal of temporary files.
    # Temporary files are created in current folder.
    # Temporary files are removed before application is exitted,
    # if --mxx-keep-tmps option isn't set.
    class TmpFiles
      include Singleton

      # Temporary file names.
      @@names = Array.new

      def initialize
        if !Mode.instance.is_keep_tmps
          ObjectSpace.define_finalizer( self, Proc.new{ TmpFiles.finalizer } )
        end

        # Counter to ensure unique file names.
        @current = 0

        # Indicies stack in @@names for push and pop methods.
        @index_stack = Array.new
      end

      def TmpFiles.finalizer
        @@names.each { |d|
            begin
              File.delete( d )
            rescue Exception => e
              puts e
            end
          }
      end

      # Temporary file creation.
      # The string given is stored to the temporary file.
      # The name of temporary file is returned.
      #
      # [_a_content_] Should be an object of String type.
      def create( a_content )

        current = (@current += 1)
        file_name = "tmpmxx_ru.#{$$}.#{current}"
        file = File.new( file_name, "w" )
        @@names << file_name
        file << a_content
        file.close

        if Mode.instance.is_show_tmps
          puts "<<<[#{file_name}]\t #{a_content}>>>"
        end

        return file_name
      end

      # Storing current vector size to the indicies stack.
      def push()
        if !Mode.instance.is_keep_tmps
          @index_stack.push( @@names.size() )
        end
      end

      # Removing all files, which are stored in a file names vector
      # starting from first element from @index_stack position.
      def pop()
        if !Mode.instance.is_keep_tmps
          index = @index_stack.pop
          while @@names.size > index
            file_to_delete = @@names.pop
            MxxRu::Util::delete_file( file_to_delete )
          end
        end
      end

    end # class TmpFiles

    # For compatibility with previous versions.
    Tmp_files = TmpFiles

    # Auxiliary function for correct execution of build method of target
    # object.  Before execution TmpFiles.push is executed, and after execution
    # of build method TmpFiles.pop is called. This way we are removing
    # temporary files created during a compilation of given target.
    def Util.build_call_wrapper( a_target )
      begin
        TmpFiles.instance.push
        state = a_target.build
        return state
      ensure
        TmpFiles.instance.pop
      end
    end

    # Create folder if it isn't exist.
    #
    # All chain of names is tracked. For example, if we trying to create
    # output/release/lib folder, and only output is exists, then release/lib
    # folders are created.
    def Util.create_dir( a_name )
      # Dividing the chain.
      chain = Array.new
      chain.unshift( a_name )

      to_split = a_name.clone
      while to_split != '.' 
        r = File.split( to_split )

        # We must stop if to_split cannot be splitted anymore
        # (when working with absolute paths in to_split can be drive name
        # on Windows platform (e.g. 'D:\') or root directory on Unix).
        # In this case r will be equal to [ <something>, '/' ].
        break if 2 == r.size and ( '\\' == r[ 1 ] or '/' == r[ 1 ] )

        chain.unshift( to_split = r[ 0 ] )
      end

      chain.each { |d|
        if !FileTest.exist?( d )
          Dir.mkdir( d )
        end
      }
    end

    # To ensure folder is exist.
    #
    # If dry-run mode wasn't set, checking for existance.
    # If it isn't exists, trying to create it.
    #
    # In a dry-run mode MxxRu always assumes folder is exists, even if it
    # isn't.
    def Util.ensure_path_exists( a_path_name )
      if !MxxRu::Util::Mode.instance.is_dry_run
        if !FileTest.exist?( a_path_name )
          MxxRu::Util.create_dir( a_path_name )
        end
      end
    end

    # Get OS name where script is exeuted.
    def Util.host_os
      if !@@host_os
        @@host_os = RbConfig::CONFIG[ "host_os" ]
      end

      return @@host_os
    end

    # This flag is set to true on Windows.
    #
    # Since v.1.5.1
    #
    IS_WINDOWS_PLATFORM = !((RbConfig::CONFIG['host_os'] =~ /mswin|mingw/).nil?)

    # Transform separators in file names to correct ones on the given platform.
    #
    # For example on mswin32 we change "/" to "\".
    #
    # [_a_pathname_] Path to the file in unix style.
    #
    # *IMPORTANT! In current version only mswin32 platform is tracked!*
    def Util.native_pathname( a_pathname )
      if IS_WINDOWS_PLATFORM
        return a_pathname.gsub( "/", "\\" )
      end

      return a_pathname
    end

    # Extract file name from 'Kernel#caller' result for using this
    # name as project alias.
    def Util.prj_alias_form_caller( caller_result )
      r = CALLER_ENTRY_REGEX.match( caller_result[ 0 ] )[ 1 ]
      # Name can starts with './' ('./some/module/prj.rb'). That name
      # must be transformed to 'some/module/prj.rb'.
      r = if r[ 0..1 ] == './'
          r[ 2...r.size ]
        else
          r
        end
      # Path could be absolute. We should extract current path name from it.
      p = Pathname.new( r )
      if p.absolute?
        p = p.relative_path_from( Pathname.pwd )
      end

      p.to_s
    end

  end # module Util

  # Manually enable 'show-brief' mode.
  #
  # Intended to use in project files:
  #   MxxRu::Cpp::exe_target( 'some.prj' ) {
  #     ...
  #     MxxRu::enable_show_brief
  #     ...
  #   }
  # 'show-brief' is global mode. Once enabled it cannot be disabled
  # form inside another project file.
  #
  # Call to MxxRu::enable_show_brief ignored if '--mxx-show-brief-disabled'
  # specified in command-line.
  #
  def MxxRu.enable_show_brief
    Util::Mode.instance.try_enable_show_brief
  end

  # Show a brief message. The message shown only if 'show-brief' mode is on.
  def MxxRu.show_brief( message )
    puts message[ 0, 1 ].capitalize << message[ 1..message.size ] << ' ...' if
        Util::Mode.instance.is_brief_desc and message
  end

end # module MxxRu

# Common Mxx_ru option parser must be added.
MxxRu::CmdLineOptionProcessor.instance.add_processor(
    MxxRu::Util::Mode::OptionParser.instance )

